(* :Author: Chris Williamson *)

BeginPackage["WebServices`", {"JLink`"}];

WebServices`Information`$Version = "Web Services Version 1.0b1 (April 2003)";

WebServices`Information`$VersionNumber = 1.0;

WebServices`Information`$ReleaseNumber = 0;

(* Functions *)

InstallService::usage=
	"InstallService[wsdlURL] installs the web service operations specified
	in the WSDL URL into a default context defined by the service name and
	port name of the service operation.
	InstallService[wsdlURL, myContext] installs the web service operations
	specified in the WSDL URL into a context specified by the second parameter.";

InstallServiceOperation::usage=
  "InstallServiceOperation[symbolName, endPoint, argumentDefs, headerDefs]
	creates a function for the web service operation defined by the end point,
	argument definitions, and header definitions.";

UninstallService::usage=
	"UninstallService[context] uninstalls all web service operations in the
	context provided.";

UninstallServiceOperation::usage=
  "UninstallServiceOperation[symbol] uninstalls the web service operation defined
	as the symbol.  The OperationPalette is cleared and the symbol is removed.";
	  
OperationPalette::usage=
  "OperationPalette[symbol] displays a palette that can be used to understand a
	web service operation.  The palette displays documentation, the function signature,
	type information, and buttons which allow the user to paste useful templates.";

ToServiceRequest::usage=
  "ToServiceRequest[parameters, headerParameters] builds a request message using
	input provided in the parameters.  ToServiceRequest[symbol, parameters___] builds
	the request message for a web service operation using input provided in the parameters
	and information linked with the symbol.";

InvokeServiceOperation::usage=
	"InvokeServiceOperation[endPointURL, requestMessage]  invokes a web service operation
	using the request message.  The message is sent to the end point specified in the
	first argument.  The response message is returned.
	InvokeServiceOperation[symbol, parameters___] invokes a web service operation using
	input provided in the parameters and information linked with the symbol.  The
	response message is returned.
	InvokeServiceOperation[symbol, requestMessage] invokes a web service operation using
	the request message and information linked with the symbol.  The response message is
	returned.";

FromServiceResponse::usage=
  "FromServiceResponse[response] converts a response message into a useful Mathematica
	expression.  For SOAP messages, the SOAP envelope and SOAP body tags are stripped
	from the results.  The values are then converted into Rule syntax.";
  
ServicesImpl::usage="ServicesImpl[name] represents a web services implemenation.";

(* Shared Utility symbols *)

Deserialize::usage = 
	"Deserialize is a utility function used to deserialize a web service response into
	a format desired by the user.";
	
Normalize::usage =
	"Normalize is a utility function that uses WSDL definitions to normalize web service
	input and output.";

Serialize::usage =
	"Serialize is a utility function used to serialize a web service request into a format
	desired by the web service.";

MapWSDL::usage = 
  "MapWSDL is a utility function used to map a WSDL file into Mathematica functions.";

GetNamespaceDefinitions::usage =
  "GetNamespaceDefinitions is a utility function that provides access to WSDL definitions.";

GeneratePalette::usage = 
  "GeneratePalette is a utility function that generates a OperationPalette palette.";

GetOperationPalette::usage = 
  "GetOperationPalette is a utility function that is used to retrieve a OperationPalette.";

InitServiceContext::usage = 
  "InitServiceContext is a utility function that is used to initialize data particular to a
	web service call.";

ClearServiceContext::usage = 
  "ClearServiceContext is a utility function that is used to clear data particular to a web
	service call.";

MapReferences::usage =
  "MapReferences is a utility function that is used to map references in a SOAP envelope to
	Mathematica symbol.";

GetHref::usage = 
  "GetHref is a utility function that is used to retrieve references from a SOAP envelope.";

NamespaceId::usage = 
  "NamespaceId is a utiltity symbol that is used to keep track of the last generic namespace
	id used while serializing a request message.";

(* Global Options *)

$ServicesImpl::usage = 	
	"$ServiceImpl specifies the default implementation that will be used to invoke Web Services.
	This should be set to the ServicesImpl object that corresponds with the Web Services
	implementation."

$InstalledServices::usage=
  "$InstalledServices is a list of the web service operations installed."

$PrintServiceRequest::usage=
	"$PrintServiceRequest uses the Mathematica Print function to print the message sent to a
	web service.  The default is False."

$PrintServiceResponse::usage=
	"$PrintServiceResponse uses the Mathematica Print function to print the message received
	from a web service before it is deserialized into a Rule syntax expression.  The default
	is False."
	
$PrintPerformanceNumbers::usage=
  "$PrintPerformanceNumbers specifies whether performance numbers will be printed.
	The default is False"

$PrintWSDLDebug::usage=
  "$PrintWSDLDebug specifies whether WSDL debugging information will be printed when installing
	a web service.  The default is False.";

$PrintShortErrorMessages::usage=
	"$PrintShortErrorMessages specifies whether error messages will be shortened for the user
	to avoid long intimidating error messages.  The default is True."  

	
(* SOAP Options *)

OperationName::usage = 
  "OperationName is an option which specifies the name of the operation tag for rpc-style
	calls and the root tag for single-document document-style calls.";

OperationStyle::usage = 
	"OperationStyle is an option which specifies the operation style of the web service
	operation.  This may be set to \"document\" or \"rpc\" only.  The default is Automatic."

SOAPActionURI::usage = 
	"SOAPActionURI is an option which specifies the SOAP action URI of the web service
	operation.  The default is Automatic."

ReturnType::usage = 
	"ReturnType is an option which specifies the return type of the Web Service operation.
	The default is Automatic."

TransportStyleURI::usage=
	"TransportStyleURI is an option which specifies the transport style URI of the web
	service operation.  The default is Automatic."

EncodingStyleURI::usage = 
	"EncodingStyleURI is an option which specifies the encoding style URI of the web service
	operation.  The default is Automatic."

HeaderEncodingStyleURI::usage=
  "HeaderEncodingStyleURI is an option which specifies the encoding style URI of the web
	service headers.  The default is Automatic."

EncodingStyle::usage=
  "EncodingStyle is an option which specifies whether the message should be encoded or use a message
	based on literal schema.  This may be set to \"encoded\" or \"literal\" only.  The default
	is Automatic."
  
Username::usage=
  "Username is an option used to authenticate a user of a Web Service."
  
Password::usage=
  "Password is an option used to authenticate a user of a Web Service."
  
(* Messaging Options *)

InvokeServiceOperation::"native" = 
	"An error occurred: `1`"

InvokeServiceOperation::"env" = 
	"Invalid SOAP envelope: `1`"

InvokeServiceOperation::"prtcl" = 
	"Incorrect protocol used in end point URL: `1`.  Should be: `2`"

InvokeServiceOperation::"opttype" = 
	"`1` is not the correct type for Option `2` "

InvokeServiceOperation::"opstyle" = 
	"OperationStyle must be set to 'rpc' or 'document'."

InvokeServiceOperation::"rspns" = 
	"Invoke Failed.  Response is $Failed or Null."

InvokeServiceOperation::"rspnsdoc" = 
	"Response cannot be parsed:\n `1`"

InvokeServiceOperation::"rspnsenv" = 
	"SOAP Envelope cannot be parsed from response:\n `1`"

InvokeServiceOperation::"rspnsbdy" = 
	"SOAP Body cannot be parsed from response:\n `1`"

InvokeServiceOperation::"rspnsrtn" = 
	"SOAP Return Type cannot be parsed from response:\n `1`"

InvokeServiceOperation::"rspnsflt" = "SOAPFault occurred: `1`"

InvokeServiceOperation::"srlz" = "Could not serialize: `1`"

InvokeServiceOperation::"hdrsrlz" = "Could not serialize SOAPHeader: `1`"

InvokeServiceOperation::"desrlz"= "Could not deserialize: `1`"

InvokeServiceOperation::"arraytype"= "Problem with arrayType: `1`"

InvokeServiceOperation::"href"= "Href not found."

InvokeServiceOperation::"mapparam" = 
	"Could not map parameters: \n`1`,\n`2`"
	
InstallService::"context" = 
	"Illegal Context: `1`, `2`"	
	
InstallService::"symbol" = 
	"Illegal Symbol (begins with a digit): `1`"	
	
InstallService::"endpnt" = 
	"The SOAP End Point URL is not defined in this port: `1`"

InstallService::"style" = "Style not supported: `1`"

InstallService::"dfntn" = 
	"The Definition element is not found for this WSDL"

InstallService::"nmspc" = 
	"The namespace is not defined for this operation: `1`"

InstallService::"opprm" = 
	"Problem parsing the operation or parameters for `1`"
	
InstallService::"elmnt" = 
	"Element not defined in part: `1`"
	
InstallService::"type" = 
	"Type not defined `1`"

InstallService::"import" = 
	"Problem importing namespace `1`"

InstallService::"prefix" = 
	"Problem finding prefix for `1`"
	
InstallService::"schema" = 
	"Problem finding schema for `1`"
	
InstallService::"prttpop" = 
  "The port type operation is Null.";

Begin["`Private`"]

$ServicesImpl = ServicesImpl["Java"];

$InstalledServices = {};

$PrintServiceRequest = False;

$PrintServiceResponse = False;
	
$PrintPerformanceNumbers = False;

$PrintShortErrorMessages = True;

$PrintWSDLDebug = False;

Options[ToServiceRequest] =
  {		
		(* SOAP Options *)
		OperationName->Automatic,
		OperationStyle -> Automatic,
		SOAPActionURI -> Automatic,
		TransportStyleURI -> Automatic,
		EncodingStyle->Automatic,
		EncodingStyleURI -> Automatic,
		HeaderEncodingStyleURI->Automatic,
		Username -> Automatic,
		Password -> Automatic
  }

Options[FromServiceResponse] = 
  {
		(* General Options *)
		LongForm -> False,
		
		(* SOAP Options *)
		ReturnType -> Automatic
  }

Options[InvokeServiceOperation] = 
  Options[ToServiceRequest]

Options[InstallServiceOperation] = 
  Union[
    {
  		AllowShortContext->True
    },
    Options[InvokeServiceOperation]
  ]
  
Options[InstallService] = 
  {
    LongForm->False,
    AllowShortContext->True
  }
  

  	
InstallService[wsdlURL_String, options___?OptionQ] := 
	InstallService[$ServicesImpl, wsdlURL, options]

InstallService[wsdlURL_String, context_String, options___?OptionQ] := 
	InstallService[$ServicesImpl, wsdlURL, context, options]

InstallServiceOperation[exprName_Symbol, 
                        endPoint_String,
                        args_List,
                        header_List,
                        options___?OptionQ] :=
  InstallServiceOperation[$ServicesImpl, exprName, endPoint, args, header, options]
  
UninstallServiceOperation[operation_Symbol] := UninstallServiceOperation[$ServicesImpl, operation]

UninstallService[context_String] := UninstallService[$ServicesImpl, context]

OperationPalette[operation_Symbol] := OperationPalette[$ServicesImpl, operation]

ToServiceRequest[parameters_List, 
								 headerParameters_List, 
								 options___?OptionQ] :=
  ToServiceRequest[$ServicesImpl, parameters, headerParameters, options];

ToServiceRequest[x___] := (
  $Failed
);

InvokeServiceOperation[endPoint_String, 
                       message:XMLObject["Document"][___], 
                       options___?OptionQ] :=
	InvokeServiceOperation[$ServicesImpl, endPoint, message, options]


FromServiceResponse[response:XMLObject["Document"][{___}, _XMLElement, ___],
                        options___?OptionQ] :=
  FromServiceResponse[$ServicesImpl, response, options];  

FromServiceResponse[x___] := (
	Message[InvokeServiceOperation::"rspnsdoc", x];
	$Failed
);

Deserialize[_ServicesImpl, x___] := (
	Message[InvokeServiceOperation::"desrlz", {x}];
	$Failed 
);

Serialize[_ServicesImpl, x___] := (
	Message[InvokeServiceOperation::"srlz", {x}];
	$Failed 
);

Normalize[_ServicesImpl, x___] := x;

Needs[ "WebServices`JavaImpl`"];

End[];

EndPackage[];