(* :Author: Chris Williamson *)

BeginPackage["WebServices`Deserialize`", {"JLink`",
                                          "WebServices`",	
                                          "WebServices`WSDL`",
							            								"WebServices`Utilities`",
							            								"WebServices`SchemaUtilities`",
							            								"WebServices`Normalize`"}]

Begin["`Private`"]

(* Serialize and Deserialize *) 
base64 = Null;

base64Decode[str_String] := (
  If[base64 === Null, base64 = JavaNew["org.apache.axis.encoding.Base64"]];
  JavaObjectToExpression[JavaNew["java.lang.String", base64@decode[str], 0]]
);

XMLSchemaTypeQ[x_String] :=   
	StringMatchQ[x, "boolean"] || 
	StringMatchQ[x, "int"] || 
	StringMatchQ[x, "integer"] ||
	StringMatchQ[x, "double"] ||
  StringMatchQ[x, "float"] || 
  StringMatchQ[x, "string"] || 
  StringMatchQ[x, "short"] || 
  StringMatchQ[x, "long"] || 
  StringMatchQ[x, "decimal"] || 
  StringMatchQ[x, "byte"] || 
  StringMatchQ[x, "nonPositiveInteger"] || 
  StringMatchQ[x, "negativeInteger"] || 
  StringMatchQ[x, "nonNegativeInteger"] || 
  StringMatchQ[x, "unsignedLong"] || 
  StringMatchQ[x, "unsignedInt"] || 
  StringMatchQ[x, "unsignedShort"] || 
  StringMatchQ[x, "unsignedByte"] || 
  StringMatchQ[x, "positiveInteger"] || 
  StringMatchQ[x, "duration"] || 
  StringMatchQ[x, "datetime"] || 
  StringMatchQ[x, "time"] || 
  StringMatchQ[x, "date"] || 
  StringMatchQ[x, "gYearMonth"] || 
  StringMatchQ[x, "gYear"] || 
  StringMatchQ[x, "gMonthDay"] || 
  StringMatchQ[x, "gDay"] || 
  StringMatchQ[x, "gMonth"] || 
  StringMatchQ[x, "base64Binary"] || 
  StringMatchQ[x, "hexBinary"] || 
  StringMatchQ[x, "anyURI"] || 
  StringMatchQ[x, "QName"] || 
  StringMatchQ[x, "normalizedString"] || 
  StringMatchQ[x, "token"] || 
  StringMatchQ[x, "language"] || 
  StringMatchQ[x, "Name"] || 
  StringMatchQ[x, "NCName"] || 
  StringMatchQ[x, "base64"]

XMLSchemaTypeQ[___] := False

Deserialize[ServicesImpl["Java"],
            m:XMLElement[name:(_String | {_String, _String}), 
                         attrs:{___}, 
                         {value_}], 
           {_?XMLSchemaStringQ, type_?XMLSchemaTypeQ}, 
           _List,
	         options___?OptionQ] :=
	Module[{prefix, val, longForm, returnName, 
	        opts = Flatten[{options}]}, 

		(* Get options *)
		{longForm, returnName} = 
			{LongForm, "ReturnName"} /. 
				opts  /. Options[InvokeServiceOperation];

		(* determine whether to return the name and/or type with the value *)
		If[longForm === True && returnName =!= False, 
			prefix = {name, attrs}, 
			If[returnName =!= False, 
				If[ListQ[name], prefix = Last[name], prefix = name], 
				prefix = Null
			];
		];
				
		(* Do the conversion *)
		Switch[ 
			type, 
			_/;StringMatchQ[type, "boolean"], 
			  If[StringMatchQ[value, "True", IgnoreCase->True], val = True, val = False],
			_/;StringMatchQ[type, "double"], val = ToExpression[value],
			_/;StringMatchQ[type, "float"], val = ToExpression[value],
			_/;StringMatchQ[type, "int"], val = ToExpression[value],
			_/;StringMatchQ[type, "integer"], val = ToExpression[value],
			_/;StringMatchQ[type, "long"], val = ToExpression[value],
			_/;StringMatchQ[type, "short"], val = ToExpression[value],
			_/;StringMatchQ[type, "decimal"], val = ToExpression[value],
			_/;StringMatchQ[type, "byte"], val = ToExpression[value]; 
																		 If[val < 0, val = val + 255];,
			_/;StringMatchQ[type, "nonPositiveInteger"], val = ToExpression[value],
			_/;StringMatchQ[type, "negativeInteger"], val = ToExpression[value],
			_/;StringMatchQ[type, "nonNegativeInteger"], val = ToExpression[value],
			_/;StringMatchQ[type, "unsignedLong"], val = ToExpression[value],
			_/;StringMatchQ[type, "positiveInteger"], val = ToExpression[value],
			_/;StringMatchQ[type, "unsignedInt"], val = ToExpression[value],
			_/;StringMatchQ[type, "unsignedShort"], val = ToExpression[value],
			_/;StringMatchQ[type, "unsignedByte"], val = ToExpression[value],
			_/;StringMatchQ[type, "base64Binary"], val = base64Decode[value],
			_/;StringMatchQ[type, "base64"], val = base64Decode[value],
			_, val = value
		];

		(* Return value *)
		If[MatchQ[prefix, Null], 
			val,
			prefix->val
		]
	]

getHrefId[m:XMLElement[(_String | {_String, _String}), 
                       {___, "href"->id_String,___}, 
                       {___}]] := StringDrop[id, 1]

getHrefId[___] := Null

(* Deserialize Array XMLElement *)
Deserialize[ServicesImpl["Java"],
            m:XMLElement[name:(_String | {_String, _String}), 
										     attrs:{___, {"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->arrayType_, ___}, 
										     children:{___}],
           (type:{_String, _String} | type:{}), 
           namespaces_List,
           options___?OptionQ] :=
  Module[{longForm, returnName, nspaces, prefix,
          namespace = "", t = Null, schema, value, n, typeSchema, 
          schemaNamespaces = {}, opts = Flatten[{options}]},
    
		(* Get options and add namespaces *)
		{longForm, returnName} = 
			{LongForm, "ReturnName"} /. 
				opts  /. Options[InvokeServiceOperation];	
		
		nspaces = MapNamespaces[m, namespaces];

		(* determine whether to return the name and/or type with the value *)
		If[longForm === True && returnName =!= False, 
			prefix = {name, attrs}, 
			If[returnName =!= False, 
				If[ListQ[name], prefix = Last[name], prefix = name], 
				prefix = Null
			];
		];
    
		opts = DeleteCases[opts, "ReturnName"->_];
    
    If[MatchQ[type, {_String, _String}], namespace  = First[type]];
 		
		If[StringQ[arrayType] && !StringMatchQ[arrayType, "*[*"], 
      t = StringTake[arrayType, {1, First[First[StringPosition[arrayType, "["]]]-1}];
      schemaNamespaces = nspaces;
 		];
 		 		
    If[MatchQ[arrayType, {_String, _String}],  		
  		If[StringMatchQ[Last[arrayType], "*[*"], 
        t = {First[arrayType], StringTake[Last[arrayType], {1, First[First[StringPosition[Last[arrayType], "["]]]-1}]};
   		];
    ];
    
	  (* If an array *)
 	  AppendTo[opts, "ReturnName"->False];
 		value = 
 			((n = Normalize[ServicesImpl["Java"], #, t, namespace, nspaces];
 			  Deserialize[ServicesImpl["Java"],
 			              n,
 		                GetType[n, MapNamespaces[n, nspaces]], 
                    nspaces, 
                    opts]) & /@ children);

    If[children === {} && 
       type =!= {"http://schemas.xmlsoap.org/soap/encoding/","array"}, 
      value = Null
    ];
                   
		(* Return value *)
		If[MatchQ[prefix, Null], 
			value,
			prefix->value
		]
  ]

(* Default Deserialize XMLElement *)
Deserialize[ServicesImpl["Java"],
            m:XMLElement[name:(_String | {_String, _String}), 
										     attrs:{___}, 
										     children:{___}],
           (type:{_String, _String} | type:{}), 
           namespaces_List,
           options___?OptionQ] :=
	Module[{longForm, nspaces, prefix, typeNS = Null, 
	        schema = Null, typeSchema = Null, value, n, schemaNamespaces = {},
	        opts = Flatten[{options}]},
    
		(* Get options and add namespaces *)
		{longForm, returnName} = 
			{LongForm, "ReturnName"} /. 
				opts  /. Options[InvokeServiceOperation];	
		
		nspaces = MapNamespaces[m, namespaces];

		(* determine whether to return the name and/or type with the value *)
		If[longForm === True && returnName =!= False, 
			prefix = {name, attrs}, 
			If[returnName =!= False, 
				If[ListQ[name], prefix = Last[name], prefix = name], 
				prefix = Null
			];
		];
    
		opts = DeleteCases[opts, "ReturnName"->_];
    
    If[MatchQ[type, {_String, _String}],
      typeNS = First[type];
   	  schema = GetNamespaceDefinitions[typeNS];
   	  schemaNamespaces = MapNamespaces[schema, {}];
  	  If[!SchemaQ[schema], 
   	    schema = GetSchema[schema, typeNS];
   	    schemaNamespaces = MapNamespaces[schema, schemaNamespaces];
      ];
      typeSchema = GetSchemaType[Last[type], schema];
      schemaNamespaces = MapNamespaces[typeSchema, schemaNamespaces];
    ];
    
    (* Do the conversion *)
    If[MatchQ[children, {_String}],
	    value = First[children],
      (* If a compound object *)
      value = 
      	((n = Normalize[ServicesImpl["Java"], #, GetElementType[GetElement[GetLocalName[#], typeSchema]], typeNS, schemaNamespaces];
      	  Deserialize[ServicesImpl["Java"],
      	              n,
                      GetType[n, MapNamespaces[n, nspaces]], 
                      nspaces, 
                      opts]) & /@ children);
    ];

    If[children === {} && 
       type =!= {"http://schemas.xmlsoap.org/soap/encoding/","array"}, 
      value = Null
    ];
                   
		(* Return value *)
		If[MatchQ[prefix, Null], 
			value,
			prefix->value
		]
	]
	
End[]

EndPackage[]
