(* :Author: Chris Williamson *)

BeginPackage["WebServices`JavaImpl`", {"JLink`",
							            				     "WebServices`",
            						               "WebServices`WSDL`",
             						               "WebServices`Deserialize`",
          						                 "WebServices`Serialize`", 
            						               "WebServices`Normalize`",
											                 "WebServices`Utilities`",
											                 "WebServices`SchemaUtilities`",
            						               "WebServices`OperationPalette`"}]

Begin["`Private`"]

(* InvokeServiceOperation *)

call = Null;

invoke[endPoint_String, env_, options___?OptionQ] :=
	Module[{service, url, response, transportURI, operationStyle, 
					soapActionURI, username, password, 
					opts = Flatten[{options}]},
	
	  If[call === Null, 
		  service = JavaNew["org.apache.axis.client.Service"];
		  call = service@createCall[];
		];
				
		(* Process options *)
		{transportURI, operationStyle, soapActionURI, username, password} = 
			{TransportStyleURI, OperationStyle, SOAPActionURI, 
			 Username, Password} /. 
			 	 opts  /. Options[InvokeServiceOperation];

		(* Add Transport URI *)

		(* Check to make sure the url contains the correct protocol *)
		If[!StringMatchQ[endPoint, "http://*"], 
			Message[InvokeServiceOperation::"prtcl", endPoint, "http"];
			Return[$Failed],
			(* Catch Exception *)	
			url = JavaNew["java.net.URL", endPoint];
			If[url === $Failed, Return[$Failed]];
			call@setTargetEndpointAddress[ url];
		];
		
		If[soapActionURI =!= Automatic,  
			If[StringQ[soapActionURI], 
				call@setSOAPActionURI[soapActionURI],
				Message[InvokeServiceOperation::"opttype", 
				        soapActionURI,
							  SOAPActionURI];				
			];
		];
										
		If[operationStyle =!= Automatic, 
			If[StringQ[operationStyle],
				call@setOperationStyle[operationStyle],
				Message[InvokeServiceOperation::"opttype", 
				        operationStyle, 
				        OperationStyle];
				Return[$Failed];
			],
			operationStyle = "rpc"
		];

		If[username =!= Automatic, 
			If[StringQ[username],
				call@setUsername[username],
				Message[InvokeServiceOperation::"opttype", 
				        username, 
				        Username];
				Return[$Failed];
			];
		];

		If[password =!= Automatic, 
			If[StringQ[password],
				call@setPassword[password],
				Message[InvokeServiceOperation::"opttype", 
				        password, 
				        Password];
				Return[$Failed];
			];
		];
				
		If[TrueQ[$PrintServiceRequest], Print[env@toString[]]];
		
    proxyHost = System`getProperties[]@getProperty["proxyHost"];
    proxyPort = System`getProperties[]@getProperty["proxyPort"];
    If[proxyHost =!= Null, 
      AxisProperties`setProperty["http.proxyHost", proxyHost];
    ];
    If[proxyPort =!= Null, 
      AxisProperties`setProperty["http.proxyPort", proxyPort];
    ];

 		(* invoke call *)
   	If[TrueQ[$PrintPerformanceNumbers], Print["Invoking: ",Date[]]];

		response = 
			If[operationStyle === Automatic || 
			   StringMatchQ[operationStyle, "rpc"] ||
			   StringMatchQ[operationStyle, "document"],
				call@invoke[ env],
				Message[InvokeServiceOperation::"opstyle"];
				Return[$Failed];
	 		];

   	If[TrueQ[$PrintPerformanceNumbers], Print["Finished Invoking: ",Date[]]];

		If[response =!= $Failed && TrueQ[$PrintServiceResponse], Print[response@toString[]]];
	   		
		(* process response *)
		If[response =!= $Failed && response =!= Null, 
			response = XML`Parser`XMLGetString[response@toString[]],
			Message[InvokeServiceOperation::"rspns"];
			Return[$Failed];
		];
		
    response
  ]

MapReferences[ServicesImpl["Java"], m_XMLElement] := 
  Module[{cases},
    cases = Cases[m, e:XMLElement[(_String | {_String, _String}), 
                                  {___, "id"->id_String, ___}, 
                                  {___}] :> {id, e}, Infinity];
    (GetHref[First[#]] = Last[#]) & /@ cases;
  ]

FromServiceResponse[ServicesImpl["Java"],
                        response:XMLObject["Document"][{___}, _XMLElement, ___],
                        options___?OptionQ] :=
  Module[{envelope, namespaces, body, return, root, type, returnType,
          typeNS = Null, schema, typeSchema = Null, nspaces = {}, 
          resp, opts = Flatten[{options}]},
		
    {returnType} = 
			{ReturnType} /. 
 			 	 opts  /. Options[FromServiceResponse];	
 			 	 
     If[returnType =!= Automatic, 
			If[!MatchQ[returnType, {_String, _String}], 
				Message[InvokeServiceOperation::"opttype", 
				        returnType, 
				        ReturnType];
				returnType = Automatic;
			];
		];
		
		(* Process SOAP Envelope *)
		If[MatchQ[response[[2]], 
		          XMLElement[
                ("Envelope" | 
                 {"http://schemas.xmlsoap.org/soap/envelope/", "Envelope"}), 
                {___}, 
                {__XMLElement}]],
			envelope = response[[2]], 
			Message[InvokeServiceOperation::"rspnsenv", 
						  response[[2]]];
			Return[$Failed]
		];
		namespaces = MapNamespaces[envelope, {}];
		
		(* Process SOAP Body *)
		body = 
			Cases[envelope[[3]], 
		        XMLElement[("Body" | 
                       {"http://schemas.xmlsoap.org/soap/envelope/",		
		                    	"Body"}), 
                       {___}, 
                       {__XMLElement}, 
                       ___], 1];
		If[Length[body] > 0, 
			body = First[body], 
			Message[InvokeServiceOperation::"rspnsbdy", 
				 			ExportString[envelope[[3]], "XML"]];
			Return[$Failed]
		];
		namespaces = MapNamespaces[body, namespaces];

		(* Process return *)
		return = Cases[body[[3]], 
		               XMLElement[__, {___}, {___XMLElement}, ___],
		               1];
		If[Length[return] > 0, 
			return = First[return], 
			Message[InvokeServiceOperation::"rspnsrtn", 
							ExportString[body[[3]], "XML"]];
			Return[$Failed]
		];
		namespaces = MapNamespaces[return, namespaces];
		
		(* Process SOAPFault *)
		(*
		fault = Cases[body[[3]], 
		              XMLElement[("respflt" | 
	                 { "http://schemas.xmlsoap.org/soap/envelope/", 
	                   "Fault"}), 
		               {___}, 
		               {__XMLElement}, 
		               ___], 1];
		If[Length[fault] > 0, 
			fault = First[fault];
			Message[InvokeServiceOperation::"rspnsflt"];
			Return[$Failed]
		];
		*)
		
		If[MatchQ[return[[3]], {__XMLElement}],
	   	resp = return[[3,1]];,
	   	Return[];
	  ];
		
		root = body;
   	
    (* Deserialize *)
   	If[TrueQ[$PrintPerformanceNumbers], Print["Deserializing: ", Date[]]];

    GetHref[x_] := Null;
    
    MapReferences[ServicesImpl["Java"], root];

    namespaces = MapNamespaces[resp, namespaces];
    type = GetType[resp, namespaces, returnType];

    If[MatchQ[type, {_String, _String}],
      typeNS = First[type];
   	  schema = GetNamespaceDefinitions[typeNS];
 	    nspaces = MapNamespaces[schema, {}];
   	  If[!SchemaQ[schema], 
   	    schema = GetSchema[schema, typeNS];
   	    nspaces = MapNamespaces[schema, nspaces];
      ];
      typeSchema = GetSchemaType[Last[type], schema];
      nspaces = MapNamespaces[typeSchema, nspaces];
    ];    
    resp = Normalize[ServicesImpl["Java"], resp, typeSchema, typeNS, nspaces];
    
    AppendTo[opts, "ReturnName"->False];

    resp = Deserialize[ServicesImpl["Java"], resp, type, namespaces, opts];
 
    Clear[GetHref];

   	If[TrueQ[$PrintPerformanceNumbers], Print["Finished Deserializing: ", Date[]]];

   	resp
	];

ToServiceRequest[ServicesImpl["Java"],
									parameters_List, 
									headerParameters_List, 
									options___?OptionQ] :=
	Module[{encodingStyleURI, headerEncodingStyleURI, encodingStyle, namespaces, 
	        params, failed, operationAttrs = {}, es = "", operation, 
	        headerAttrs = {}, headerParams, children, envelope, 
	        ns = Null, name = Null},
	        
	  NamespaceId = 1;
	  
		(* Process options *)
		{encodingStyleURI, headerEncodingStyleURI, encodingStyle, operationName} = 
			{EncodingStyleURI, HeaderEncodingStyleURI, EncodingStyle, OperationName} /. 
			 	 Flatten[{options}]  /. Options[ToServiceRequest];

   	If[TrueQ[$PrintPerformanceNumbers], Print["Building SOAPEnvelope: ", Date[]]];

    If[MatchQ[operationName, {_String, _String}],
      ns = First[operationName];
      name = Last[operationName],
			If[operationName =!= Automatic, 
			  Message[InvokeServiceOperation::"opttype", 
  		          soapActionURI,
			  			  SOAPActionURI];
			];
    ];
	
		namespaces = {{"http://www.w3.org/2001/xmlns/", "soapenv"}->"http://schemas.xmlsoap.org/soap/envelope/",
		              {"http://www.w3.org/2001/xmlns/", "soapenc"}->"http://schemas.xmlsoap.org/soap/encoding/",
		              {"http://www.w3.org/2001/xmlns/", "xsd"}->"http://www.w3.org/2001/XMLSchema",
		              {"http://www.w3.org/2001/xmlns/", "xsi"}->"http://www.w3.org/2001/XMLSchema-instance",
		              {"http://www.w3.org/2001/xmlns/", "xmlns"}->"http://www.w3.org/2001/xmlns/"};

		If[headerEncodingStyleURI =!= Automatic && headerEncodingStyle =!= {}, 
			If[MatchQ[headerEncodingStyleURI, {___String}], 
				es = StringReplace[ExportString[headerEncodingStyleURI, "Words"], {" "->"", "\n"->""}];
				AppendTo[headerAttrs, {"soapenv", "encodingStyle"}->es],
				Message[InvokeServiceOperation::"opttype", 
				        encodingStyleURI, 
				        EncodingStyleURI];
			];
		];
		headerParams = 
		  ((n = Normalize[ServicesImpl["Java"], #, GetType[#, namespaces], Null, namespaces];
			  Serialize[ServicesImpl["Java"], n, namespaces, options]) & /@ headerParameters);
		failed = Cases[headerParams, $Failed, Infinity];
		If[Length[failed] > 0, Return[$Failed]];
    
    If[ns =!= Null, AppendTo[namespaces, {"http://www.w3.org/2001/xmlns/", "ns0"}->ns]];
		(* serialize parameters *)		
		params = 
		  ((If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
			    n = Normalize[ServicesImpl["Java"], #, GetType[#, namespaces], Null, namespaces],
			    n = Normalize[ServicesImpl["Java"], #, Null, Null, namespaces];
			  ];
			  Serialize[ServicesImpl["Java"], n, namespaces, options]) & /@ parameters);
		failed = Cases[params, $Failed, Infinity];
		If[Length[failed] > 0, Return[$Failed]];

		(* Add operation *)
		If[encodingStyleURI =!= Automatic && encodingStyleURI =!= {}, 
			If[MatchQ[encodingStyleURI, {___String}], 
				es = StringReplace[ExportString[encodingStyleURI, "Words"], {" "->"", "\n"->""}];
				AppendTo[operationAttrs, {"soapenv", "encodingStyle"}->es],
				Message[InvokeServiceOperation::"opttype", 
				        encodingStyleURI, 
				        EncodingStyleURI];
			];
		];
		
		If[ns =!= Null && name =!= Null, 
      AppendTo[operationAttrs, {"xmlns", "ns0"}->ns];
      children = {XMLElement[{"ns0", name}, operationAttrs, params]},
      children = params;
		];		
		
		children = {XMLElement[{"soapenv", "Body"}, {}, children]};
		If[headerParams =!= {}, 
		  PrependTo[children, XMLElement[{"soapenv", "Header"}, headerAttrs, headerParams]]
		];
		
		(* Create envelope *)
		envelope = XMLObject["Document"][{}, 
		                                 XMLElement[{"soapenv", "Envelope"}, 
		                                            {{"xmlns", "soapenv"}->"http://schemas.xmlsoap.org/soap/envelope/",
		                                             {"xmlns", "xsd"}->"http://www.w3.org/2001/XMLSchema",
		                                             {"xmlns", "xsi"}->"http://www.w3.org/2001/XMLSchema-instance", 
		                                             {"xmlns", "soapenc"}->"http://schemas.xmlsoap.org/soap/encoding/"},
                  		                          children],
                       		           {}];
		
   	If[TrueQ[$PrintPerformanceNumbers], Print["Finished Building SOAPEnvelope: ", Date[]]];
		
		envelope
	]

convertMessageToJavaObject[message:XMLObject["Document"][___]] :=
  Module[{xml, is, envelope},
		(* convert to a string *)
		xml = ExportString[message, "XML"];
		If[xml === $Failed, 
			Message[InvokeServiceOperation::"env", message];
			Return[$Failed]
		];
			
		(* convert to an InputStream *)
		is = JavaNew["java.io.ByteArrayInputStream", 
		             MakeJavaObject[xml]@getBytes[]];
  	If[is === $Failed, 
			Message[InvokeServiceOperation::"env", xml];
			Return[$Failed]
		];
  
		(* convert to a SOAPEnvelope object *)
		envelope = JavaNew["org.apache.axis.message.SOAPEnvelope", is];
		If[envelope === $Failed, 
			Message[InvokeServiceOperation::"env", xml];
			Return[$Failed]
		];
		envelope
  ]
	
InvokeServiceOperation[ServicesImpl["Java"],
                       endPoint_String, 
                       message:XMLObject["Document"][___], 
                       options___?OptionQ] :=
  JavaBlock[
		Module[{envelope, messages, result},

			InstallJava[];
			
      Block[{JLink`Exceptions`Private`reportException},
      
        If[TrueQ[$PrintShortErrorMessages], 
          JLink`Exceptions`Private`reportException = reportShortError,
          JLink`Exceptions`Private`reportException = reportError
        ];
        
        envelope = convertMessageToJavaObject[message];
        If[envelope === $Failed, Return[$Failed]];

  			result = invoke[endPoint, envelope, options];
			];
			
			result
		], KeepObjects->{call}
	]
	

(* Used for mapping parameters definitions with their values 
 * on methods that use a SOAPCall. 
 *)

mapParameters[paramDesc_List, parameters_List] := 
	Module[{},
		If[Length[paramDesc] == Length[parameters],
			MapThread[
				If[MatchQ[#1, (_String | 
				              {{_String, _String}, {___}} | 
				              {_String, {___}})], 
					{GetName[#1], GetAttributes[#1]}->#2, 
					Message[InvokeServiceOperation::"mapparam", #1, #2];
					Return[$Failed]
				] &, 
				{paramDesc, parameters}
			],
			Message[InvokeServiceOperation::"mapparam", 
			        paramDesc, 
			        parameters];
			Return[$Failed];
		]
	]

(* InstallServiceOperation *)

InstallServiceOperation[ServicesImpl["Java"],
                        exprName_Symbol, 
                        endPoint_String,
                        args_List,
                        header_List,
                        options___?OptionQ] :=
    JavaBlock[
			Module[{argNames, argTypes, argSymbols, argPats, 
			        headerNames, headerTypes, headerSymbols},
				InstallJava[];

        {allowShortContext} = {AllowShortContext} /. 
			 	  Flatten[{options}]  /. Options[InstallServiceOperation];

				If[TrueQ[allowShortContext],
				  context = Context[exprName];
				  If[!MemberQ[$ContextPath, context], AppendTo[$ContextPath, context]];
        ];
			 	  
				argNames = (Context[exprName] <> "Private`" <> NormalizeSymbolName[GetLocalName[#]] & /@ 	args);
		   	argTypes = GetType[#, MapNamespaces[#, {}]] & /@ args;
				argSymbols = Unique[ToExpression[#]] & /@ argNames;
				argPats = Pattern[#, _] & /@ argSymbols;

        headerNames = (Context[exprName] <> "Private`" <> NormalizeSymbolName[GetLocalName[#]] & /@	header);   
			  headerTypes = GetType[#, MapNamespaces[#, {}]] & /@ header;
			  headerSymbols = Unique[ToExpression[#]] & /@ headerNames;
				headerPats = Pattern[#, _] & /@ headerSymbols;  
				                   
				With[{exprName = exprName, argSymbols = argSymbols, 
				      argDesc = args, argPats = Sequence @@ argPats, 
				      headerSymbols = headerSymbols, headerDesc = header, 
				      headerPats = Sequence @@ headerPats,
				      o = options, optSymbol = ToExpression[Context[exprName] <> "Private`" <> "opts"], 
				      optPat = Pattern @@ {ToExpression[Context[exprName] <> "Private`" <> "opts"], ___?OptionQ}},

					exprName[envelope:XMLObject["Document"][{___}, 
					           XMLElement[
                       ("Envelope" | 
                        {_String, "Envelope"}), 
                       {___}, 
                       {__XMLElement}], 
                     {___}], 
                   optPat] := 
						Module[{opts},
  					  opts = Union[Flatten[{optSymbol}], Flatten[{o}]];
										                  								  
              InvokeServiceOperation[ServicesImpl["Java"], endPoint, envelope, opts]   						    
						];

					exprName[argPats, headerPats, optPat] := 
						Module[{envelope, result, opts, params, headerParams},
										                  								  
  					  opts = Union[Flatten[{optSymbol}], Flatten[{o}]];
						  params = mapParameters[argDesc, argSymbols];
						  headerParams = mapParameters[headerDesc, headerSymbols];

						  envelope = ToServiceRequest[ServicesImpl["Java"], params, headerParams, opts];
						  If[envelope === $Failed, Return[$Failed]];	

              result = InvokeServiceOperation[ServicesImpl["Java"], endPoint, envelope, opts];
              If[result === $Failed, Return[$Failed]];	
                 
              FromServiceResponse[ServicesImpl["Java"], result, opts]
						    
						];
                    
          ToServiceRequest[exprName, argPats, headerPats, optPat] := 
            Module[{opts, params, headerParams},
						  opts = Union[Flatten[{optSymbol}], Flatten[{o}]];
						  params = mapParameters[argDesc, argSymbols];
						  headerParams = mapParameters[headerDesc, headerSymbols];

						  ToServiceRequest[ServicesImpl["Java"], params, headerParams, opts]
						];
            
					InvokeServiceOperation[
					        exprName,
					        envelope:XMLObject["Document"][{___}, 
					           XMLElement[
                       ("Envelope" | 
                        {_String, "Envelope"}), 
                       {___}, 
                       {__XMLElement}], 
                     {___}], 
                   optPat] := 
						Module[{opts},
  					  opts = Union[Flatten[{optSymbol}], Flatten[{o}]];
										                  								  
              InvokeServiceOperation[ServicesImpl["Java"], endPoint, envelope, opts]   						    
						];

          InvokeServiceOperation[exprName, argPats, headerPats, optPat] := 
            Module[{opts, params, headerParams, envelope}, 
              opts = Union[Flatten[{optSymbol}], Flatten[{o}]];
						  params = mapParameters[argDesc, argSymbols];
						  headerParams = mapParameters[headerDesc, headerSymbols];

						  envelope = ToServiceRequest[ServicesImpl["Java"], params, headerParams, opts];
						  If[envelope === $Failed, Return[$Failed]];	

              InvokeServiceOperation[ServicesImpl["Java"], endPoint, envelope, opts]
            ];
				];
				
				$InstalledServices = Union[$InstalledServices, {exprName}];
				
				exprName
			]
		]
		
InstallService[ServicesImpl["Java"],
               wsdlURL_String, 
               options___?OptionQ] := 
  Module[{result},
   	InstallJava[];
   				 	 
   	If[TrueQ[$PrintPerformanceNumbers], Print["Installing Service: ", Date[]]];

  	result = MapWSDL[ServicesImpl["Java"], wsdlURL, options];
          
   	If[TrueQ[$PrintPerformanceNumbers], Print["Finished Installing Service: ", Date[]]];
	  
	  result
	]

InstallService[ServicesImpl["Java"],
               wsdlURL_String,
               context_String,
               options___?OptionQ] :=
  Module[{result},
   	InstallJava[];
   	   	
   	If[TrueQ[$PrintPerformanceNumbers], Print["Installing Service: ", Date[]]];
          
 	  result = MapWSDL[ServicesImpl["Java"], wsdlURL, context, options];	    
   	
   	If[TrueQ[$PrintPerformanceNumbers], Print["Finished Installing Service: ", Date[]]];
	  	  
	  result
	]

UninstallServiceOperation[ServicesImpl["Java"], operation_Symbol] :=
  (
    GetOperationPalette[operation] =.;
        
    $InstalledServices = Select[$InstalledServices, # =!= operation &];
    
    Remove[operation];
    
    (* 
    Clear[ToServiceRequest[operation, ___]];
    Clear[InvokeServiceOperation[operation, ___]];
    *)
  )
  
UninstallService[ServicesImpl["Java"], context_String] :=
  (     
    UninstallServiceOperation /@ Intersection[$InstalledServices, ToExpression /@ Names[context]];
  )

OperationPalette[ServicesImpl["Java"], operation_Symbol] := 
	Module[{notebook},			  
	  notebook = GetOperationPalette[operation];
	  If[notebook === Null,
			Return["Palette cannot be found."];
     ];
    nb = NotebookPut[notebook];
  ]

  (* reportError is used to present Java errors in a prettier and less intimidating manner. *)    
  reportShortError[e_JLink`Package`manualException, 
              isPost : (True | False) : False] := 
    reportShortError[First[e], isPost]

  reportShortError[e_JLink`Package`autoException, 
              isPost : (True | False) : False]:= 
    reportShortError[First[e], isPost]    
    
  reportShortError[e_String, isPost : (True | False)] := 
   Block[{$MessagePrePrint, str = e, pos1, pos2},
     (*Sometimes get newline or 0x00 chars at end of message.Remove 
       them, and also the terminating period if there is one.*)
     str = FromCharacterCode[ToCharacterCode[e] /. {a__, junk___?(#<=13&)} :> {a}];
	   If[StringTake[str, -1] === ".", str = StringDrop[str, -1]];

	   If[StringMatchQ[str, "AxisFault*"],
	     pos1 = StringPosition[str, "faultString: "];
	     pos2 = StringPosition[str, "faultActor: "];
	     str = StringTake[str, {Last[First[pos1]], First[First[pos2]] - 3}],
 	     If[StringMatchQ[str, "*\tat "],
  	     pos1 = StringPosition[str, "\tat "];
  	     str = StringTake[str, {1, First[First[pos1]] - 3}]
       ];
	   ];
	   Message[InvokeServiceOperation::"native", str]
   ]
   
   reportError[e_JLink`Package`manualException, 
                   isPost:(True | False):False] := 
     reportError[First[e], isPost]
     
	 reportError[e_JLink`Package`autoException, 
	                 isPost:(True | False):False] := 	                 
	   reportError[First[e], isPost]
	   
	 reportError[e_String, isPost:(True | False)] :=
	 	Block[{$MessagePrePrint, str = e},
	 		(* Sometimes get newline or 0x00 chars at end of message. Remove them, and also the terminating
	 		   period if there is one.
	 		*)
	 		str = FromCharacterCode[ToCharacterCode[e] /. {a__, junk___?(#<=13&)} :> {a}];
	 		If[StringTake[str, -1] === ".", str = StringDrop[str, -1]];
	 		If[isPost, 
	 			Message[Java::pexcptn, str],
	 		(* else *)
	 			Message[Java::excptn, str]
	 		]
	 	]
  
End[]

EndPackage[]