(* :Author: Chris Williamson *)

BeginPackage["WebServices`Normalize`", {"WebServices`",	
									            					"WebServices`WSDL`", 
									            					"WebServices`Utilities`",
									            					"WebServices`SchemaUtilities`"}]

Begin["`Private`"]

setName[_, name:{_String, _String}] := name;

setName[namespace_String, name_String] := {namespace, name};

setName[_, name_String] := {"", name};

setType[m:({(name_String | name:{ns_String, _String}), 
              attrs:{___,{_String?XMLSchemaInstanceStringQ, "type"}->_,___}}->val_), 
        t:{_String, _String}] := Null;

setType[m:XMLElement[(name_String | name:{_String, _String}), 			
												attrs:{___,{_String?XMLSchemaInstanceStringQ, "type"}->_,___}, 
												val:{___}],
        t:{_String, _String}] := Null;

setType[_, t:{_String, _String}] := {"http://www.w3.org/2001/XMLSchema-instance","type"}->t;

setType[___] := Null;

setArrayType[m:XMLElement[(name_String | name:{_String, _String}), 			
							attrs:{___,{"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->_,___}, 
							val:{___}], 
             t_, 
             namespaces_List] := Null;

setArrayType[m:({(name_String | name:{ns_String, _String}), 
              attrs:{___,{"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->_,___}}->val_), 
             t_, 
             namespaces_List] := Null;

setArrayType[_, 
             t:{_String, _String},
             namespaces_List] := 
  Module[{schema, schemaNamespaces, typeSchema, arrayType},

    schema = GetNamespaceDefinitions[First[t]];
  	schemaNamespaces = MapNamespaces[schema, {}];
    If[!SchemaQ[schema], 
   	  schema = GetSchema[schema, First[t]];
   	  schemaNamespaces = MapNamespaces[schema, schemaNamespaces];
   	  If[!SchemaQ[schema],
     	  Return[Null];
     	];
    ];   	    
    typeSchema = GetSchemaType[Last[t], schema];
    schemaNamespaces = MapNamespaces[schema, schemaNamespaces];    
    
    If[(arrayType = GetArrayType[typeSchema]) =!= Null,
      {"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->GetType[arrayType, schemaNamespaces],
      Null
    ]    
  ];

setArrayType[_, 
        type:XMLElement[{_?XMLSchemaStringQ, ("complexType" | "simpleType")}, {___}, {___}],
        namespaces_List] :=
  Module[{arrayType},
    If[(arrayType = GetArrayType[type]) =!= Null,
      {"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->GetType[arrayType, namespaces],
      Null
    ]        
  ];

setArrayType[___] := Null;

(* Normalize base type with string type *)
Normalize[ServicesImpl["Java"],
          (val:(True | False) | val_Integer | val_Real | val_List), 
          type_String,
          namespace_,
          namespaces_List] :=

	Module[{t, newAttrs = {}, newName = {"", "item"}, typeAttr}, 

    t = GetType[type, namespaces];
	
    (* set type *)     
    If[(typeAttr = setType[val, t]) =!= Null, AppendTo[newAttrs, typeAttr]];      

 		{newName, newAttrs}->val
	];
  
(* Normalize base type with list type *)
Normalize[ServicesImpl["Java"],
          (val:(True | False) | val_Integer | val_Real | val_List), 
          type:({_String, _String} | {}),
          namespace_,
          namespaces_List] :=
  Module[{newAttrs = {}, newName = {"", "item"}, typeAttr},
      
    (* set type *)     
    If[(typeAttr = setType[m, type]) =!= Null, AppendTo[newAttrs, typeAttr]];      

 		{newName, newAttrs}->val
  ];

(* Normalize base type with schema type *)
Normalize[ServicesImpl["Java"],
          (val:(True | False) | val_Integer | val_Real | val_List), 
          type:XMLElement[{_?XMLSchemaStringQ, ("complexType" | "simpleType")}, 
            {___,{"", "name"}->typeName,___}, {___}],
          namespace_,
          namespaces_List] :=
  Module[{newAttrs = {}, newName = {"", "item"}, typeAttr},
      
    (* set type *)     
    If[!StringQ[namespace], ns = ""];    
    If[(typeAttr = setType[m, {namespace, typeName}]) =!= Null, AppendTo[newAttrs, typeAttr]];      
    
 		{newName, newAttrs}->val
  ];
 
(* Normalize base type with null type *)
Normalize[ServicesImpl["Java"],
          (val:(True | False) | val_Integer | val_Real | val_List), 
          Null,
          namespace_,
          namespaces_List] := {{"", "item"}, {}}->val;

(* Normalize base type with string type *)
Normalize[ServicesImpl["Java"],
          val_String,
          type_String,
          namespace_,
          namespaces_List] :=

	Module[{t, newAttrs = {}, newName = {"", "item"}, typeAttr}, 

    t = GetType[type, namespaces];
	
    (* set type *)     
    If[(typeAttr = setType[val, t]) =!= Null, AppendTo[newAttrs, typeAttr]];      

 		XMLElement[newName, newAttrs, {val}]
	];
  
(* Normalize base type with list type *)
Normalize[ServicesImpl["Java"],
          val_String,
          type:({_String, _String} | {}),
          namespace_,
          namespaces_List] :=
  Module[{newAttrs = {}, newName = {"", "item"}, typeAttr},
      
    (* set type *)     
    If[(typeAttr = setType[m, type]) =!= Null, AppendTo[newAttrs, typeAttr]];      

 		XMLElement[newName, newAttrs, {val}]
  ];

(* Normalize base type with schema type *)
Normalize[ServicesImpl["Java"],
          val_String,
          type:XMLElement[{_?XMLSchemaStringQ, ("complexType" | "simpleType")}, 
            {___,{"", "name"}->typeName,___}, {___}],
          namespace_,
          namespaces_List] :=
  Module[{newAttrs = {}, newName = {"", "item"}, typeAttr},
      
    (* set type *)     
    If[!StringQ[namespace], ns = ""];    
    If[(typeAttr = setType[m, {namespace, typeName}]) =!= Null, AppendTo[newAttrs, typeAttr]];      
    
 		XMLElement[newName, newAttrs, {val}]
  ];
 
(* Normalize base type with null type *)
Normalize[ServicesImpl["Java"],
          val_String,
          Null,
          namespace_,
          namespaces_List] := XMLElement[{"", "item"}, {}, {val}];  
  
(* Normalize Rule with string type *)
Normalize[ServicesImpl["Java"],
          m:((name_String | 
             {(name_String | name:{ns_String, _String}), 
              attrs:{___}})->val_), 
          type_String,
          namespace_,
          namespaces_List] :=
  Module[{t, newName = name, newAttrs = {}, typeAttr, arrayTypeAttr},
          
    If[Head[Unevaluated[attrs]] === List, newAttrs = attrs];

    t = GetType[type, namespaces];
    
    (* set name *)
    newName = setName[namespace, name];

    (* set type *)   
    If[(typeAttr = setType[m, t]) =!= Null, AppendTo[newAttrs, typeAttr]];      

    (* set array type *)    
    If[(arrayTypeAttr = setArrayType[m, t, namespaces]) =!= Null, AppendTo[newAttrs, arrayTypeAttr]];
    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];
    
 		{newName, newAttrs}->val
  ];

(* Normalize Rule with list type *)
Normalize[ServicesImpl["Java"],
          m:((name_String | 
             {(name_String | name:{ns_String, _String}), 
              attrs:{___}})->val_), 
          type:({_String, _String} | {}),
          namespace_,
          namespaces_List] :=
  Module[{newName = name, newAttrs = {}, typeAttr, arrayTypeAttr},
      
    If[Head[Unevaluated[attrs]] === List, newAttrs = attrs];

    (* set name *)
    newName = setName[namespace, name];

    (* set type *)     
    If[(typeAttr = setType[m, type]) =!= Null, AppendTo[newAttrs, typeAttr]];      

    (* set array type *)    
    If[(arrayTypeAttr = setArrayType[m, type, namespaces]) =!= Null, AppendTo[newAttrs, arrayTypeAttr]];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		{newName, newAttrs}->val
  ];

(* Normalize Rule with schema type *)
Normalize[ServicesImpl["Java"],
          m:((name_String | 
             {(name_String | name:{_String, _String}), 
              attrs:{___}})->val_), 
          type:XMLElement[{_?XMLSchemaStringQ, ("complexType" | "simpleType")}, 
            {___,{"", "name"}->typeName,___}, {___}],
          namespace_,
          namespaces_List] :=
  Module[{newName = name, newAttrs = {}, typeAttr, arrayTypeAttr},
      
    If[Head[Unevaluated[attrs]] === List, newAttrs = attrs];

    (* set name *)
    newName = setName[namespace, name];

    (* set type *)     
    If[!StringQ[namespace], ns = ""];    
    If[(typeAttr = setType[m, {namespace, typeName}]) =!= Null, AppendTo[newAttrs, typeAttr]];      

    (* set array type *)    
    If[(arrayTypeAttr = setArrayType[m, t, namespaces]) =!= Null, AppendTo[newAttrs, arrayTypeAttr]];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		{newName, newAttrs}->val
  ];
 
(* Normalize Rule with null type *)
Normalize[ServicesImpl["Java"],
          m:((name_String | 
             {(name_String | name:{ns_String, _String}), 
              attrs:{___}})->val_), 
          Null,
          namespace_,
          namespaces_List] :=
  Module[{newName = name, newAttrs = {}},
      
    If[Head[Unevaluated[attrs]] === List, newAttrs = attrs];

    (* set name *)
    newName = setName[namespace, name];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		{newName, newAttrs}->val
  ];
          

(* Normalize Referenced XMLElement *)
Normalize[ServicesImpl["Java"],
          m:XMLElement[name:(_String | {_String, _String}), 
										   attrs:{___, "href"->id_String,___}, 
										   children:{___}],
          type_,
          namespace_,
          namespaces_List] :=
  Module[{elmnt, t = Null, nspaces},

  	elmnt = GetHref[StringDrop[id, 1]];
  	If[elmnt === Null, 
  		Message[InvokeServiceOperation::"href"];
  		m,
  	  Normalize[ServicesImpl["Java"], elmnt, type, namespace, namespaces]
  	]
  ];
     
(* Normalize XMLElement with String type *)
Normalize[ServicesImpl["Java"],
          m:XMLElement[(name_String | name:{_String, _String}), 			
												attrs:{___}, 
												val:{___}], 
          type_String,
          namespace_,
          namespaces_List] :=
  Module[{t, newName = name, newAttrs = attrs, typeAttr, arrayTypeAttr},
    
    t = GetType[type, namespaces];
    
    (* set name *)
    newName = setName[namespace, name];

    (* set type *)     
    If[(typeAttr = setType[m, t]) =!= Null, AppendTo[newAttrs, typeAttr]];      

    (* set array type *)    
    If[(arrayTypeAttr = setArrayType[m, t, namespaces]) =!= Null, AppendTo[newAttrs, arrayTypeAttr]];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		XMLElement[newName, newAttrs, val]
  ]

(* Normalize XMLElement with list type *)
Normalize[ServicesImpl["Java"],
          m:XMLElement[(name_String | name:{_String, _String}), 			
												attrs:{___}, 
												val:{___}], 
          type:({_String, _String} | {}),
          namespace_,
          namespaces_List] :=
  Module[{newName = name, newAttrs = attrs, typeAttr, arrayTypeAttr},
      
    (* set name *)
    newName = setName[namespace, name];

    (* set type *)     
    If[(typeAttr = setType[m, type]) =!= Null, AppendTo[newAttrs, typeAttr]];      

    (* set array type *)    
    If[(arrayTypeAttr = setArrayType[m, type, namespaces]) =!= Null, AppendTo[newAttrs, arrayTypeAttr]];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		XMLElement[newName, newAttrs, val]
  ];

(* Normalize XMLElement with schema type *)
Normalize[ServicesImpl["Java"],
          m:XMLElement[(name_String | name:{_String, _String}), 			
												attrs:{___}, 
												val:{___}], 
          type:XMLElement[{_?XMLSchemaStringQ, ("complexType" | "simpleType")}, 
            {___,{"", "name"}->typeName_String,___}, {___}],
          namespace_,
          namespaces_List] :=
  Module[{newName = name, newAttrs = attrs, typeAttr, arrayTypeAttr},
      
    (* set name *)
    newName = setName[namespace, name];

    (* set type *)     
    If[!StringQ[namespace], ns = ""];    
    If[(typeAttr = setType[m, {namespace, typeName}]) =!= Null, AppendTo[newAttrs, typeAttr]];      

    (* set array type *)    
    If[(arrayTypeAttr = setArrayType[m, type, namespaces]) =!= Null, AppendTo[newAttrs, arrayTypeAttr]];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		XMLElement[newName, newAttrs, val]
  ];

(* Normalize XMLElement with null type *)
Normalize[ServicesImpl["Java"],
          m:XMLElement[(name_String | name:{_String, _String}), 			
												attrs:{___}, 
												val:{___}, ___], 
          Null,
          namespace_,
          namespaces_List] :=
  Module[{newName = name, newAttrs = attrs},
     
    (* set name *)
    newName = setName[namespace, name];

    (* set null *)     
    If[val === Null, 
      AppendTo[newAttrs, {"http://www.w3.org/2001/XMLSchema-instance","null"}->"1"];
    ];

 		XMLElement[newName, newAttrs, val]
  ];
      
End[];

EndPackage[];