(* :Author: Chris Williamson *)

BeginPackage["WebServices`SchemaUtilities`", {"WebServices`Utilities`"}];

SchemaQ;

GetSchema;

GetSchemaType;

GetElementType;

GetElementName;

GetElement;

GetElements;

GetTypeName;

GetTypeAttributes;

GetMaxOccurs;

GetMinOccurs;

GetArrayType;

GetRestrictionBase;

GetEnumerations;

Begin["`Private`"];

SchemaQ[schema:XMLElement[{_?XMLSchemaStringQ, "schema"}, {___}, {___}]] := True;

SchemaQ[___] := False;
	
GetSchema[def:XMLElement[{"http://schemas.xmlsoap.org/wsdl/", "definitions"}, {___}, 
            {___, types:XMLElement[{"http://schemas.xmlsoap.org/wsdl/", "types"}, {___}, 
              children:{___}], ___}],
          schemaName_String] := 
  Module[{results},	
    results = Cases[children, XMLElement[{_?XMLSchemaStringQ, "schema"}, 
                       {___, {"", "targetNamespace"}->schemaName, ___}, {___}]];
    If[Length[results] > 0, First[results], Null]
  ];

GetSchema[___] := Null;
  
GetSchemaTargetNamespace[schema:XMLElement[{_?XMLSchemaStringQ,"schema"},
          {___, {"","targetNamespace"}->targetNamespace_String,___},children:{___}]] := targetNamespace;
          
getSchemaTargetNamespace[___] := "";

GetSchemaType[typeName_String, schema:XMLElement[{_?XMLSchemaStringQ,"schema"},
          {___},children:{___}]] := 
  Module[{results},	
    results = Cases[children, XMLElement[{_?XMLSchemaStringQ, 
                     ("complexType" | "simpleType")}, 
                       {___, {"", "name"}->typeName, ___}, {___}]];
    If[Length[results] > 0, First[results], Null]
  ];

GetSchemaType[___] := Null;

GetElementName[part:XMLElement[{_?XMLSchemaStringQ, "element"}, 
             {___, {"", "name"}->name_String, ___}, {___}]] := name;

GetElementName[___] := Null;


GetElementType[schema:XMLElement[{_?XMLSchemaStringQ,"element"},
          {___,{"", "type"}->type_String,___},children:{___}]] := type;


GetElementType[schema:XMLElement[{_?XMLSchemaStringQ,"element"},
          {___},children:{___, type:XMLElement[{_?XMLSchemaStringQ, 
            ("complexType" | "simpleType")}, {___}, {___}],___}]] := type;

GetElementType[___] := Null;

GetElement[elementName_String, schema:XMLElement[{_?XMLSchemaStringQ,"schema"},
            {___},children:{___}]] := 
  Module[{results},	
    results = Cases[children, XMLElement[{_?XMLSchemaStringQ, "element"}, 
                       {___, {"", "name"}->elementName, ___}, {___}], Infinity];
    If[Length[results] > 0, First[results], Null]
  ];


GetElement[elementName_String, schema:XMLElement[{_?XMLSchemaStringQ, "complexType"},
            {___},children:{___}]] := 
  Module[{results},	
    results = Cases[children, XMLElement[{_?XMLSchemaStringQ, "element"}, 
                       {___, {"", "name"}->elementName, ___}, {___}], Infinity];
    If[Length[results] > 0, First[results], Null]
  ];

GetElement[___] := Null;

GetElements[type:XMLElement[{_?XMLSchemaStringQ, "complexType"}, {___}, children:{___}]] :=
  Cases[children, XMLElement[{_?XMLSchemaStringQ, "element"}, {___}, {___}], 3];

GetElements[___] := {};

GetTypeName[type:XMLElement[{_?XMLSchemaStringQ, ("simpleType" | "complexType")}, 
                {___,{"", "name"}->name_String,___}, children:{___}]] := name;

GetTypeName[___] := Null;

GetTypeAttributes[type:XMLElement[{_?XMLSchemaStringQ, ("simpleType" | "complexType")}, 
                {___}, children:{___}]] :=
  Cases[children, XMLElement[{_?XMLSchemaStringQ, "attribute"}, {___}, {___}], Infinity];

GetTypeAttributes[___] := {};

GetMaxOccurs[XMLElement[{_?XMLSchemaStringQ, "element"}, 
                       {___, {"", "maxOccurs"}->maxOccurs_String, ___}, {___}]] := maxOccurs;
                       
GetMaxOccurs[___] := Null;

GetMinOccurs[XMLElement[{_?XMLSchemaStringQ, "element"}, 
                       {___, {"", "minOccurs"}->minOccurs_String, ___}, {___}]] := minOccurs;

GetMinOccurs[___] := Null;

GetArrayType[type:XMLElement[{_?XMLSchemaStringQ, "complexType"}, {___}, children:{___}]] :=
  Module[{results},	
    
    results = Cases[children, XMLElement[{_?XMLSchemaStringQ, "attribute"}, 
                        {___, {"http://schemas.xmlsoap.org/wsdl/","arrayType"}->arrayType_String,
                           ___}, {___}] :> arrayType, Infinity];
    If[Length[results] > 0, Return[First[results]]];
    
    elements = GetElements[type];
    If[Length[elements] =!= 1, Return[Null]];

    maxOccurs = GetMaxOccurs[First[elements]];
    If[maxOccurs === Null, Return[Null]];
    If[maxOccurs === "unbounded", Return[GetElementType[First[elements]]]];
    If[DigitQ[maxOccurs] && ToExpression[maxOccurs] > 1, Return[GetElementType[First[elements]]]];
    Null
  ];      

GetArrayType[___] := Null;

GetRestrictionBase[type:XMLElement[{_?XMLSchemaStringQ, ("simpleType" | "complexType")}, {___}, children:{___}]] :=
  Module[{results},	
    results = Cases[children, XMLElement[{_?XMLSchemaStringQ, "restriction"}, 
                       {___, {"", "base"}->base_String, ___}, {___}] :> base, 2];
    If[Length[results] > 0, First[results], Null]
  ];  
  
GetRestrictionBase[___] := Null;

GetEnumerations[type:XMLElement[{_?XMLSchemaStringQ, ("simpleType" | "complexType")}, {___}, children:{___}]] :=
  Cases[children, XMLElement[{_?XMLSchemaStringQ, "enumeration"}, 
                       {___, {"", "value"}->value_String, ___}, {___}] :> value, 3]

GetEnumerations[___] := {};

End[];

EndPackage[];


