(* :Author: Chris Williamson *)

BeginPackage["WebServices`Serialize`", {"JLink`",
									            					"WebServices`",	
                                        "WebServices`WSDL`",
                                        "WebServices`Utilities`",
									            					"WebServices`SchemaUtilities`",
									            				  "WebServices`Normalize`"}]

Begin["`Private`"]

(* Serialize and Deserialize *) 
base64 = Null;

base64Encode[str_String] := (
  If[base64 === Null, base64 = JavaNew["org.apache.axis.encoding.Base64"]];
  JavaObjectToExpression[base64@encode[MakeJavaObject[str]@getBytes[]]]
);


(* Serialize Boolean *)
Serialize[ServicesImpl["Java"], val:(True | False), _List, options___?OptionQ] :=
		XMLElement["item", {{"http://www.w3.org/2001/XMLSchema-instance", 
		                     "type"}->"xsd:boolean"}, {ToString[val]}];

(* Serialize String *)
Serialize[ServicesImpl["Java"], val_String, _List, options___?OptionQ] :=
		XMLElement["item", {{"http://www.w3.org/2001/XMLSchema-instance", 
		                     "type"}->"xsd:string"}, {val}];

(* Serialize Integer *)
Serialize[ServicesImpl["Java"], val_Integer, _List, options___?OptionQ] :=
		XMLElement["item", {{"http://www.w3.org/2001/XMLSchema-instance", 
		                     "type"}->"xsd:int"}, {ToString[val]}];

(* Serialize Real *)
Serialize[ServicesImpl["Java"], val_Real, _List, options___?OptionQ] :=
		XMLElement["item", {{"http://www.w3.org/2001/XMLSchema-instance", 
		                     "type"}->"xsd:double"}, {ToString[val]}];


(* Used for mapping attributes to an element. If the value of the attribute is not a 
   String, Integer, or Real, the attribute will not be set.  Also the type
   attribute will not be set, as it is set earlier in Serialize.
 *)

namespaceQ = {};

getNamespacePrefix[ns_String, namespaces_List] := 
  Module[{nsMappings},
    If[ns === "", Return[""]];
    nsMappings = Union[namespaces, namespaceQ];
    prefixes = Select[nsMappings, MatchQ[#, {_?XMLNamespaceStringQ, val_String}->ns] &];
    If[Length[prefixes] > 0, 
      prefix = prefixes[[1, 1, 2]],
      prefix = "ns"<> ToString[NamespaceId];
      NamespaceId++;
      AppendTo[namespaceQ, {"http://www.w3.org/2001/xmlns/", prefix}->ns];
    ];
    prefix
  ]

(* QName Attributes in a List - only in Rule syntax *)

mapAttribute[{ns_String, name_String}->{attrNS_String, attrName_String}, 
             attributes:{___}, 
             namespaces_List] :=
 	Append[attributes, {getNamespacePrefix[ns, namespaces], name}->getNamespacePrefix[attrNS, namespaces]<>":"<>attrName]

mapAttribute[name_String->{attrNS_String, attrName_String}, 
             attributes:{___},
             namespaces_List] :=
	Append[attributes, name->getNamespacePrefix[attrNS, namespaces]<>":"<>attrName]

(* QName Attributes in a String *)

mapAttribute[{ns_String, name_String}->attrVal_String?WrappedStringQ,
             attributes:{___},
             namespaces_List] :=
  Module[{val = attrVal, position, namespace},
		position = Part[StringPosition[val, "}"], 1, 2];
		namespace = StringTake[val, {2, position - 1}];
		val = StringTake[val, {position+1, StringLength[val]}];
   	Append[attributes, {getNamespacePrefix[ns, namespaces], name}->getNamespacePrefix[namespace,namespaces]<>":"<>val]
	]

mapAttribute[name_String->attrVal_String?WrappedStringQ,
             attributes:{___},
             namespaces_List] :=
  Module[{val = attrVal, position, namespace},
		position = Part[StringPosition[val, "}"], 1, 2];
		namespace = StringTake[val, {2, position - 1}];
		val = StringTake[val, {position+1, StringLength[val]}];
   	Append[attributes, {ns, name}->getNamespacePrefix[namespace, namespaces]<>":"<>val]
	]

(* Regualar Attributes *)

mapAttribute[{ns_String, name_String}->val_, 
             attributes:{___},
             namespaces_List] :=
	Module[{attrVal},
		Switch[ 
			val, 
			True, attrVal = "true",
			False, attrVal = "false",
			_String, attrVal = val,
			_Integer, attrVal = ToString[val],
			_Real, attrVal = ToString[NumberForm[val, 16, NumberFormat -> (If[#3 == "", #1,SequenceForm[#1, "e", #3]] &)]],
			_, attrVal = Null;
		];	
		If[attrVal =!= Null,
			Append[attributes, {getNamespacePrefix[ns, namespaces], name}->attrVal],
			attributes
	  ]
	]

mapAttribute[name_String->val_, 
             element:XMLElement[name_, attributes:{___}, value:{___}], 
             namespaces_List] :=
	Module[{attrVal},
		Switch[ 
			val, 
			True, attrVal = "true",
			False, attrVal = "false",
			_String, attrVal = val,
			_Integer, attrVal = ToString[val],
			_Real, attrVal = ToString[NumberForm[val, 16, NumberFormat -> (If[#3 == "", #1,SequenceForm[#1, "e", #3]] &)]],
			_, attrVal = Null;
		];
		If[attrVal =!= Null,
			Append[attributes, name->attrVal],
			attributes
	  ]
	]

setValue[parameter:((name_String | 
                    {(name_String | name:{_String, _String}), 
                     attrs:{___, {"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->arrayType_,___}})->val_List), 
         type_, 
         nspaces_List, 
         options___?OptionQ] := 
  Module[{t = Null, n, namespace = "", encodingStyle, encodingStyleURI}, 
  
    {encodingStyle, encodingStyleURI} = {EncodingStyle, EncodingStyleURI} /. 
      Flatten[{options}]  /. Options[InvokeServiceOperation];
 
    If[MatchQ[type, {_String, _String}], namespace  = First[type]];

		If[StringQ[arrayType] && !StringMatchQ[arrayType, "*[*"], 
      t = StringTake[arrayType, {1, First[First[StringPosition[arrayType, "["]]]-1}];
 		];
 		 		
    If[MatchQ[arrayType, {_String, _String}],  		
  		If[StringMatchQ[Last[arrayType], "*[*"], 
        t = {First[arrayType], StringTake[Last[arrayType], {1, First[First[StringPosition[Last[arrayType], "["]]]-1}]};
   		];
    ];

    children = 
 			((If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
 			    n = Normalize[ServicesImpl["Java"], #, t, namespace, nspaces],
 			    n = Normalize[ServicesImpl["Java"], #, Null, namespace, nspaces];;
 			  ];
 			  Serialize[ServicesImpl["Java"], n, nspaces, options]) & /@ val)
  ]
  
setValue[parameter:((name_String | 
                    {(name_String | {namespace_String, name_String}), 
                     attrs:{___}})->val_),
         type_, 
         nspaces_List, 
         options___?OptionQ] := 
  Module[{encodingStyle, encodingStyleURI, typeNS = namespace, schema, typeSchema, 
          n, children, failed, schemaNamespaces = {}},

    {encodingStyle, encodingStyleURI} = {EncodingStyle, EncodingStyleURI} /. 
      Flatten[{options}]  /. Options[InvokeServiceOperation];
       
    (* Get schema for type, so it can be used to normalize children *)
    If[MatchQ[type, {_String, _String}],
      typeNS = First[type];
   	  schema = GetNamespaceDefinitions[typeNS];
   	  schemaNamespaces = MapNamespaces[schema, {}];
  	  If[!SchemaQ[schema], 
   	    schema = GetSchema[schema, typeNS];
   	    schemaNamespaces = MapNamespaces[schema, schemaNamespaces];
      ];
      typeSchema = GetSchemaType[Last[type], schema];
      schemaNamespaces = MapNamespaces[typeSchema, schemaNamespaces];
    ];
    
		Switch[
			val,
			_List, 
      	children = 
			    ((If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
			        n = Normalize[ServicesImpl["Java"], #, GetElementType[GetElement[GetLocalName[#], typeSchema]], typeNS, schemaNamespaces],
			        n = Normalize[ServicesImpl["Java"], #, Null, typeNS, schemaNamespaces];
			      ];
			 	    Serialize[ServicesImpl["Java"], n, nspaces, options]) & /@ val);
		    failed = Cases[children, $Failed, Infinity];
			  If[Length[failed] > 0, Return[$Failed]],
			_Rule,
			  If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
			    n = Normalize[ServicesImpl["Java"], val, GetElementType[GetElement[GetLocalName[#], typeSchema]], typeNS, schemaNamespaces],
			    n = Normalize[ServicesImpl["Java"], val, Null, typeNS, schemaNamespaces];
			  ];
				children = Serialize[ServicesImpl["Java"], n, nspaces, options];
				If[children =!= $Failed, children = {children}],
			_XMLElement, 
			  If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
			    n = Normalize[ServicesImpl["Java"], val, GetElementType[GetElement[GetLocalName[#], typeSchema]], typeNS, schemaNamespaces];
			    n = Normalize[ServicesImpl["Java"], val, Null, typeNS, schemaNamespaces];
			  ];
				children = Serialize[ServicesImpl["Java"], n, nspaces, options];
				If[children =!= $Failed, children = {children}],
			_String, 
			  If[MatchQ[type, {_String?XMLSchemaStringQ, ("base64"|"base64Binary")}], 
			    children = {base64Encode[val]},
			    children = {val};
			  ],
			_Integer, children = {ToString[val]},
			_Real, children = {ToString[NumberForm[val, 16, NumberFormat -> (If[#3 == "", #1, SequenceForm[#1, "e", #3]] &)]]},
			True, children = {"true"},
			False, children = {"false"},
			Null, children = {},
			_, Message[InvokeServiceOperation::"srlz", val];Return[$Failed];
		];
		children
  ]


(* Serialize Rule *)
Serialize[ServicesImpl["Java"], 
          parameter:((name_String | 
                     {(name_String | {namespace_String, name_String}), attrs:{___}})->val_), 
          namespaces_List,
          options___?OptionQ] :=
	Module[{encodingStyle, encodingStyleURI, nspaces, nm, attributes = {}, type, children},
    
    {encodingStyle, encodingStyleURI} = {EncodingStyle, EncodingStyleURI} /. 
      Flatten[{options}]  /. Options[InvokeServiceOperation];
				 
		(* Get options and add namespaces *)
		nspaces = MapNamespaces[parameter, namespaces];
				
	  If[Head[Unevaluated[namespace]] === String,
	    nm = {getNamespacePrefix[namespace, nspaces], name},
	    nm = name
	  ];
	  		
	  (* Add attributes *)
		If[Head[Unevaluated[attrs]] === List, 
		  (attributes = mapAttribute[#, attributes, nspaces]) & /@ attrs;
		];
		
		(* Add namespaces required for attributes *)
    (attributes = mapAttribute[#, attributes, nspaces]) & /@ namespaceQ;
    nspaces = Union[nspaces, namespaceQ];
    namespaceQ = {};
	  
	  type = GetType[parameter, nspaces];
	  children = setValue[parameter, type, nspaces, options];
	  If[children === $Failed, Return[$Failed]];
	  
		XMLElement[nm, attributes, children]
	]

XMLElementQ[x_XMLElement] := True

XMLElementQ[x___] := (Message[InvokeServiceOperation::"srlz", x];False)


setValue[m:XMLElement[(name_String | {ns_String, name_String}), 			
											attrs:{___, {"http://schemas.xmlsoap.org/soap/encoding/","arrayType"}->arrayType_,___}, 
											children:{___}], 
         type_, 
         nspaces_List, 
         options___?OptionQ] := 
  Module[{t = Null, n, c, namespace = "", encodingStyle, encodingStyleURI}
    {encodingStyle, encodingStyleURI} = {EncodingStyle, EncodingStyleURI} /. 
      Flatten[{options}]  /. Options[InvokeServiceOperation];
 
    If[MatchQ[type, {_String, _String}], namespace  = First[type]];

		If[StringQ[arrayType] && !StringMatchQ[arrayType, "*[*"], 
      t = StringTake[arrayType, {1, First[First[StringPosition[arrayType, "["]]]-1}];
 		];
 		 		
    If[MatchQ[arrayType, {_String, _String}],  		
  		If[StringMatchQ[Last[arrayType], "*[*"], 
        t = {First[arrayType], StringTake[Last[arrayType], {1, First[First[StringPosition[Last[arrayType], "["]]]-1}]};
   		];
    ];

		c = (If[XMLElementQ[#], 
  	     (If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
 			      n = Normalize[ServicesImpl["Java"], #, t, namespace, nspaces],
 			      n = Normalize[ServicesImpl["Java"], #, Null, typeNS, nspaces];
 			    ];
  	      s = Serialize[ServicesImpl["Java"], n, nspaces, options]),
  				$Failed
  			 ] & /@ children),
		c = Cases[c, $Failed, Infinity];
		If[Length[c] > 0, $Failed, c]
  ]
  
setValue[m:XMLElement[(name_String | {ns_String, name_String}), 			
											attrs:{___}, 
											children:{___}],
         type_, 
         nspaces_List, 
         options___?OptionQ] := 
  Module[{encodingStyle, encodingStyleURI, typeNS = ns, schema, typeSchema, 
          n, failed, schemaNamespaces = {}, c = children},

    {encodingStyle, encodingStyleURI} = {EncodingStyle, EncodingStyleURI} /. 
      Flatten[{options}]  /. Options[InvokeServiceOperation];

    If[MatchQ[type, {_String, _String}],
      typeNS = First[type];
   	  schema = GetNamespaceDefinitions[typeNS];
   	  schemaNamespaces = MapNamespaces[schema, {}];
  	  If[!SchemaQ[schema], 
   	    schema = GetSchema[schema, typeNS];
   	    schemaNamespaces = MapNamespaces[schema, schemaNamespaces];
      ];
      typeSchema = GetSchemaType[Last[type], schema];
      schemaNamespaces = MapNamespaces[typeSchema, schemaNamespaces];
    ];

		(* Process children and/or set value *)
		Switch[children,
		  {___String}, c = children,
		  {___XMLElement}, 
     		c = 
     		  (If[XMLElementQ[#], 
  	        (If[encodingStyle === "encoded" && encodingStyleURI === {"http://schemas.xmlsoap.org/soap/encoding/"},
			         n = Normalize[ServicesImpl["Java"], #, GetElementType[GetElement[GetLocalName[#], typeSchema]], typeNS, schemaNamespaces],
			         n = Normalize[ServicesImpl["Java"], #, Null, typeNS, schemaNamespaces];
			       ];
  	         s = Serialize[ServicesImpl["Java"], n, nspaces, options]),
  			  	$Failed
    			 ] & /@ children);
				c = Cases[c, $Failed, Infinity];
 				If[Length[c] > 0, Return[$Failed]],
 			_, c = $Failed
		];
		c
  ]

(* Serialize XMLElement *)
Serialize[ServicesImpl["Java"], m:XMLElement[(name_String | {ns_String, name_String}), 			
												attrs:{___}, 
												children:{___}], 
					namespaces_List,
				  options___?OptionQ] :=
	Module[{nspaces, nm, type, attributes = {}, qname, jtype, node, refType = Null, c}, 
		
		(* Get options *)
		nspaces = MapNamespaces[m, namespaces];

    If[Head[Unevaluated[namespace]] === String,
	    nm = {getNamespacePrefix[namespace, nspaces], name},
	    nm = name
	  ];
	  
		(* Add attributes *)
		If[Head[Unevaluated[attrs]] === List, 
		  (attributes = mapAttribute[#, attributes, nspaces]) & /@ attrs;
		];
		
		(* Add namespaces required for attributes *)
    (attributes = mapAttribute[#, attributes, nspaces]) & /@ namespaceQ;
    nspaces = Union[nspaces, namespaceQ];
    namespaceQ = {};

		(* Process children and/or set value *)
    If[symbolTable =!= Null && type =!= {},
   	  qname = getQName @@ type;
		  jtype = getJType[qname, symbolTable];
      node = getNode[jtype];		  
      If[jtype =!= Null, refType = getRefType[jtype]];
      If[refType === Null, refType = getArrayType[node, symbolTable]];
		];	  
	  
	  c = setValue[m, type, nspaces, options];
	  If[c === $Failed, Return[$Failed]];
	  
		XMLElement[nm, attributes, c]
	]
	
End[]

EndPackage[]
