(* :Author: Chris Williamson *)

BeginPackage["WebServices`Utilities`"];

GetType;

GetName;

GetQName;

GetLocalName;

GetAttribute;

GetAttributes;

GetValue;

GetValueAttribute;

GetSymbolicXML;

MapNamespaces;

$Base64LineLength;

$Base64Characters;

NormalizeSymbolName;

XMLNamespaceStringQ;

XMLSchemaStringQ;

XMLSchemaBaseTypeQ

XMLSchemaInstanceStringQ;

WrappedStringQ;

PrefixedStringQ;

Begin["`Private`"]

XMLSchemaStringQ[x_String] :=   
	StringMatchQ[x, "http://www.w3.org/1999/XMLSchema"] || 
	StringMatchQ[x, "http://www.w3.org/2000/XMLSchema"] || 
	StringMatchQ[x, "http://www.w3.org/2001/XMLSchema"	] ||
  StringMatchQ[x, "http://schemas.xmlsoap.org/soap/encoding/"]
  
XMLSchemaStringQ[___] := False

XMLSchemaBaseTypeQ[x_String] :=   
	StringMatchQ[x, "boolean"] || 
	StringMatchQ[x, "int"] || 
	StringMatchQ[x, "integer"] ||
	StringMatchQ[x, "double"] ||
  StringMatchQ[x, "float"] || 
  StringMatchQ[x, "string"] || 
  StringMatchQ[x, "short"] || 
  StringMatchQ[x, "long"] || 
  StringMatchQ[x, "decimal"] || 
  StringMatchQ[x, "byte"] || 
  StringMatchQ[x, "nonPositiveInteger"] || 
  StringMatchQ[x, "negativeInteger"] || 
  StringMatchQ[x, "nonNegativeInteger"] || 
  StringMatchQ[x, "unsignedLong"] || 
  StringMatchQ[x, "unsignedInt"] || 
  StringMatchQ[x, "unsignedShort"] || 
  StringMatchQ[x, "unsignedByte"] || 
  StringMatchQ[x, "positiveInteger"] || 
  StringMatchQ[x, "duration"] || 
  StringMatchQ[x, "dateTime"] || 
  StringMatchQ[x, "time"] || 
  StringMatchQ[x, "date"] || 
  StringMatchQ[x, "gYearMonth"] || 
  StringMatchQ[x, "gYear"] || 
  StringMatchQ[x, "gMonthDay"] || 
  StringMatchQ[x, "gDay"] || 
  StringMatchQ[x, "gMonth"] || 
  StringMatchQ[x, "base64Binary"] || 
  StringMatchQ[x, "hexBinary"] || 
  StringMatchQ[x, "anyURI"] || 
  StringMatchQ[x, "QName"] || 
  StringMatchQ[x, "normalizedString"] || 
  StringMatchQ[x, "token"] || 
  StringMatchQ[x, "language"] || 
  StringMatchQ[x, "Name"] || 
  StringMatchQ[x, "NCName"] || 
  StringMatchQ[x, "base64"] || 
  StringMatchQ[x, "anyType"]

XMLSchemaBaseTypeQ[___] := False

(* Used to identify type *)
XMLSchemaInstanceStringQ[x_String] :=   StringMatchQ[x, "http://www.w3.org/1999/XMLSchema-instance"] || 
				        StringMatchQ[x, "http://www.w3.org/2000/XMLSchema-instance"] || 
				        StringMatchQ[x, "http://www.w3.org/2001/XMLSchema-instance"]

XMLSchemaInstanceStringQ[___] := False
	
(* 
 * GetType for XMLElement with type defined in an attribute as a List
 *)
GetType[m:XMLElement[(_String | {_String, _String}), 
                     {___, {_String?XMLSchemaInstanceStringQ, 			
                     		"type"}->type:{_String,_String}, ___}, 
                     {___}, 
                     ___], 
        namespaces_List, 
        ___] := type

(* 
 * GetType for XMLElement with type defined in an attribute as a String
 *) 
GetType[m:XMLElement[(_String | {_String, _String}), 
                     {___, {_String?XMLSchemaInstanceStringQ, 			
                     		"type"}->type_String, ___}, 
                     {___}, 
                     ___], 
        namespaces_List, 
        ___] := 
	GetType[type, namespaces]

(* 
 * GetType for XMLElement with type not defined in an attribute, so 
 * the default is used.
 *) 
GetType[m:XMLElement[(_String | {_String, _String}), 
                     {___}, 
                     {child__}, 
                     opts___], 
        namespaces_List,
        default:{_String, _String}] := default

(* 
 * GetType for XMLElement syntax.  The child is used to determine the type. 
 *)
GetType[m:XMLElement[(_String | {_String, _String}), 
                     {___}, 
                     {val_String},
                     opts___],  
	      namespaces_List,
	      ___] := {"http://www.w3.org/2001/XMLSchema", "string"}

(* 
 * GetType for XMLElement syntax.  The name is used to determine the type. 
 *)
GetType[m:XMLElement[name:{_String, _String}, 
                     {___}, 
                     {___},
                     opts___],  
	      namespaces_List,
	      ___] := name
	
(* 
 * GetType for Rule syntax with type defined in an attribute as a List
 *)
GetType[parameter:({(_String | {_String, _String}), 
                   attrs:{___, {_String?XMLSchemaInstanceStringQ, 			
                   		"type"}->type:{_String,_String}, ___}}->val_), 
        namespaces_List,
        ___] := type

(* 
 * GetType for Rule syntax with type defined in an attribute as a String
 *)
GetType[parameter:({(_String | {_String, _String}), 
                   attrs:{___, {_String?XMLSchemaInstanceStringQ, 			
                   		"type"}->type_String, ___}}->val_), 
        namespaces_List,
        ___] := 
	GetType[type, namespaces]

(* 
 * GetType for Rule syntax with type not defined in an attribute, so 
 * the default is used.
 *) 
GetType[parameter:((name_String | {(name_String | 
	                 		{ns_String, name_String}), 
	                 	attrs:{___}})->val_), 
	      namespaces_List,
        default:{_String, _String}] := default


(* 
 * GetType for Rule syntax.  The child is used to determine the type.
 *)
GetType[parameter:((name_String | {(name_String | 
	                 		{ns_String, name_String}), 
	                 	attrs:{___}})->val_), 
	      namespaces_List,
	      ___] := 
	Module[{},
		Switch[
			val,
	 		True | False, {"http://www.w3.org/2001/XMLSchema", "boolean"},
   		_String, {"http://www.w3.org/2001/XMLSchema", "string"},
			_Integer, {"http://www.w3.org/2001/XMLSchema", "int"},
   		_Real, {"http://www.w3.org/2001/XMLSchema", "double"},
			_, {}
		]
	]

(*
 * GetType where the name is used to determine the type for 
 * complex objects 
 *)
GetType[parameter:({name:{_String, _String}, attrs:{___}}->___), 
	      namespaces_List,
        default:{_String, _String}] := name

(* 
 * GetType for Rule syntax definition with type defined in an 
 * attribute as a List 
 *)
GetType[parameter:({(name_String | {ns_String, name_String}), 
          				  attrs:{___, 
        			      {_String?XMLSchemaInstanceStringQ, "type"}->
        			      	type:{_String, _String}, ___}}), 
				        	namespaces_List] := type
	
(* 
 * GetType for Rule syntax definition with type defined in an 
 * attribute as a String 
 *)
GetType[parameter:({(name_String | {ns_String, name_String}), 
        					  attrs:{___, 
        			      {_String?XMLSchemaInstanceStringQ, "type"}->
        			      	type_String, ___}}), 
				        	namespaces_List] := 
	GetType[type, namespaces]

(* 
 * GetType for QName String representation
 *)
 
WrappedStringQ[x_String] := StringMatchQ[x, "{*}*"]

WrappedStringQ[___] := False

GetType[t_String?WrappedStringQ, namespaces_List] := 
	Module[{position, prefix, namespace, type},
		type = t;
		position = Part[StringPosition[type, "}"], 1, 2];
		namespace = StringTake[type, {2, position - 1}];
		type = StringTake[type, {position+1, StringLength[type]}];
		{namespace, type}
	]

(* 
 * GetType for prefixed string value 
 *)	
PrefixedStringQ[x_String] :=   StringMatchQ[x, "*:*"]

PrefixedStringQ[___] := False

GetType[t_String?PrefixedStringQ, namespaces_List] :=
	Module[{position, prefix, namespace, type},
		type = t;
		position = First[First[StringPosition[type, ":"]]];
		prefix = StringTake[type, {1, position - 1}];
		namespace = Select[namespaces, MatchQ[#, {_?XMLNamespaceStringQ, prefix}->_]&];
		If[MatchQ[namespace, {__Rule}], namespace = namespace[[1,2]], prefix = "None"];
		type = StringTake[type, {position+1, StringLength[type]}];
		If[StringMatchQ[prefix, "None"], t, {namespace, type}]
	]	

(* 
 * QName for Strings without a prefix or namespace
 *) 
GetType[t_String, namespaces_List] := {"", t}

(* 
 * Default GetType,  Could perhaps return something more useful or 
 * something that conforms to {_String, _String}
 *)
GetType[___] := {}

(* 
 * GetName for XMLElements
 *)
GetName[XMLElement[(name_String | name:{_String, _String}), 
                   {___}, 
                   {___}]] := name

(* 
 * GetName for Rule syntax
 *)
GetName[(name_String | {(name_String | name:{_String, _String}), 
        {___}})->_] := name

(* 
 * GetName for Rule syntax definition 
 *)
GetName[(name_String | {(name_String | name:{_String, _String}), {___}})] := name
 
GetName[___] := Null

(* 
 * GetQName for XMLElements for a List
 *)
GetQName[XMLElement[name:{_String, _String}, {___}, {___}]] := name

(* 
 * GetQName for XMLElements for a String
 *)
GetQName[XMLElement[name_String, {___}, {___}]] := {"", name}

(* 
 * GetName for Rule syntax for a List
 *)
GetQName[{name:{_String, _String}, {___}}->_] := name

(* 
 * GetName for Rule syntax for a String
 *)
GetQName[(name_String | {name_String, {___}})->_] := {"", name}

(* 
 * GetQName for Rule syntax definition for a List
 *)
GetQName[{name:{_String, _String}, {___}}] := name

(* 
 * GetQName for Rule syntax definition for a String
 *)
GetQName[{name_String, {___}}] := {"", name}

GetQName[___] := {}

(* 
 * GetLocalName for XMLElements
 *)
GetLocalName[XMLElement[(name_String | {_String, name_String}), 
                        {___},
                        {___}]] := name

(* 
 * GetLocalName for Rule syntax
 *)
GetLocalName[(name_String | {(name_String | {_String, name_String}), 
             {___}})->_] := name

(* 
 * GetLocalName for Rule syntax definition
 *)
GetLocalName[(name_String | {(name_String | {_String, name_String}), 
             {___}})] := name

GetLocalName[___] := Null

(* 
 * GetAttribute for XMLElements for attributes with namespace
 *)
GetAttribute[namespace_String,
             localpart_String,
             XMLElement[(_String | {_String, _String}), 
                        attrs:{___}, 
                        {___}]] :=
  Module[{cases},
    cases = Select[attrs, MatchQ[#, {namespace, localpart}->_]&];
    If[cases === {}, Null, cases]
  ]

(* 
 * GetAttribute for XMLElements for attributes without a namespace
 *)
GetAttribute[localpart_String,
             XMLElement[(_String | {_String, _String}), 
                        attrs:{___}, 
                        {___}]] :=
  Module[{cases},
    cases = Select[attrs, MatchQ[#, (localpart | {_String, localpart})->_]&];
    If[cases === {}, Null, cases]
  ]

(* 
 * GetAttribute for Rule syntax for attributes with namespace
 *)
GetAttribute[namespace_String,
             localpart_String,
             {(_String | {_String, _String}), attrs:{___}}->_] :=
  Module[{cases},
    cases = Select[attrs, MatchQ[#, {namespace, localpart}->_]&];
    If[cases === {}, Null, cases]
  ]

(* 
 * GetAttribute for Rule syntax for attributes without a namespace
 *)
GetAttribute[localpart_String,
             {(_String | {_String, _String}), attrs:{___}}->_] :=
  Module[{cases},
    cases = Select[attrs, MatchQ[#, (localpart | {_String, localpart})->_]&];
    If[cases === {}, Null, cases]
  ]

(* 
 * GetAttribute for Rule syntax definition for attributes with namespace
 *)
GetAttribute[namespace_String,
             localpart_String,
             {(_String | {_String, _String}), attrs:{___}}] :=
  Module[{cases},
    cases = Select[attrs, MatchQ[#, {namespace, localpart}->_]&];
    If[cases === {}, Null, cases]
  ]

(* 
 * GetAttribute for Rule syntax definition for attributes without a namespace
 *)
GetAttribute[localpart_String,
             {(_String | {_String, _String}), attrs:{___}}] :=
  Module[{cases},
    cases = Select[attrs, MatchQ[#, (localpart | {_String, localpart})->_]&];
    If[cases === {}, Null, cases]
  ]

GetAttribute[___] := Null

(* 
 * GetAttributes for XMLElements
 *)
GetAttributes[XMLElement[(_String | {_String, _String}), 
                         attrs:{___}, 
                         {___}]] := attrs

(* 
 * GetAttributes for Rule syntax
 *)
GetAttributes[{(_String | {_String, _String}), attrs:{___}}->_] := attrs


(* 
 * GetAttributes for Rule syntax definition
 *)
GetAttributes[{(_String | {_String, _String}), attrs:{___}}] := attrs

GetAttributes[___] := {}

(* 
 * GetValue for XMLElements
 *)
GetValue[XMLElement[(_String | {_String, _String}), 
                         {___}, 
                         value:{___}]] := value
                         
(* 
 * GetValue for Rule syntax
 *)
GetValue[{(_String | {_String, _String}), {___}}->value_] := value

GetValue[___] := Null

(* 
 * GetValueAttribute for XMLElements for attributes with namespace
 *)
GetValueAttribute[namespace_String,
                  localpart_String,
                  XMLElement[(_String | {_String, _String}), 
                             {___}, 
                        values:{___}]] :=
  Module[{cases},
    cases = Select[values, GetQName[#] === {namespace, localpart}&];
    If[cases === {}, Null, If[Length[cases] === 1, First[cases], cases]]
  ]

(* 
 * GetValueAttribute for XMLElements for attributes without a namespace
 *)
GetValueAttribute[localpart_String,
             XMLElement[(_String | {_String, _String}), 
                        {___}, 
                        values:{___}]] :=
  Module[{cases},  
    cases = Select[values, GetLocalName[#] === localpart &];
    If[cases === {}, Null, If[Length[cases] === 1, First[cases], cases]]
  ]

(* 
 * GetValueAttribute for Rule syntax for attributes with namespace
 *)
GetValueAttribute[namespace_String,
                  localpart_String,
                  {(_String | {_String, _String}), {___}}->value_] :=
  Module[{cases},  
    cases = Select[values, GetQName[#] === {namespace, localpart}&];
    If[cases === {}, Null, If[Length[cases] === 1, First[cases], cases]]
  ]
  
(* 
 * GetValueAttribute for Rule syntax for attributes without a namespace
 *)
GetValueAttribute[localpart_String,
             {(_String | {_String, _String}), attrs:{___}}->_] :=
  Module[{cases},
    cases = Select[values, GetLocalName[#] === localpart &];
    If[cases === {}, Null, If[Length[cases] === 1, First[cases], cases]]
  ]

GetValueAttribute[___] := Null

(* 
 * Used for converting Rule syntax to symbolic XML
 *) 
GetSymbolicXML[(name_String | {(name_String | name:{_String, _String}), 
               attrs:{___}})->val_] :=
  Module[{newAttrs = attrs, newVal = val},
		If[Head[Unevaluated[attrs]] =!= List, newAttrs = {}];
		If[!ListQ[val], newVal = {val}];
		newVal = GetSymbolicXML /@ newVal;
		XMLElement[name, newAttrs, newVal]
  ]

GetSymbolicXML[x___] := ToString[x]

(* Used for mapping namespaces contained in the Namespaces option and attributes.  This is needed
   because the type attribute generally contained a prefixed qualified name.  The Namespaces option
   is used to look up the prefix.  The attributes defining a namespace are found and added the 
   option (which can be set at runtime.)
 *)

(* Used to identify new namespaces *)
XMLNamespaceStringQ[x_String] := StringMatchQ[x, "http://www.w3.org/1999/xmlns/"] || 
				 StringMatchQ[x, "http://www.w3.org/2000/xmlns/"] || 
				 StringMatchQ[x, "http://www.w3.org/2001/xmlns/"] || 
				 StringMatchQ[x, "xmlns"]

XMLNamespaceStringQ[___] := False

namespaceSameQ[{namespace1_String, local1_String}->_, {namespace2_String, local2_String}->_] :=
  (namespace1 === namespace2 && local1 === local2);
  
namespaceSameQ[___] := False;

MapNamespaces[m:XMLElement[name:(_String | {_String, _String}), attrs:{___}, children:{___}], 
	      namespaces_List] :=
	Module[{},
		Join[Select[attrs, MatchQ[#, {_?XMLNamespaceStringQ, _String}->_String]&], 
		     namespaces
		]
	]

MapNamespaces[parameter:((name_String | {(name_String | {ns_String, name_String}), attrs:{___}})->val_), 
	      namespaces_List] :=
	Module[{},
		If[Head[Unevaluated[attrs]] === List, 
			Join[Select[attrs, 
		                     MatchQ[#, {_?XMLNamespaceStringQ, _String}->_String]&], 
		        namespaces],	              
			namespaces
		]
	]

MapNamespaces[parameter:((name_String | {(name_String | {ns_String, name_String}), attrs:{___}})), 
	      namespaces_List] :=
	Module[{},
		If[Head[Unevaluated[attrs]] === List, 
			Join[Select[attrs, 
		                     MatchQ[#, {_?XMLNamespaceStringQ, _String}->_String]&], 
		        namespaces],	              
			namespaces
		]
	]	
	
MapNamespaces[_, namespaces_List] := namespaces;
	
(* Jeff Adams originally wrote this code. *)

(*   Base 64 encoding and decoding 
--------------------------------------------------------------------------------- 
*)

(*
	Note: I include the explicit characters here rather than 
expressed as CharacterRange's since
	CharacterRange is not actually definied when the autoload 
package is created and this creates
	problems.
*)

$Base64Characters = {
	"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
	"N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z",
	"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
	"n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
	"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "+", "/" };

$Base64LineLength = 74;

(* Decodes a string from Base64 encoding *)
Base64Decode @ str_String :=
   With[{cs = Flatten[ (IntegerDigits[# - 1, 2, 6]) & /@
             Flatten [ Position[ $Base64Characters, #] & /@ Characters @ str]]},
   StringJoin[ FromCharacterCode @ FromDigits[#, 2] & /@
     Partition[Drop[cs, - Mod[Length @ cs, 8]], 8]]
   ];

(* Encodes a string into Base64 encoding *)
Base64Encode @ str_String :=
   Block[{cs = ToCharacterCode @ str, res},
   res = Flatten[ ConvertOneGroup /@
       Partition[Join[cs, Table[Null, {3 - Mod[Length @ cs, 3]}]], 3]];
   If[ Length @ res > $Base64LineLength,
     res = Join[ {#, "\n"}& /@ Partition[res, $Base64LineLength],
      {Take[res,-Mod[Length @ res, $Base64LineLength]]}] ];
   StringJoin @@ res
   ];
   
ConvertOneGroup @ {Null, Null, Null} = "";

ConvertOneGroup @ {f_  , Null, Null} :=
   { $Base64Characters[[1+FromDigits[#,2] ]]&/@
     Partition[ Flatten @ { IntegerDigits[f, 2, 8], {0,0,0,0} }, 6] , "=", "="};

ConvertOneGroup @ {f_  , s_  , Null} :=
   { $Base64Characters[[1+FromDigits[#,2] ]]&/@
       Partition[ Flatten @ { IntegerDigits[f, 2, 8], IntegerDigits[s, 
2, 8], {0,0} }, 6], "="};

ConvertOneGroup @ l_ :=
   $Base64Characters[[1+FromDigits[#,2] ]]&/@
     Partition[ Flatten[IntegerDigits[#, 2, 8]& /@ l], 6];

(*
 * Function used to clean up Strings used to make Symbols for InstallService and
 * InstallServiceOperation.
 *)     
NormalizeSymbolName[name_String] :=
  StringReplace[name, {"."->"", "_"->"", "~"->"", "!"->"", 
 				               "@"->"", "#"->"", "$"->"", "%"->"", "^"->"", 
							         "&"->"", "*"->"", "("->"", ")"->"", "-"->"", 
							         "+"->"", "="->"", "{"->"", "["->"", "}"->"", 
							         "]"->"", "|"->"", "\\"->"", ":"->"", ";"->"",
							         "\""->"", "\'"->"", "<"->"", ","->"", ">"->"",
		                   "?"->"", "/"->"", " "->""}]
     
End[]

EndPackage[]