(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 4.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    110439,       2639]*)
(*NotebookOutlinePosition[    123198,       3009]*)
(*  CellTagsIndexPosition[    122832,       2993]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["WebServices", "Title"],

Cell[CellGroupData[{

Cell["Introduction", "Section",
  CellTags->"b:0.0"],

Cell[TextData[{
  "More and more the World Wide Web is being used for application to \
application communication. The programmatic interfaces made available over \
the web are often referred to as web services.  There are many types of \
applications that can be considered web services but interoperability between \
applications is enhanced most by the use of familiar technologies such as XML \
and HTTP.  These technologies allow applications using differing languages \
and platforms to interface with one another in a familiar way.  More \
information on web services can be found at ",
  ButtonBox["http://www.w3.org/2002/ws/.",
    ButtonData:>{
      URL[ "http://www.w3.org/2002/ws/"], None},
    ButtonStyle->"Hyperlink"]
}], "Text",
  FontWeight->"Plain",
  CellTags->"b:0.0"],

Cell[CellGroupData[{

Cell["Technology", "Subsubsection",
  CellTags->"b:0.0"],

Cell[TextData[{
  "The web services client for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " allows users to discover and use web services in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  This is just one side to web services (providing Web Services being the \
other), but it is very important.  With this client, users can call \
operations that are based on other platforms or languages that are not \
immediately accessible to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  This opens up a whole new realm to users of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text",
  CellTags->"b:0.0"],

Cell["\<\
The World Wide Web Consortium (W3C) has introduced two API's that are highly \
useful for application to application communication.  These API's are Simple \
Object Access Protocol (SOAP) and Web Service Definition Language (WSDL).  \
These API's are the foundation for this web services client.\
\>", "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["XML and HTTP", "Subsubsection",
  CellTags->"b:0.0"],

Cell["\<\
A user interacts with the web service interfaces by sending an XML messages \
over HTTP.  XML and HTTP are useful for creating and sending these messages \
because they are flexible and widely supported on many platforms and \
languages.  This promotes interoperability between applications and allows \
applications using differing languages and platforms to interface with one \
another.\
\>", "Text",
  CellTags->"b:0.0"],

Cell["\<\
The user may find knowledge of XML to be useful..  However, it is not \
required.  The Expressions built to interact with the Web Service functions \
will mimic XML to a degree.  \
\>", "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["SOAP", "Subsubsection",
  CellTags->"b:0.0"],

Cell[TextData[{
  "SOAP formally sets the conventions governing the format and processing \
rules of an XML message.  SOAP is important because it standardizes the XML \
message used by the applications to communicate.  Without SOAP or a similar \
protocol, applications would not know how to process the XML messages.  For \
more information please see ",
  ButtonBox["http://www.w3.org/TR/SOAP/",
    ButtonData:>{
      URL[ "http://www.w3.org/TR/SOAP/"], None},
    ButtonStyle->"Hyperlink"],
  "."
}], "Text",
  FontWeight->"Plain",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["WSDL", "Subsubsection",
  CellTags->"b:0.0"],

Cell[TextData[{
  "WSDL formally sets the format for describing a web service.  By providing \
a URI to a WSDL,  applications can discover what operations a particular web \
service provides and what an operation's messages look like.  With this \
information a client can bind to the operation and use it.  A web services \
operation may be used without a WSDL, however, the WSDL is key for promoting \
a Web Service.  For information please see ",
  ButtonBox["http://www.w3.org/TR/wsdl",
    ButtonData:>{
      URL[ "http://www.w3.org/TR/wsdl"], None},
    ButtonStyle->"Hyperlink"],
  "."
}], "Text",
  FontWeight->"Plain",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Apache Axis", "Subsubsection",
  CellTags->"b:0.0"],

Cell[TextData[{
  "Rather than implement a whole new engine, Apache AXIS is used to implement \
part of the web services package technology.  It is a Java-based SOAP engine \
with extensive support for WSDL.  This engine is used by the web services \
package to invoke web service operations.  AXIS follows the SOAP 1.1 and WSDL \
1.1 specifications.  For more details on Axis see ",
  ButtonBox["http://xml.apache.org/axis.",
    ButtonData:>{
      URL[ "http://xml.apache.org/axis"], None},
    ButtonStyle->"Hyperlink"]
}], "Text",
  CellTags->"b:0.0"],

Cell["\<\
Despite being implemented in Java, the Web Services package is designed to be \
Java independent.  This means the user will not be required to provide Java \
parameters to functions, and Java objects will not be returned from \
functions.  Therefor the user is not required to know a thing about Java in \
order to use this package.\
\>", "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Version Requirements", "Subsubsection",
  CellTags->"b:0.0"],

Cell["\<\
The Web Services package requiresMathematica4.2 or later and is compatible \
with all platforms on which Mathematica available.The Web Services package \
also requires a Java runtime that is 1.3.1 or later.\
\>", "Text",
  CellTags->"b:0.0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Getting Started", "Section"],

Cell[CellGroupData[{

Cell["Base Types", "Subsection",
  CellTags->"b:0.1.0"],

Cell[TextData[{
  "Using WSDL technology, a user can discover and generate ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions for each web service operation.  The WSDL file specifies how to \
interact with a particular web service operation.  A user generates functions \
for interacting with these operations by passing a WSDL file to the \
InstallService function.  InstallService creates a  function for each \
operation defined in the WSDL.  Each function parses user input, formats the \
input into appropriate format, and then invokes the appropriate server \
operation.  When the function receives the results of the invocation, it \
parses the results and returns them in the appropriate format."
}], "Text",
  CellTags->"b:0.1.0"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[1]:=",
  CellTags->"b:0.1.0"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallService["\<http://www.xmethods.net/sd/BabelFishService.wsdl\>"]\)\
], "Input",
  CellLabel->"In[2]:=",
  CellTags->"b:0.1.0"],

Cell[BoxData[
    \({BabelFish}\)], "Output",
  CellLabel->"Out[2]=",
  CellTags->"b:0.1.0"]
}, Open  ]],

Cell["\<\
The functions created by InstallService are returned in a list.  After the \
service is installed, a user can find out information about the operations by \
using the OperationPalette.  The OperationPalette lists the parameters needed \
to invoke the operation and the data returned by the operation.\
\>", "Text",
  CellTags->"b:0.1.0"],

Cell[BoxData[
    \(OperationPalette[BabelFish]\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->"b:0.1.0"],

Cell[TextData[{
  "The OperationPalette reveals two parameters of type ",
  StyleBox["{http://www.w3.org/1999/XMLSchema}string",
    FontSlant->"Italic"],
  " are required to invoke the BabelFish function.  The OperationPalette \
contains a Types section which may be used to view information about each \
type used by an operation.  A user can view information about a particular \
type by clicking on a blue hyperlink.  The Types section may also be \
navigated manually.  The Types section specifies that the parameters for this \
function are represented by Strings in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  A user can make calls to BabelFish using two Strings as parameters.  "
}], "Text",
  CellTags->"b:0.1.0"],

Cell[TextData[{
  "As a convenience the user can paste a template for a function call by \
clicking the Paste button in the OperationPalette.  This will paste the \
function as it should be called in ",
  StyleBox["Mathematica, ",
    FontSlant->"Italic"],
  " but it will contain place holders for the values."
}], "Text",
  CellTags->"b:0.1.0"],

Cell[CellGroupData[{

Cell[BoxData[
    \(BabelFish["\<en_de\>", "\<Hello my friend!\>"]\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->"b:0.1.0"],

Cell[BoxData[
    \("hallo mein Freund!"\)], "Output",
  CellLabel->"Out[4]=",
  CellTags->"b:0.1.0"]
}, Open  ]],

Cell[TextData[{
  "To find out more information about the BabelFish Web Service please visit \
",
  ButtonBox["http://babelfish.altavista.com/.",
    ButtonData:>{
      URL[ "http://babelfish.altavista.com/"], None},
    ButtonStyle->"Hyperlink"]
}], "Text",
  CellTags->"b:0.1.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compound Types", "Subsection",
  CellTags->"b:0.1.1"],

Cell[TextData[{
  "Some web service operations require compound types as parameters.  \
Compound types are data that cannot be represented by a single simple value \
such as a String, Integer, Real, True or False.  Compound types are a \
sequence of base types.  They are bounded by a List in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  Items inside a compound type are represented by a Rule.  The left side \
of the rule is the description of the element.  The description of the \
element contains its name but may also contain its type and other attributes \
describing it (See ",
  ButtonBox["Type Mapping",
    ButtonData:>"Type Mapping",
    ButtonStyle->"Hyperlink"],
  ").  The right side of the Rule will be the value of the element.  The \
value of the element may be a simple or compound type."
}], "Text",
  CellTags->"b:0.1.1"],

Cell["\<\
In the example below, AuthorSearchRequest has a parameter that requires a \
compound value.\
\>", "Text",
  CellTags->"b:0.1.1"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[5]:=",
  CellTags->"b:0.1.1"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallService["\<http://soap.amazon.com/schemas/AmazonWebServices.wsdl\
\>"]\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->"b:0.1.1"],

Cell[BoxData[
    \({KeywordSearchRequest, BrowseNodeSearchRequest, AsinSearchRequest, 
      UpcSearchRequest, AuthorSearchRequest, ArtistSearchRequest, 
      ActorSearchRequest, ManufacturerSearchRequest, DirectorSearchRequest, 
      ListManiaSearchRequest, SimilaritySearchRequest}\)], "Output",
  CellLabel->"Out[6]=",
  CellTags->"b:0.1.1"]
}, Open  ]],

Cell[TextData[{
  "As shown above you can use the OperationPalette to see the signature of \
the method.  Using the OperationPalette, you can see that this method \
requires a complex type with a qualified name of ",
  StyleBox["{urn:PI/DevCentral/SoapService}AuthorRequest",
    FontSlant->"Italic"],
  "."
}], "Text",
  CellTags->"b:0.1.1"],

Cell[BoxData[
    \(OperationPalette[AuthorSearchRequest]\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->"b:0.1.1"],

Cell[TextData[{
  "As mentioned before, the OperationPalette can be used to inspect \
information about the type.  The blue hyperlinks to can be used to jump to \
information about each type.  The OperationPalette can be also be used to \
paste a template of the type needed for the function.  In this case a user \
may click the gray ",
  StyleBox["AuthorSearchRequest",
    FontSlant->"Italic"],
  " button, and it pastes an expression representing an ",
  StyleBox["{urn:PI/DevCentral/SoapService}AuthorRequest",
    FontSlant->"Italic"],
  ".  A user can tab through the placeholders to easily input the values of \
the expression.  The OperationPalette may be used as a reference for \
inputting these values.  A user may click on the blue hyperlink, ",
  StyleBox["{urn:PI/DevCentral/SoapService}AuthorRequest",
    FontSlant->"Italic"],
  ", to go to documentation describing the type.  Once the placeholders are \
filled, the user may invoke the function."
}], "Text",
  CellTags->"b:0.1.1"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AuthorSearchRequest[{"\<author\>" -> "\<Stephen Wolfram\>", \ \
"\<page\>" -> "\<1\>", \ "\<mode\>" -> \ "\<books\>", \ "\<tag\>" -> \
"\<webservices-20\>", "\<type\>" -> "\<lite\>", \ "\<devtag\>" -> \
"\<D3VZJS46JD5U9D\>", \ \ "\<version\>" -> "\<1.0\>"}]\)], "Input",
  CellLabel->"In[8]:=",
  CellTags->"b:0.1.1"],

Cell[BoxData[
    \({"Details" \[Rule] {{"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/1579550088",\
 "Asin" \[Rule] "1579550088", "ProductName" \[Rule] "A New Kind of Science", 
            "Catalog" \[Rule] "Book", "Authors" \[Rule] {"Stephen Wolfram"}, 
            "ReleaseDate" \[Rule] "14 May, 2002", 
            "Manufacturer" \[Rule] "Wolfram Media, Inc.", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/1579550088.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/1579550088.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/1579550088.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$44.95", "OurPrice" \[Rule] "$44.95", 
            "UsedPrice" \[Rule] "$37.95"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/1579550207",\
 "Asin" \[Rule] "1579550207", 
            "ProductName" \[Rule] 
              "A New Kind of Science: A New Kind of Science Explorer bundle", 
            "Catalog" \[Rule] "Book", 
            "Authors" \[Rule] {"Stephen Wolfram", "Wolfram Research Inc."}, 
            "ReleaseDate" \[Rule] "01 September, 2002", 
            "Manufacturer" \[Rule] "Wolfram Media, Inc.", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/1579550207.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/1579550207.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/1579550207.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$95.00", 
            "OurPrice" \[Rule] "$95.00"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0521643147",\
 "Asin" \[Rule] "0521643147", "ProductName" \[Rule] "The Mathematica Book", 
            "Catalog" \[Rule] "Book", "Authors" \[Rule] {"Stephen Wolfram"}, 
            "ReleaseDate" \[Rule] "15 March, 1999", 
            "Manufacturer" \[Rule] "Cambridge Univ Pr (Trd)", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0521643147.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0521643147.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0521643147.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$53.00", "OurPrice" \[Rule] "$37.10", 
            "UsedPrice" \[Rule] "$29.50"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0201626640",\
 "Asin" \[Rule] "0201626640", 
            "ProductName" \[Rule] "Cellular Automata and Complexity", 
            "Catalog" \[Rule] "Book", "Authors" \[Rule] {"Stephen Wolfram"}, 
            "ReleaseDate" \[Rule] "July, 2002", 
            "Manufacturer" \[Rule] "Perseus Publishing", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0201626640.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0201626640.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0201626640.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$35.00", "OurPrice" \[Rule] "$35.00", 
            "UsedPrice" \[Rule] "$9.99"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/020151012X",\
 "Asin" \[Rule] "020151012X", 
            "ProductName" \[Rule] 
              "Mathematica: Reference Guide/for Mathematica Version 2", 
            "Catalog" \[Rule] "Book", "Authors" \[Rule] {"Stephen Wolfram"}, 
            "ReleaseDate" \[Rule] "May, 1992", 
            "Manufacturer" \[Rule] "Addison-Wesley Publishing", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/020151012X.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/020151012X.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/020151012X.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$21.95", "OurPrice" \[Rule] "$21.95", 
            "UsedPrice" \[Rule] "$9.95"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0124649904",\
 "Asin" \[Rule] "0124649904", 
            "ProductName" \[Rule] "The Mathematica Programmer", 
            "Catalog" \[Rule] "Book", 
            "Authors" \[Rule] {"Roman E. Maeder", "Stephen Wolfram"}, 
            "ReleaseDate" \[Rule] "January, 1994", 
            "Manufacturer" \[Rule] "Academic Press", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0124649904.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0124649904.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0124649904.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$48.00", "OurPrice" \[Rule] "$48.00", 
            "UsedPrice" \[Rule] "$15.00"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0521585856",\
 "Asin" \[Rule] "0521585856", 
            "ProductName" \[Rule] 
              "Mathematica \[RegisteredTrademark] 3.0 Standard Add-on \
Packages", "Catalog" \[Rule] "Book", 
            "Authors" \[Rule] {"Stephen Wolfram", 
                "Inc. Staff Wolfram Research"}, 
            "ReleaseDate" \[Rule] "December, 1996", 
            "Manufacturer" \[Rule] "Cambridge Univ Pr (Trd)", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0521585856.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0521585856.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0521585856.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$35.00", "OurPrice" \[Rule] "$35.00", 
            "UsedPrice" \[Rule] "$3.82"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0201515024",\
 "Asin" \[Rule] "0201515024", 
            "ProductName" \[Rule] 
              "Mathematica: A System for Doing Mathematics by Computer", 
            "Catalog" \[Rule] "Book", "Authors" \[Rule] {"Stephen Wolfram"}, 
            "ReleaseDate" \[Rule] "March, 1991", 
            "Manufacturer" \[Rule] "Addison-Wesley Pub Co", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0201515024.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0201515024.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0201515024.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$51.95", "OurPrice" \[Rule] "$51.95", 
            "UsedPrice" \[Rule] "$2.19"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0201554798",\
 "Asin" \[Rule] "0201554798", 
            "ProductName" \[Rule] "Mathematica: The Student Book", 
            "Catalog" \[Rule] "Book", 
            "Authors" \[Rule] {"Stephen Wolfram", "George Beck"}, 
            "ReleaseDate" \[Rule] "July, 1900", 
            "Manufacturer" \[Rule] "Addison-Wesley Pub Co", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0201554798.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0201554798.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0201554798.01.LZZZZZZZ.jpg", 
            "ListPrice" \[Rule] "$32.95", "OurPrice" \[Rule] "$32.95", 
            "UsedPrice" \[Rule] "$3.95"}, {"Url" \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0521644283",\
 "Asin" \[Rule] "0521644283", 
            "ProductName" \[Rule] 
              "The MATHEMATICA \[RegisteredTrademark] Book, Version 4", 
            "Catalog" \[Rule] "Book", "Authors" \[Rule] {"Stephen Wolfram"}, 
            "Manufacturer" \[Rule] "Cambridge University Press", 
            "ImageUrlSmall" \[Rule] 
              "http://images.amazon.com/images/P/0521644283.01.THUMBZZZ.jpg", 
            "ImageUrlMedium" \[Rule] 
              "http://images.amazon.com/images/P/0521644283.01.MZZZZZZZ.jpg", 
            "ImageUrlLarge" \[Rule] 
              "http://images.amazon.com/images/P/0521644283.01.LZZZZZZZ.jpg"}}\
}\)], "Output",
  CellLabel->"Out[8]=",
  CellTags->"b:0.1.1"]
}, Open  ]],

Cell["\<\
The result that is returned is similar to the input that provided except that \
it is a bit more lengthy.  It is a compound type as well.  The syntax for the \
output is the same as the input.  The OperationPalette can also be used to \
inspect information about the output type.\
\>", "Text",
  CellTags->"b:0.1.1"],

Cell[TextData[{
  "To find out more information about the Amazon Web Services please visit ",
  ButtonBox["http://www.amazon.com/webservices.",
    ButtonData:>{
      URL[ "http://www.amazon.com/webservices"], None},
    ButtonStyle->"Hyperlink"]
}], "Text",
  CellTags->"b:0.1.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Examples", "Section",
  CellTags->"b:0.2"],

Cell[TextData[{
  "These are some examples which demonstrate packaging Web Services ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " code into a function or interface that is friendly to the user.\n"
}], "Text",
  CellTags->"b:0.2"],

Cell[TextData[ButtonBox["Amazon Web Service",
  ButtonData:>{
    FrontEnd`FileName[ {"../../Examples"}, "Amazon.nb", CharacterEncoding -> 
      "WindowsANSI"], None},
  ButtonStyle->"Hyperlink"]], "Text",
  CellDingbat->"\[FilledCircle]",
  CellTags->"b:0.2"],

Cell[TextData[ButtonBox["XMethods Query Web Service",
  ButtonData:>{
    FrontEnd`FileName[ {"../../Examples"}, "XMethods.nb", CharacterEncoding -> 
      "WindowsANSI"], None},
  ButtonStyle->"Hyperlink"]], "Text",
  CellDingbat->"\[FilledCircle]",
  CellTags->"b:0.2"],

Cell[TextData[ButtonBox["TerraService",
  ButtonData:>{
    FrontEnd`FileName[ {"../../Examples"}, "TerraServiceUI.nb", 
      CharacterEncoding -> "WindowsANSI"], None},
  ButtonStyle->"Hyperlink"]], "Text",
  CellDingbat->"\[FilledCircle]",
  CellTags->"b:0.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Advanced Topics", "Section"],

Cell[CellGroupData[{

Cell["Rule Syntax", "Subsection",
  CellTags->{"Type Mapping", "b:0.3.0"}],

Cell[TextData[{
  "While invoking a web service, the user may be required to input parameters \
of several different types.  Since Web Services is an XML based standard, \
there needs to be some mapping between ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " types and XML types.   The Rule syntax is used to provide this mapping.  \
The Rule syntax may be compared to Symbolic XML in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  However, the Rule syntax offers two advantages over Symbolic XML.  The \
first advantage is that it may contain values other than Strings.  The user \
will not be required to transform the values to and from Strings.  The second \
advantage of the Rule syntax is that it offers a shorter way to display an \
XML element."
}], "Text",
  CellTags->"b:0.3.0"],

Cell[CellGroupData[{

Cell["Basic Rule Syntax", "Subsubsection",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "The basic format of the Rule syntax is ",
  StyleBox["name_String->value_",
    FontSlant->"Italic"],
  ".   This syntax is much shorter than the XMLElement of Symbolic XML.  The \
name of the element is specified on the left side and the value is specified \
on the right side.  This syntax is used at two different times.  When the \
user invokes a web service, the parameters, in Rule syntax, are passed to the \
function and serialized into XML.  When the user receives the result from the \
function, the message is deserialized from XML into a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " Expression.  This Expression follows the Rule syntax.  Therefor the Rule \
syntax plays a very important role in the Web Services package.  "
}], "Text",
  CellTags->"b:0.3.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["LongForm", "Subsubsection",
  CellTags->"b:0.3.0"],

Cell["\<\
One advantage of using the Rule syntax is that there are multiple forms of \
the Rule syntax.  So it is possible to return more or less of the information \
returned.  The option LongForm controls how much information is returned \
during the deserialization of results.  LongForm determines if the attributes \
will be returned with the names and values.  If LongForm is set to False, \
then the function will return the basic rule syntax.  If LongForm is set to \
True, then the names, attributes, and values will all be returned.  LongForm \
is set to False by default.\
\>", "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "The basic Rule syntax does not allow for the user to input attributes or a \
qualified name for an element.  Most of the time this is ok, if the operation \
being called has a WSDL file associated with it.  However, sometimes it is \
important for the user to be able to input the qualified name or the \
attributes of an element.  Perhaps the user is using a web service without a \
WSDL.  The long form syntax is ",
  StyleBox["{{namespace_String, name_String}, attributes:{___Rule}}->value_.  \
",
    FontSlant->"Italic"],
  "The difference between the basic and full Rule syntax is entirely on the \
left side.  The full syntax contains a list of two items on the left side of \
the rule.  The first item is the name.  This is specified on the left side of \
the Rule in the basic syntax.  The second item is the attributes list.  This \
is something that cannot be specified in the basic Rule syntax."
}], "Text",
  CellTags->"b:0.3.0"],

Cell["\<\
When specifying the name in the full Rule syntax, the user may specify the \
name as a single String.  The user may also specify a qualified name as a \
list of two Strings.  The first string specifies the namespace URI for the \
qualified name and the second string specifies the local name.\
\>", "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "When specifying attributes in the full Rule syntax, the user may specify \
an attribute using a Rule.  The left hand side of the Rule contains the name \
of the attribute.  It may be a String or a qualified name as described above \
for the name.  The right hand side is the value of the attribute.  The value \
may be a String, Integer, Real, True, False, or a qualified name as described \
above for the name.  A qualified name may also be specified using a special \
String syntax of \"",
  StyleBox["{<namespaceURI>}name\".  ",
    FontSlant->"Italic"]
}], "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "One important attribute that may need to be set in order to use a web \
service is the type attribute.   A type can be set by setting the type \
attribute in the ",
  StyleBox["http://www.w3.org/2001/XMLSchema-instance",
    FontSlant->"Italic"],
  " namespace.  An example of this is shown below.  The bold code is used to \
set the type."
}], "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  StyleBox["{\"AuthorSearchRequest\", {\n    ", "InlineFormula"],
  StyleBox["{\"http://www.w3.org/2001/XMLSchema-instance\",\"type\"}->\n      \
\"{", "InlineFormula",
    FontWeight->"Bold"],
  StyleBox["urn:PI/DevCentral/SoapService",
    FontWeight->"Bold"],
  StyleBox["}", "InlineFormula",
    FontWeight->"Bold"],
  StyleBox["ActorRequest",
    FontWeight->"Bold"],
  StyleBox["\"", "InlineFormula",
    FontWeight->"Bold"],
  StyleBox["}}->{...}", "InlineFormula"]
}], "InlineFormula",
  CellTags->"b:0.3.0"],

Cell["\<\
As noted above, an attribute requiring a qualified name, such as the type, \
can also be setting using a list as such.  The example below specifies the \
type as an array.  Sometimes for arrays it is important to also set the \
arrayType attribute.  This is demonstrated below in italics.\
\>", "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  StyleBox["{\"doubleArray\", {\n    ", "InlineFormula"],
  StyleBox["{\"http://www.w3.org/2001/XMLSchema-instance\",\"type\"}->\n      \
  {\"", "InlineFormula",
    FontWeight->"Bold"],
  StyleBox["http://schemas.xmlsoap.org/soap/encoding/",
    FontWeight->"Bold"],
  StyleBox["\"",
    FontWeight->"Bold"],
  StyleBox[",\"", "InlineFormula",
    FontWeight->"Bold"],
  StyleBox["array",
    FontWeight->"Bold"],
  StyleBox["\"}, \n    ", "InlineFormula",
    FontWeight->"Bold"],
  StyleBox["{\"", "InlineFormula",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox["http://schemas.xmlsoap.org/soap/encoding/",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox["\",\"arrayType\"}->\"xsd:double[]\"", "InlineFormula",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox["}->{...}", "InlineFormula"]
}], "InlineFormula",
  CellTags->"b:0.3.0"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[9]:=",
  CellTags->"b:0.3.0"],

Cell[CellGroupData[{

Cell[BoxData[
    StyleBox[\(InstallService["\<http://soap.amazon.com/schemas/\
AmazonWebServices.wsdl\>"]\),
      FontSize->12]], "Input",
  CellLabel->"In[10]:=",
  CellTags->"b:0.3.0"],

Cell[BoxData[
    \({KeywordSearchRequest, BrowseNodeSearchRequest, AsinSearchRequest, 
      UpcSearchRequest, AuthorSearchRequest, ArtistSearchRequest, 
      ActorSearchRequest, ManufacturerSearchRequest, DirectorSearchRequest, 
      ListManiaSearchRequest, SimilaritySearchRequest}\)], "Output",
  CellLabel->"Out[10]=",
  CellTags->"b:0.3.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ActorSearchRequest[{{{"\<urn:PI/DevCentral/SoapService\>", \
"\<actor\>"}, {{"\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \
\[Rule] {"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] \
"\<Cruise\>", {{"\<urn:PI/DevCentral/SoapService\>", "\<page\>"}, \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \[Rule] \
{"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] "\<1\>", {{"\
\<urn:PI/DevCentral/SoapService\>", "\<mode\>"}, \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \[Rule] \
{"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] "\<dvd\>", \
{{"\<urn:PI/DevCentral/SoapService\>", "\<tag\>"}, \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \[Rule] \
{"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] \
"\<webservices-20\>", {{"\<urn:PI/DevCentral/SoapService\>", "\<type\>"}, {{"\
\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \[Rule] \
{"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] "\<lite\>", \
{{"\<urn:PI/DevCentral/SoapService\>", "\<devtag\>"}, \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \[Rule] \
{"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] \
"\<D3VZJS46JD5U9D\>", {{"\<urn:PI/DevCentral/SoapService\>", "\<version\>"}, \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", "\<type\>"} \[Rule] \
{"\<http://www.w3.org/1999/XMLSchema\>", "\<string\>"}}} \[Rule] "\<1.0\>"}, 
      LongForm \[Rule] True]\)], "Input",
  CellLabel->"In[11]:=",
  CellTags->"b:0.3.0"],

Cell[BoxData[
    \({{{"urn:PI/DevCentral/SoapService", 
            "Details"}, {{"http://schemas.xmlsoap.org/soap/encoding/", 
                "arrayType"} \[Rule] 
              "amazon:Details[10]", \
{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
              "SOAP-ENC:Array"}} \[Rule] {{{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B00005JL78",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B00005JL78", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Minority Report (Widescreen Edition)", \
{{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "17 December, 2002", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Universal Studios", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005JL78.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005JL78.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005JL78.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$29.99", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$22.49", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$12.00"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B000063UR2",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B000063UR2", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Legend (Ultimate Edition)", {{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "24 February, 2003", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Umvd", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B000063UR2.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B000063UR2.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B000063UR2.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$24.98", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$21.49", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$9.99"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/6305127689",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "6305127689", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Top Gun", {{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "08 January, 2002", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Paramount Studio", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/6305127689.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/6305127689.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/6305127689.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$24.99", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$21.49", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$12.99"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B0000639FE",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B0000639FE", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Jerry Maguire (Special Edition)", \
{{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "30 April, 2002", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Columbia Tristar Hom", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B0000639FE.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B0000639FE.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B0000639FE.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$27.95", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$23.99", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$15.49"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B00000ILBL",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B00000ILBL", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Days of Thunder", {{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "08 January, 2002", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Paramount Studio", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00000ILBL.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00000ILBL.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00000ILBL.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$24.99", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$22.49", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$16.19"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/0783226810",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "0783226810", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Far and Away", {{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "24 February, 2003", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Universal Studios", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/0783226810.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/0783226810.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/0783226810.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$26.98", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$24.28", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$22.79"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B00005B6JZ",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B00005B6JZ", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "A Few Good Men (Special Edition)", \
{{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "29 May, 2001", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Columbia/Tristar Studios", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005B6JZ.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005B6JZ.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005B6JZ.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$14.95", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$12.99", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$11.85"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B00005JKMZ",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B00005JKMZ", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Vanilla Sky", {{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "13 May, 2003", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Paramount Home Video", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005JKMZ.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005JKMZ.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00005JKMZ.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$29.99", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$22.49", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$7.99"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B00002E237",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B00002E237", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "The Outsiders", {{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "30 November, 1999", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Warner Studios", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00002E237.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00002E237.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00002E237.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$14.98", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$12.99", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$7.99"}, {{{"urn:PI/DevCentral/SoapService", 
                  "Url"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "http://www.amazon.com/exec/obidos/redirect?tag=webservices-20%\
26creative=D3VZJS46JD5U9D%26camp=2025%26link_code=sp1%26path=ASIN/B00003CWTI",\
 {{"urn:PI/DevCentral/SoapService", 
                  "Asin"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "B00003CWTI", {{"urn:PI/DevCentral/SoapService", 
                  "ProductName"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "Magnolia - New Line Platinum Series", \
{{"urn:PI/DevCentral/SoapService", 
                  "Catalog"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "DVD", {{"urn:PI/DevCentral/SoapService", 
                  "ReleaseDate"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "29 August, 2000", {{"urn:PI/DevCentral/SoapService", 
                  "Manufacturer"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "New Line Studios", {{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlSmall"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00003CWTI.01.THUMBZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlMedium"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00003CWTI.01.MZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ImageUrlLarge"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "http://images.amazon.com/images/P/B00003CWTI.01.LZZZZZZZ.jpg", \
{{"urn:PI/DevCentral/SoapService", 
                  "ListPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] 
              "$29.95", {{"urn:PI/DevCentral/SoapService", 
                  "OurPrice"}, {{"http://www.w3.org/1999/XMLSchema-instance", 
                      "type"} \[Rule] "xsd:string"}} \[Rule] 
              "$25.49", {{"urn:PI/DevCentral/SoapService", 
                  "UsedPrice"}, \
{{"http://www.w3.org/1999/XMLSchema-instance", "type"} \[Rule] 
                    "xsd:string"}} \[Rule] "$15.99"}}}\)], "Output",
  CellLabel->"Out[11]=",
  CellTags->"b:0.3.0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Type Mapping", "Subsubsection",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "The main reason Rule syntax can use types such as Integers, Reals, etc. is \
because web services defines strong types.  Through SOAP encoding and WSDL \
files, data passed as parameters and returned as results is strongly typed.  \
This means that either the data encoding or the WSDL file tells the client \
how to decode the data.  This allows the client to serialize these ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " types other than Strings.  It also allows the client to deserialize the \
XML types into these ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " types.  WSDL files also minimize the use of attributes allowing the basic \
Rule syntax to contain a simple name and value Rule."
}], "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "The types of values supported by the Rule syntax are Rule, XMLElement, \
List, String, Integer, Real, True, False, and Null.  This client does not \
have a validation mechanism to check the users input.  Therefor when passing \
",
  StyleBox["Mathematica ",
    FontSlant->"Italic"],
  "parameters to a web service function any type supported by the Web \
Services client may be used.  Parameters are all eventually converted to \
Strings.  Type information and other information is also added to these \
parameters using data from the WSDL before it is sent to the web service.  \
This helps the web service discern the data sent."
}], "Text",
  CellTags->"b:0.3.0"],

Cell[TextData[{
  "When the results are returned from the web service, XML is deserialized \
into ",
  "Rule syntax",
  ".  XML base types are mapped into ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " base types.  The mappings for XML to ",
  StyleBox["Mathematica ",
    FontSlant->"Italic"],
  "are defined in the table below.  The base64 and base64binary types get \
decoded and stored as a string."
}], "Text",
  CellTags->"b:0.3.0"],

Cell[BoxData[GridBox[{
        {\(XML\ type\), \(Mathematica\ Type\)},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            boolean\), \(True, \ False\)},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} double\), 
          "Real"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} float\), 
          "Real"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} decimal\),
           "Real"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} int\), 
          "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} integer\),
           "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} long\), 
          "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} short\), 
          "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} byte\), 
          "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            nonPositiveInteger\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            negativeInteger\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            nonNegativeInteger\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            nonsignedLong\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            unsignedInteger\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            unsignedInt\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            unsignedShort\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            unsignedByte\), "Integer"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            base64Binary\), \(Decoded\ String\)},
        {\({\(\(http\)\(:\)\) // \(\(\(schemas . xmlsoap . org/soap\)/
                    encoding\)\(/\)\)} base64\), \(Decoded\ String\)},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} string\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            duration\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            datetime\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} time\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} date\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            gYearMonth\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} gYear\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            gMonthDay\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} gDay\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} gMonth\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            hexBinary\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} anyURI\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} QName\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            normalizedString\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} token\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            language\), "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} Name\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} NCName\), 
          "String"},
        {\({\(\(http\)\(:\)\) // \(\(\(schemas . xmlsoap . org/soap\)/
                    encoding\)\(/\)\)} array\), "List"},
        {\(Compound\ Type\), "List"}
        },
      GridFrame->True,
      RowLines->True,
      ColumnLines->True]], "Text",
  TextAlignment->Center,
  CellTags->"b:0.3.0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Naming Services", "Subsection",
  CellTags->"b:0.3.1"],

Cell[TextData[{
  "There are two options for installing a web service.  The first option \
allows the user to specify which context a web service will be installed \
into.  This option may be utilized by supplying a second argument which \
specifies the context.  The context should be specified using a String and \
must be defined under the legal rules specified by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text",
  CellTags->"b:0.3.1"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[12]:=",
  CellTags->"b:0.3.1"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallService["\<http://www.xmethods.net/sd/2001/BNQuoteService.wsdl\>\
", "\<BarnesAndNoble`\>"]\)], "Input",
  CellLabel->"In[13]:=",
  CellTags->"b:0.3.1"],

Cell[BoxData[
    \({getPrice}\)], "Output",
  CellLabel->"Out[13]=",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Context[getPrice]\)], "Input",
  CellLabel->"In[14]:=",
  CellTags->"b:0.3.1"],

Cell[BoxData[
    \("BarnesAndNoble`"\)], "Output",
  CellLabel->"Out[14]=",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(getPrice["\<043935806X\>"]\)], "Input",
  CellLabel->"In[15]:=",
  CellTags->"b:0.3.1"],

Cell[BoxData[
    \(17.99`\)], "Output",
  CellLabel->"Out[15]=",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell["\<\
The second option available for installing a web service is the \
AllowShortContext option.  By default the context of a web service is added \
to the $ContextPath.  This allows users to call the functions using short \
names.  However, if a user does not want the web service context added to \
$ContextPath, then the user should set AllowShortContext->False.  If this \
option is set to False, then the entire context will need to be specified in \
order to call a function.  \
\>", "Text",
  CellTags->"b:0.3.1"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallService["\<http://www.xmethods.net/sd/2001/\
CurrencyExchangeService.wsdl\>", AllowShortContext \[Rule] False]\)], "Input",\

  CellLabel->"In[16]:=",
  CellTags->"b:0.3.1"],

Cell[BoxData[
    \({CurrencyExchangeService`CurrencyExchangePort`getRate}\)], "Output",
  CellLabel->"Out[16]=",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CurrencyExchangeService`CurrencyExchangePort`getRate["\<us\>", \
"\<england\>"]\)], "Input",
  CellLabel->"In[17]:=",
  CellTags->"b:0.3.1"],

Cell[BoxData[
    \(0.6432`\)], "Output",
  CellLabel->"Out[17]=",
  CellTags->"b:0.3.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Working with Messages", "Subsection",
  CellTags->"b:0.3.2"],

Cell["\<\
Sometimes it is useful to work directly with the request message that is sent \
to a web service.  This message may be retrieved using the ToServiceRequest \
function.  ToServiceRequest is called by providing a web service function as \
the first parameter and its parameters as the additional parameters.  The \
function will not invoke the web service operation.  Rather the function will \
return the symbolic XML representation of the request message that would \
generally be sent to the web service.\
\>", "Text",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[18]:=",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallService["\<http://www.xmethods.net/sd/2001/CATrafficService.wsdl\
\>"]\)], "Input",
  CellLabel->"In[19]:=",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \({getTraffic}\)], "Output",
  CellLabel->"Out[19]=",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(symbolicXML\  = ToServiceRequest[getTraffic, "\<101\>"]\)], "Input",
  CellLabel->"In[20]:=",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"soapenv", 
          "Envelope"}, {{"xmlns", "soapenv"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", {"xmlns", 
              "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema", {"xmlns", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", {"xmlns", 
              "soapenc"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/encoding/"}, \
{XMLElement[{"soapenv", 
              "Body"}, {}, {XMLElement[{"ns0", 
                  "getTraffic"}, {{"soapenv", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", {"xmlns", 
                      "ns0"} \[Rule] 
                    "urn:xmethods-CATraffic"}, {XMLElement[{\*"\<\"\"\>", 
                      "hwynums"}, {{"xsi", "type"} \[Rule] 
                        "xsd:string"}, {"101"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[20]=",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell["\<\
A user may change the request message to have different values or different \
attributes.  In fact, the user may create an entirely new message.  However, \
the message must still follow the format specified in the SOAP specification. \
 Using ToServiceRequest gives the user a nice starting point.  The request \
message can then be invoked using the InvokeServiceOperation function.  The \
first parameter of this function is the symbol representing a web service \
function.  The web service function specifies the endpoint URI internally \
along with other options for invoking the service.  The second argument is \
the symbolic XML representation of the request message.\
\>", "Text",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InvokeServiceOperation["\<http://services.xmethods.net:80/soap/servlet/\
rpcrouter\>", symbolicXML]\)], "Input",
  CellLabel->"In[21]:=",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
          "Envelope"}, {{"http://www.w3.org/2000/xmlns/", "SOAP-ENV"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", \
{"http://www.w3.org/2000/xmlns/", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", \
{"http://www.w3.org/2000/xmlns/", "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema"}, \
{XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
              "Body"}, {}, {XMLElement[{"urn:xmethods-CATraffic", 
                  "getTrafficResponse"}, \
{{"http://schemas.xmlsoap.org/soap/envelope/", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", \
{"http://www.w3.org/2000/xmlns/", "ns1"} \[Rule] 
                    "urn:xmethods-CATraffic"}, {XMLElement[
                    "return", {{"http://www.w3.org/2001/XMLSchema-instance", 
                          "type"} \[Rule] 
                        "xsd:string"}, {"reported as of Monday, April 7, 2003 \
at 14:54 . US 101 [LOS ANGELES & VENTURA CO.'S] NO TRAFFIC RESTRICTIONS ARE \
REPORTED FOR THIS AREA. [CENTRAL COAST] NO TRAFFIC RESTRICTIONS ARE REPORTED \
FOR THIS AREA. [SAN FRANCISCO BAY AREA] NO TRAFFIC RESTRICTIONS ARE REPORTED \
FOR THIS AREA. [NORTHWEST CALIFORNIA] /AT THE CONFUSION HILL SLIDE/ FROM 7 MI \
NORTH OF LEGGETT TO THE JCT OF SR 271 (MENDOCINO CO) IS CURRENTLY OPEN. \
MOTORISTS ARE SUBJECT TO 20 MINUTE DELAYS MONDAY THRU FRIDAY FROM 8AM TO 6PM \
- DUE TO CONSTRUCTION"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[21]=",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(InvokeServiceOperation[getTraffic, symbolicXML]\)], "Input",
  CellLabel->"In[22]:=",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
          "Envelope"}, {{"http://www.w3.org/2000/xmlns/", "SOAP-ENV"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", \
{"http://www.w3.org/2000/xmlns/", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", \
{"http://www.w3.org/2000/xmlns/", "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema"}, \
{XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
              "Body"}, {}, {XMLElement[{"urn:xmethods-CATraffic", 
                  "getTrafficResponse"}, \
{{"http://schemas.xmlsoap.org/soap/envelope/", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", \
{"http://www.w3.org/2000/xmlns/", "ns1"} \[Rule] 
                    "urn:xmethods-CATraffic"}, {XMLElement[
                    "return", {{"http://www.w3.org/2001/XMLSchema-instance", 
                          "type"} \[Rule] 
                        "xsd:string"}, {"reported as of Monday, April 7, 2003 \
at 14:54 . US 101 [LOS ANGELES & VENTURA CO.'S] NO TRAFFIC RESTRICTIONS ARE \
REPORTED FOR THIS AREA. [CENTRAL COAST] NO TRAFFIC RESTRICTIONS ARE REPORTED \
FOR THIS AREA. [SAN FRANCISCO BAY AREA] NO TRAFFIC RESTRICTIONS ARE REPORTED \
FOR THIS AREA. [NORTHWEST CALIFORNIA] /AT THE CONFUSION HILL SLIDE/ FROM 7 MI \
NORTH OF LEGGETT TO THE JCT OF SR 271 (MENDOCINO CO) IS CURRENTLY OPEN. \
MOTORISTS ARE SUBJECT TO 20 MINUTE DELAYS MONDAY THRU FRIDAY FROM 8AM TO 6PM \
- DUE TO CONSTRUCTION"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[22]=",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell[TextData[{
  "The result of this call is the response message.  Sometimes it is useful \
to work directly with the response message.  However, web service functions \
by default return a deserialized version of the response message.  The \
original response message may be retrieved using the InvokeServiceOperation \
function.  The first parameter is the web service function and the function's \
parameters follow.  InvokeServiceOperation invokes the service and returns \
the entire response message in Symbolic XML.  This will allow the user to \
customize how the data is deserialized into ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " representation."
}], "Text",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(response\  = \ InvokeServiceOperation[getTraffic, "\<101\>"]\)], "Input",\

  CellLabel->"In[23]:=",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
          "Envelope"}, {{"http://www.w3.org/2000/xmlns/", "SOAP-ENV"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", \
{"http://www.w3.org/2000/xmlns/", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", \
{"http://www.w3.org/2000/xmlns/", "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema"}, \
{XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
              "Body"}, {}, {XMLElement[{"urn:xmethods-CATraffic", 
                  "getTrafficResponse"}, \
{{"http://schemas.xmlsoap.org/soap/envelope/", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", \
{"http://www.w3.org/2000/xmlns/", "ns1"} \[Rule] 
                    "urn:xmethods-CATraffic"}, {XMLElement[
                    "return", {{"http://www.w3.org/2001/XMLSchema-instance", 
                          "type"} \[Rule] 
                        "xsd:string"}, {"reported as of Monday, April 7, 2003 \
at 14:54 . US 101 [LOS ANGELES & VENTURA CO.'S] NO TRAFFIC RESTRICTIONS ARE \
REPORTED FOR THIS AREA. [CENTRAL COAST] NO TRAFFIC RESTRICTIONS ARE REPORTED \
FOR THIS AREA. [SAN FRANCISCO BAY AREA] NO TRAFFIC RESTRICTIONS ARE REPORTED \
FOR THIS AREA. [NORTHWEST CALIFORNIA] /AT THE CONFUSION HILL SLIDE/ FROM 7 MI \
NORTH OF LEGGETT TO THE JCT OF SR 271 (MENDOCINO CO) IS CURRENTLY OPEN. \
MOTORISTS ARE SUBJECT TO 20 MINUTE DELAYS MONDAY THRU FRIDAY FROM 8AM TO 6PM \
- DUE TO CONSTRUCTION"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[23]=",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell[TextData[{
  "Once a web service function is called using InvokeServiceOperation the \
results come back in the form of a SOAPEnvelope.FromServiceResponse can be \
used to extract the results from the SOAPEnvelope.If the message uses SOAP \
encoding and provides types in the response message,the data will be \
deserialized into the proper ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " types.  If not,the SOAPEnvelope wrappers will be stripped and the \
response will be deserialized into a Rule syntax expression with String \
values."
}], "Text",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FromServiceResponse[response]\)], "Input",
  CellLabel->"In[24]:=",
  CellTags->"b:0.3.2"],

Cell[BoxData[
    \("reported as of Monday, April 7, 2003 at 14:54 . US 101 [LOS ANGELES & \
VENTURA CO.'S] NO TRAFFIC RESTRICTIONS ARE REPORTED FOR THIS AREA. [CENTRAL \
COAST] NO TRAFFIC RESTRICTIONS ARE REPORTED FOR THIS AREA. [SAN FRANCISCO BAY \
AREA] NO TRAFFIC RESTRICTIONS ARE REPORTED FOR THIS AREA. [NORTHWEST \
CALIFORNIA] /AT THE CONFUSION HILL SLIDE/ FROM 7 MI NORTH OF LEGGETT TO THE \
JCT OF SR 271 (MENDOCINO CO) IS CURRENTLY OPEN. MOTORISTS ARE SUBJECT TO 20 \
MINUTE DELAYS MONDAY THRU FRIDAY FROM 8AM TO 6PM - DUE TO CONSTRUCTION"\)], \
"Output",
  CellLabel->"Out[24]=",
  CellTags->"b:0.3.2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Working without a WSDL", "Subsection",
  CellTags->"b:0.3.3"],

Cell[TextData[{
  "Web services may work without a WSDL file and sometimes the user may be \
forced into a situation where a WSDL file is not available.  It is possible \
to use this client without a WSDL.  Although it is much more convenient to \
use with a WSDL file.  In order to use the client without a WSDL it may be \
useful for the user to take a look at the SOAP specification.  This is \
available at  ",
  ButtonBox["http://www.w3.org/TR/SOAP/.",
    ButtonData:>{
      URL[ "http://www.w3.org/TR/SOAP/"], None},
    ButtonStyle->"Hyperlink"]
}], "Text",
  CellTags->"b:0.3.3"],

Cell["\<\
A user can invoke a web service using the InvokeServiceOperation.  In order \
to properly invoke this function the user must provide a proper SOAP message. \
 The ToServiceRequest function can be used to build a request message.  In \
order to use this function, two parameters must be provided.  The first \
parameter is a list of the parameter of the web service operation which will \
be invoked.  The second parameter is a list of the header's of the web \
service operation which will be invoked.  Both of these parameters should \
passed in as a List of  Rule syntax expressions or Symbolic XML elements.  It \
is necessary to pass the full definition of the parameters as a WSDL is not \
present to provide information about the parameter.  This includes the name, \
type, or any other information needed by the web service.  The options are \
also very important to building the SOAP envelope.  These are defined in the \
table below.  \
\>", "Text",
  CellTags->"b:0.3.3"],

Cell[BoxData[GridBox[{
        {
          StyleBox["Option",
            "Text"], 
          StyleBox["Description",
            "Text"]},
        {
          StyleBox["OperationName",
            "Text"], 
          
          StyleBox[\(OperationName\ is\ an\ option\ which\ specifies\ the\ \
name\ of\ \[IndentingNewLine]
            
            the\ operation\ tag\ for\ rpc - 
              style\ calls\ and\ the\ root\ tag\ \[IndentingNewLine]
            for\ single - document\ document - style\ \(\(calls\)\(.\)\)\),
            "Text"]},
        {
          StyleBox["OperationStyle",
            "Text"], 
          
          StyleBox[\(OperationStyle\ is\ an\ option\ which\ specifies\ the\ \
\[IndentingNewLine]
            
            operation\ style\ of\ the\ web\ service\ operation . 
                This\ may\ be\ \[IndentingNewLine]
            
            set\ to\ ' document'\ or\ ' 
              rpc'\ only . \ \ The\ default\ is\ \[IndentingNewLine]
            \(\(Automatic\)\(.\)\(\ \ \)\)\),
            "Text"]},
        {
          StyleBox["SOAPActionURI",
            "Text"], 
          
          StyleBox[\(SOAPActionURI\ is\ an\ option\ which\ specifies\ the\ \
SOAP\ \[IndentingNewLine]
            
            action\ URI\ of\ the\ web\ service\ operation . 
                The\ default\ is\ \[IndentingNewLine]
            \(\(Automatic\)\(.\)\)\),
            "Text"]},
        {
          StyleBox["ReturnType",
            "Text"], 
          
          StyleBox[\(ReturnType\ is\ used\ to\ specify\ the\ qualified\ name\ \
of\ the\ \[IndentingNewLine]
            
            root\ element\ returned\ in\ the\ results . \ \ This\ is\ used\ \
for\[IndentingNewLine]
            
            deserializing\ the\ results . \ \ The\ value\ should\ be\ a\ list\
\ \[IndentingNewLine]
            
            of\ two\ strings . \ \ The\ first\ string\ being\ the\ namespace\ \
\[IndentingNewLine]
            
            and\ the\ second\ string\ is\ the\ name\ of\ the\ \(\(type\)\(.\)\
\)\),
            "Text"]},
        {
          StyleBox["TransportStyleURI",
            "Text"], 
          
          StyleBox[\(TransportStyleURI\ is\ used\ to\ specify\ the\ style\ of\
\ \[IndentingNewLine]
            
            transport\ used\ to\ invoke\ the\ call . \ \ The\ default\ value\ \
is\[IndentingNewLine]
            \(\(Automatic\)\(.\)\(\ \ \)\)\),
            "Text"]},
        {
          StyleBox["EncodingStyle",
            "Text"], 
          
          StyleBox[\(EncodingStyle\ is\ an\ option\ which\ specifies\ whether\
\ the\ \[IndentingNewLine]
            
            message\ should\ be\ encoded\ or\ use\ a\ message\ based\ on\ \
\[IndentingNewLine]
            
            literal\ schema . \ \ This\ may\ be\ set\ to\ ' 
              "encoded"'\ or\ \[IndentingNewLine]
            \(\('\) \(literal'\)\(\ \)\(only . \ \ The\)\(\ \)\(default\)\(\ \
\)\(is\)\(\ \)\(\(Automatic\)\(.\)\)\)\),
            "Text"]},
        {
          StyleBox["EncodingStyleURI",
            "Text"], 
          
          StyleBox[\(EncodingStyleURI\ is\ used\ to\ specify\ the\ encoding\ \
of\ the\ \[IndentingNewLine]
            
            parameters\ passed\ to\ a\ web\ service\ operation . \ \ The\ \
value\[IndentingNewLine]
            
            should\ be\ a\ list\ of\ string\ representing\ a\ \
\(\(URI\)\(.\)\)\),
            "Text"]},
        {
          StyleBox["HeaderEncodingStyleURI",
            "Text"], 
          
          StyleBox[\(HeaderEncodingStyleURI\ is\ used\ to\ specify\ the\ \
encoding\ \[IndentingNewLine]
            
            of\ the\ header\ parameters\ passed\ to\ a\ web\ service\ \
\[IndentingNewLine]
            
            operation . \ \ The\ value\ should\ be\ a\ list\ of\ string\ \
\[IndentingNewLine]
            representing\ a\ \(\(URI\)\(.\)\)\),
            "Text"]},
        {
          StyleBox["Username",
            "Text"], 
          
          StyleBox[\(Username\ is\ used\ to\ supply\ authentication\ for\ an\ \
\[IndentingNewLine]
            operation . \ \ The\ value\ should\ be\ a\ \(\(string\)\(.\)\)\),
            "Text"]},
        {
          StyleBox["Password",
            "Text"], 
          
          StyleBox[\(Password\ is\ used\ to\ supply\ authentication\ for\ an\ \
\[IndentingNewLine]
            operation . \ \ The\ value\ should\ be\ a\ \(\(string\)\(.\)\)\),
            "Text"]}
        },
      GridFrame->True,
      RowLines->True,
      ColumnLines->True]], "Text",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[25]:=",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(msg = \ 
      ToServiceRequest[\[IndentingNewLine]{"\<zipcode\>" -> "\<61821\>"}, {}, \
\ OperationName -> {"\<urn:xmethods-Temperature\>", "\<getTemp\>"}]\)], \
"Input",
  CellLabel->"In[26]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"soapenv", 
          "Envelope"}, {{"xmlns", "soapenv"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", {"xmlns", 
              "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema", {"xmlns", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", {"xmlns", 
              "soapenc"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/encoding/"}, \
{XMLElement[{"soapenv", 
              "Body"}, {}, {XMLElement[{"ns0", 
                  "getTemp"}, {{"xmlns", "ns0"} \[Rule] 
                    "urn:xmethods-Temperature"}, {XMLElement[{\*"\<\"\"\>", 
                      "zipcode"}, {}, {"61821"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[26]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell["\<\
The request message can then be invoked using the InvokeServiceOperation \
function.  The first parameter of this function is the endpoint URI.  The \
endpoint specifies where to send the request message.  This URI is generally \
found in the WSDL.\
\>", "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InvokeServiceOperation["\<http://services.xmethods.net:80/soap/servlet/\
rpcrouter\>", \ msg]\)], "Input",
  CellLabel->"In[27]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
          "Envelope"}, {{"http://www.w3.org/2000/xmlns/", "SOAP-ENV"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", \
{"http://www.w3.org/2000/xmlns/", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", \
{"http://www.w3.org/2000/xmlns/", "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema"}, \
{XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
              "Body"}, {}, {XMLElement[{"urn:xmethods-Temperature", 
                  "getTempResponse"}, \
{{"http://schemas.xmlsoap.org/soap/envelope/", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", \
{"http://www.w3.org/2000/xmlns/", "ns1"} \[Rule] 
                    "urn:xmethods-Temperature"}, {XMLElement[
                    "return", {{"http://www.w3.org/2001/XMLSchema-instance", 
                          "type"} \[Rule] 
                        "xsd:float"}, {"43.0"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[27]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[TextData[{
  "When using SOAP encoding, the client will add types to the elements in the \
parameters list passed to ToServiceRequest depending on the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " type.  Therefor in the example below it is not necessary to add type \
information because SOAP encoding is specified.  The simple types supported \
are defined below.  However, when the results are returned, if the types are \
not encoded in the data returned, the values will not be converted from \
Strings."
}], "Text",
  CellTags->"b:0.3.3"],

Cell[BoxData[GridBox[{
        {\(Mathematica\ Type\), \(XML\ type\)},
        {"String", \({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            string\)},
        {"Integer", \({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/
                  XMLSchema} int\)},
        {"Real", \({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            double\)},
        {\(True, \ 
          False\), \({\(\(http\)\(:\)\) // \(www . w3 . org/1999\)/XMLSchema} 
            boolean\)}
        },
      GridFrame->True,
      RowLines->True,
      ColumnLines->True]], "Text",
  TextAlignment->Center,
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(msg = \ 
      ToServiceRequest[\[IndentingNewLine]{"\<zipcode\>" -> "\<61821\>"}, {}, \
\ OperationName -> {"\<urn:xmethods-Temperature\>", "\<getTemp\>"}, \ 
        EncodingStyle -> "\<encoded\>", 
        EncodingStyleURI \[Rule] \
{"\<http://schemas.xmlsoap.org/soap/encoding/\>"}]\)], "Input",
  CellLabel->"In[28]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"soapenv", 
          "Envelope"}, {{"xmlns", "soapenv"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", {"xmlns", 
              "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema", {"xmlns", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", {"xmlns", 
              "soapenc"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/encoding/"}, \
{XMLElement[{"soapenv", 
              "Body"}, {}, {XMLElement[{"ns0", 
                  "getTemp"}, {{"soapenv", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", {"xmlns", 
                      "ns0"} \[Rule] 
                    "urn:xmethods-Temperature"}, {XMLElement[{\*"\<\"\"\>", 
                      "zipcode"}, {{"xsi", "type"} \[Rule] 
                        "xsd:string"}, {"61821"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[28]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(InvokeServiceOperation["\<http://services.xmethods.net:80/soap/servlet/\
rpcrouter\>", \ msg]\)], "Input",
  CellLabel->"In[29]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
          "Envelope"}, {{"http://www.w3.org/2000/xmlns/", "SOAP-ENV"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", \
{"http://www.w3.org/2000/xmlns/", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", \
{"http://www.w3.org/2000/xmlns/", "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema"}, \
{XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
              "Body"}, {}, {XMLElement[{"urn:xmethods-Temperature", 
                  "getTempResponse"}, \
{{"http://schemas.xmlsoap.org/soap/envelope/", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", \
{"http://www.w3.org/2000/xmlns/", "ns1"} \[Rule] 
                    "urn:xmethods-Temperature"}, {XMLElement[
                    "return", {{"http://www.w3.org/2001/XMLSchema-instance", 
                          "type"} \[Rule] 
                        "xsd:float"}, {"43.0"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[29]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[TextData[{
  "However, if the web service is expecting an XML type other than the type \
mapped from a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " type, it may be necessary to add type information.  For example a Real \
maps to a double.  However, a Real also represents a float in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  If a Real value is passed as parameter it will be encoded as a double.  \
A web service expecting a float may not handle this correctly.  To correct \
this the user should specify the type as shown below.  It may also be \
necessary to specify the type for compound types.  Compound types do not have \
an XML type automatically associated with them."
}], "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(msg\  = \ 
      ToServiceRequest[{{"\<zipcode\>", \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", \ "\<type\>"} \[Rule] \
{"\<http://www.w3.org/2001/XMLSchema\>", "\<string\>"}}} -> "\<61821\>"}, {}, \
\ \[IndentingNewLine]OperationName -> {"\<urn:xmethods-Temperature\>", \
"\<getTemp\>"}]\)], "Input",
  CellLabel->"In[30]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"soapenv", 
          "Envelope"}, {{"xmlns", "soapenv"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", {"xmlns", 
              "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema", {"xmlns", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", {"xmlns", 
              "soapenc"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/encoding/"}, \
{XMLElement[{"soapenv", 
              "Body"}, {}, {XMLElement[{"ns0", 
                  "getTemp"}, {{"xmlns", "ns0"} \[Rule] 
                    "urn:xmethods-Temperature"}, {XMLElement[{\*"\<\"\"\>", 
                      "zipcode"}, {{"xsi", "type"} \[Rule] 
                        "xsd:string"}, {"61821"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[30]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(response\  = \ 
      InvokeServiceOperation["\<http://services.xmethods.net:80/soap/servlet/\
rpcrouter\>", \ msg]\)], "Input",
  CellLabel->"In[31]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(\(XMLObject["Document"]\)[{}, 
      XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
          "Envelope"}, {{"http://www.w3.org/2000/xmlns/", "SOAP-ENV"} \[Rule] 
            "http://schemas.xmlsoap.org/soap/envelope/", \
{"http://www.w3.org/2000/xmlns/", "xsi"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema-instance", \
{"http://www.w3.org/2000/xmlns/", "xsd"} \[Rule] 
            "http://www.w3.org/2001/XMLSchema"}, \
{XMLElement[{"http://schemas.xmlsoap.org/soap/envelope/", 
              "Body"}, {}, {XMLElement[{"urn:xmethods-Temperature", 
                  "getTempResponse"}, \
{{"http://schemas.xmlsoap.org/soap/envelope/", "encodingStyle"} \[Rule] 
                    "http://schemas.xmlsoap.org/soap/encoding/", \
{"http://www.w3.org/2000/xmlns/", "ns1"} \[Rule] 
                    "urn:xmethods-Temperature"}, {XMLElement[
                    "return", {{"http://www.w3.org/2001/XMLSchema-instance", 
                          "type"} \[Rule] 
                        "xsd:float"}, {"43.0"}]}]}]}], {}]\)], "Output",
  CellLabel->"Out[31]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[TextData[{
  "Once a web service function is called using InvokeServiceOperation the \
results come back in the form of a SOAPEnvelope.  FromServiceResponse can be \
used to extract the results from the SOAPEnvelope.  If the message uses SOAP \
encoding and provides types in the response message, the data will be \
deserialized into the proper ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " types.  If not, the SOAPEnvelope wrappers will be stripped and the \
response will be deserialized into a Rule syntax expression with String \
values."
}], "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FromServiceResponse[response]\)], "Input",
  CellLabel->"In[32]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(43.`\)], "Output",
  CellLabel->"Out[32]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[TextData[{
  "If a web service operation is used a lot, the operation can be installed \
more permanently using InstallServiceOperation.  InstallServiceOperation \
installs the operation as a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " function.  The first argument specifies the function name.  It is a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " symbol.  The second argument is the end point URI.  The third argument is \
a List of parameters and the fourth argument is a List of headers.  These \
last two arguments work similarly to BuildSOAPEnvelope.  The major difference \
between InstallServiceOperation and ToServiceRequest is how the parameters \
are passed.  ToServiceRequest requires the values to be passed in the \
parameters.  However, for InstallServiceOperation, the parameter values will \
be supplied later when the function is called  Therefor only the left side of \
the a parameter in Rule syntax needs to be passed to InstallServiceOperation. \
 Passing the left side specifies to the function what the parameter is.  \
However, it also allows the user to omit this information when invoking the \
function.  An example is provided below.  It specifies a method named getTemp \
which requires a parameter called zipcode. "
}], "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallServiceOperation[
      TempService`getTemperature, \ \
"\<http://services.xmethods.net:80/soap/servlet/rpcrouter\>", \ \
\[IndentingNewLine]{"\<zipcode\>"}, {}, 
      OperationName -> {"\<urn:xmethods-Temperature\>", "\<getTemp\>"}]\)], \
"Input",
  CellLabel->"In[33]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(getTemperature\)], "Output",
  CellLabel->"Out[33]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(getTemperature["\<61821\>"]\)], "Input",
  CellLabel->"In[34]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(43.`\)], "Output",
  CellLabel->"Out[34]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell["\<\
If the user wants to specify more information than just the name, that can \
accomplished in a similar way.  The user just needs to use the left side of \
the Rule syntax.  An example is provided below.\
\>", "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallServiceOperation[
      TempService`getTempFromZipCode, \ \
"\<http://services.xmethods.net:80/soap/servlet/rpcrouter\>", \ \
\[IndentingNewLine]{{"\<zipcode\>", \
{{"\<http://www.w3.org/2001/XMLSchema-instance\>", \ "\<type\>"} \[Rule] \
{"\<http://www.w3.org/2001/XMLSchema\>", "\<string\>"}}}}, {}, 
      OperationName -> {"\<urn:xmethods-Temperature\>", "\<getTemp\>"}]\)], \
"Input",
  CellLabel->"In[35]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(getTempFromZipCode\)], "Output",
  CellLabel->"Out[35]=",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(getTempFromZipCode["\<61821\>"]\)], "Input",
  CellLabel->"In[36]:=",
  CellTags->"b:0.3.3"],

Cell[BoxData[
    \(43.`\)], "Output",
  CellLabel->"Out[36]=",
  CellTags->"b:0.3.3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Uninstalling a Web Service", "Subsection",
  CellTags->"b:0.3.4"],

Cell[TextData[{
  "UninstallServiceOperation and UninstallService can be used to remove a web \
service from ",
  StyleBox["Mathematica.  ",
    FontSlant->"Italic"],
  "Uninstalling a service amounts to removing the operation palette and the \
",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " function defining the operation.  UninstallServiceOperation takes a \
function as parameter.  If the function, defines a web service operation, the \
operation is uninstalled.  UninstallService takes a context pattern as its \
only parameter.  It then uses UninstallServiceOperation to uninstall any \
function that is a returned from the context pattern and is an installed \
service."
}], "Text",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \(Needs["\<WebServices`\>"]\)], "Input",
  CellLabel->"In[37]:=",
  CellTags->"b:0.3.4"],

Cell[CellGroupData[{

Cell[BoxData[
    \(InstallService["\<http://terraservice.net/TerraService.asmx?WSDL \
\>"]\)], "Input",
  CellLabel->"In[38]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \({ConvertLonLatPtToNearestPlace, ConvertLonLatPtToUtmPt, 
      ConvertUtmPtToLonLatPt, ConvertPlaceToLonLatPt, CountPlacesInRect, 
      GetAreaFromPt, GetAreaFromRect, GetAreaFromTileId, GetLatLonMetrics, 
      GetPlaceFacts, GetPlaceList, GetPlaceListInRect, GetTheme, 
      GetTileMetaFromLonLatPt, GetTileMetaFromTileId, GetTile}\)], "Output",
  CellLabel->"Out[38]=",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \($InstalledServices\)], "Input",
  CellLabel->"In[39]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \({ActorSearchRequest, ArtistSearchRequest, AsinSearchRequest, 
      AuthorSearchRequest, BabelFish, BrowseNodeSearchRequest, 
      ConvertLonLatPtToNearestPlace, ConvertLonLatPtToUtmPt, 
      ConvertPlaceToLonLatPt, ConvertUtmPtToLonLatPt, CountPlacesInRect, 
      DirectorSearchRequest, GetAreaFromPt, GetAreaFromRect, 
      GetAreaFromTileId, GetLatLonMetrics, GetPlaceFacts, GetPlaceList, 
      GetPlaceListInRect, getPrice, 
      CurrencyExchangeService`CurrencyExchangePort`getRate, getTemperature, 
      getTempFromZipCode, GetTheme, GetTile, GetTileMetaFromLonLatPt, 
      GetTileMetaFromTileId, getTraffic, KeywordSearchRequest, 
      ListManiaSearchRequest, ManufacturerSearchRequest, 
      SimilaritySearchRequest, UpcSearchRequest}\)], "Output",
  CellLabel->"Out[39]=",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[TextData[{
  "Once a service is installed a user can use $InstalledServices to see which \
services are installed.  The example below demonstrates \
UninstallServiceOperation.  $InstalledServices demonstrates that is removed \
from ",
  StyleBox["Mathematica.",
    FontSlant->"Italic"]
}], "Text",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \(UninstallServiceOperation[ConvertLonLatPtToNearestPlace]\)], "Input",
  CellLabel->"In[40]:=",
  CellTags->"b:0.3.4"],

Cell[CellGroupData[{

Cell[BoxData[
    \($InstalledServices\)], "Input",
  CellLabel->"In[41]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \({ActorSearchRequest, ArtistSearchRequest, AsinSearchRequest, 
      AuthorSearchRequest, BabelFish, BrowseNodeSearchRequest, 
      ConvertLonLatPtToUtmPt, ConvertPlaceToLonLatPt, ConvertUtmPtToLonLatPt, 
      CountPlacesInRect, DirectorSearchRequest, GetAreaFromPt, 
      GetAreaFromRect, GetAreaFromTileId, GetLatLonMetrics, GetPlaceFacts, 
      GetPlaceList, GetPlaceListInRect, getPrice, 
      CurrencyExchangeService`CurrencyExchangePort`getRate, getTemperature, 
      getTempFromZipCode, GetTheme, GetTile, GetTileMetaFromLonLatPt, 
      GetTileMetaFromTileId, getTraffic, KeywordSearchRequest, 
      ListManiaSearchRequest, ManufacturerSearchRequest, 
      SimilaritySearchRequest, UpcSearchRequest}\)], "Output",
  CellLabel->"Out[41]=",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(OperationPalette[ConvertLonLatPtToNearestPlace]\)], "Input",
  CellLabel->"In[42]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \("Palette cannot be found."\)], "Output",
  CellLabel->"Out[42]=",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Context[ConvertLonLatPtToNearestPlace]\)], "Input",
  CellLabel->"In[43]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \("Global`"\)], "Output",
  CellLabel->"Out[43]=",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[TextData[{
  "The example below demonstrates UninstallService.  $InstalledServices \
demonstrates that all the symbols in the context are removed from ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \(UninstallService["\<TerraService`TerraServiceSoap`*\>"]\)], "Input",
  CellLabel->"In[44]:=",
  CellTags->"b:0.3.4"],

Cell[CellGroupData[{

Cell[BoxData[
    \($InstalledServices\)], "Input",
  CellLabel->"In[45]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \({ActorSearchRequest, ArtistSearchRequest, AsinSearchRequest, 
      AuthorSearchRequest, BabelFish, BrowseNodeSearchRequest, 
      DirectorSearchRequest, getPrice, 
      CurrencyExchangeService`CurrencyExchangePort`getRate, getTemperature, 
      getTempFromZipCode, getTraffic, KeywordSearchRequest, 
      ListManiaSearchRequest, ManufacturerSearchRequest, 
      SimilaritySearchRequest, UpcSearchRequest}\)], "Output",
  CellLabel->"Out[45]=",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(OperationPalette[ConvertPlaceToLonLatPt]\)], "Input",
  CellLabel->"In[46]:=",
  CellTags->"b:0.3.4"],

Cell[BoxData[
    \("Palette cannot be found."\)], "Output",
  CellLabel->"Out[46]=",
  CellTags->"b:0.3.4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Debugging", "Subsection",
  CellTags->"b:0.3.5"],

Cell["\<\
While using a package, it is useful to include tools that help a user debug \
problems experienced using the package.  The Web Services package provides a \
few symbols that should help debugging.  The debugging symbols are defined in \
the table below.\
\>", "Text",
  CellTags->"b:0.3.5"],

Cell[BoxData[GridBox[{
        {"Symbol", "Description"},
        {"$PrintServiceRequest", \($PrintServiceRequest\ uses\ the\ \
Mathematica\ Print\ function\[IndentingNewLine]
          
          to\ print\ the\ message\ in\ XML\ format\ sent\ to\ a\ web\ \
\(\(service\)\(.\)\)\)},
        {"$PrintServiceResponse", \($PrintServiceResponse\ uses\ the\ \
Mathematica\ Print\ function\[IndentingNewLine]
          
          to\ print\ the\ message\ in\ XML\ format\ received\ from\ a\ web\ \
\[IndentingNewLine]
          
          service\ before\ it\ is\ deserialized\ into\ a\ Rule\ syntax\ \
\[IndentingNewLine]
          \(\(expression\)\(.\)\)\)},
        {"$PrintPerformanceNumbers", \($PrintPerformanceNumbers\ prints\ \
timing\ information\ that\ \[IndentingNewLine]
          
          can\ be\ used\ to\ measure\ the\ performance\ of\ a\ web\ service\ \
\[IndentingNewLine]
          \(\(call\)\(.\)\)\)},
        {"$PrintWSDLDebug", \($PrintWSDLDebug\ prints\ information\ about\ \
web\ services\ \[IndentingNewLine]
          
          being\ \ \ installed\ that\ can\ be\ used\ to\ debug . \ \ \
Information\ \[IndentingNewLine]
          printed\ includes\ option\ values, \ 
          parameter\ signatures\ \[IndentingNewLine]
          
          and\ the\ end\ point\ URI\ for\ \ web\ service\ \
\(\(operation\)\(.\)\)\)},
        {"$PrintShortErrorMessages", \($PrintShortErrorMessages\ is\ an\ \
option\ which\ specifies\ \[IndentingNewLine]
          
          whether\ error\ messages\ will\ be\ shortened\ for\ the\ user\ to\ \
\[IndentingNewLine]
          
          avoid\ long\ intimidating\ error\ messages . \ \ The\ default\ is\ \
\(\(True\)\(.\)\)\)}
        },
      GridFrame->True,
      RowLines->True,
      ColumnLines->True]], "Text",
  CellTags->"b:0.3.5"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.2 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 951}},
ScreenStyleEnvironment->"Working",
WindowSize->{644, 696},
WindowMargins->{{35, Automatic}, {Automatic, 17}},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:0.0"->{
    Cell[1829, 57, 52, 1, 56, "Section",
      CellTags->"b:0.0"],
    Cell[1884, 60, 788, 15, 93, "Text",
      CellTags->"b:0.0"],
    Cell[2697, 79, 56, 1, 35, "Subsubsection",
      CellTags->"b:0.0"],
    Cell[2756, 82, 657, 18, 76, "Text",
      CellTags->"b:0.0"],
    Cell[3416, 102, 340, 6, 59, "Text",
      CellTags->"b:0.0"],
    Cell[3793, 113, 58, 1, 35, "Subsubsection",
      CellTags->"b:0.0"],
    Cell[3854, 116, 434, 8, 76, "Text",
      CellTags->"b:0.0"],
    Cell[4291, 126, 224, 5, 42, "Text",
      CellTags->"b:0.0"],
    Cell[4552, 136, 50, 1, 35, "Subsubsection",
      CellTags->"b:0.0"],
    Cell[4605, 139, 558, 13, 76, "Text",
      CellTags->"b:0.0"],
    Cell[5200, 157, 50, 1, 35, "Subsubsection",
      CellTags->"b:0.0"],
    Cell[5253, 160, 649, 14, 76, "Text",
      CellTags->"b:0.0"],
    Cell[5939, 179, 57, 1, 35, "Subsubsection",
      CellTags->"b:0.0"],
    Cell[5999, 182, 556, 11, 76, "Text",
      CellTags->"b:0.0"],
    Cell[6558, 195, 377, 7, 59, "Text",
      CellTags->"b:0.0"],
    Cell[6972, 207, 66, 1, 35, "Subsubsection",
      CellTags->"b:0.0"],
    Cell[7041, 210, 251, 5, 42, "Text",
      CellTags->"b:0.0"]},
  "b:0.1.0"->{
    Cell[7400, 225, 55, 1, 39, "Subsection",
      CellTags->"b:0.1.0"],
    Cell[7458, 228, 757, 13, 110, "Text",
      CellTags->"b:0.1.0"],
    Cell[8218, 243, 105, 3, 29, "Input",
      CellTags->"b:0.1.0"],
    Cell[8348, 250, 152, 4, 29, "Input",
      CellTags->"b:0.1.0"],
    Cell[8503, 256, 92, 3, 23, "Output",
      CellTags->"b:0.1.0"],
    Cell[8610, 262, 347, 6, 59, "Text",
      CellTags->"b:0.1.0"],
    Cell[8960, 270, 107, 3, 29, "Input",
      CellTags->"b:0.1.0"],
    Cell[9070, 275, 739, 14, 93, "Text",
      CellTags->"b:0.1.0"],
    Cell[9812, 291, 346, 8, 42, "Text",
      CellTags->"b:0.1.0"],
    Cell[10183, 303, 126, 3, 29, "Input",
      CellTags->"b:0.1.0"],
    Cell[10312, 308, 101, 3, 23, "Output",
      CellTags->"b:0.1.0"],
    Cell[10428, 314, 282, 8, 25, "Text",
      CellTags->"b:0.1.0"]},
  "b:0.1.1"->{
    Cell[10747, 327, 59, 1, 39, "Subsection",
      CellTags->"b:0.1.1"],
    Cell[10809, 330, 857, 17, 110, "Text",
      CellTags->"b:0.1.1"],
    Cell[11669, 349, 138, 4, 25, "Text",
      CellTags->"b:0.1.1"],
    Cell[11810, 355, 105, 3, 29, "Input",
      CellTags->"b:0.1.1"],
    Cell[11940, 362, 157, 4, 29, "Input",
      CellTags->"b:0.1.1"],
    Cell[12100, 368, 347, 6, 80, "Output",
      CellTags->"b:0.1.1"],
    Cell[12462, 377, 342, 8, 42, "Text",
      CellTags->"b:0.1.1"],
    Cell[12807, 387, 117, 3, 29, "Input",
      CellTags->"b:0.1.1"],
    Cell[12927, 392, 999, 19, 127, "Text",
      CellTags->"b:0.1.1"],
    Cell[13951, 415, 337, 6, 69, "Input",
      CellTags->"b:0.1.1"],
    Cell[14291, 423, 9505, 159, 2132, "Output",
      CellTags->"b:0.1.1"],
    Cell[23811, 585, 326, 6, 59, "Text",
      CellTags->"b:0.1.1"],
    Cell[24140, 593, 282, 7, 25, "Text",
      CellTags->"b:0.1.1"]},
  "b:0.2"->{
    Cell[24471, 606, 48, 1, 56, "Section",
      CellTags->"b:0.2"],
    Cell[24522, 609, 242, 6, 67, "Text",
      CellTags->"b:0.2"],
    Cell[24767, 617, 261, 6, 25, "Text",
      CellTags->"b:0.2"],
    Cell[25031, 625, 271, 6, 25, "Text",
      CellTags->"b:0.2"],
    Cell[25305, 633, 263, 6, 25, "Text",
      CellTags->"b:0.2"]},
  "Type Mapping"->{
    Cell[25664, 648, 74, 1, 39, "Subsection",
      CellTags->{"Type Mapping", "b:0.3.0"}]},
  "b:0.3.0"->{
    Cell[25664, 648, 74, 1, 39, "Subsection",
      CellTags->{"Type Mapping", "b:0.3.0"}],
    Cell[25741, 651, 813, 16, 110, "Text",
      CellTags->"b:0.3.0"],
    Cell[26579, 671, 65, 1, 35, "Subsubsection",
      CellTags->"b:0.3.0"],
    Cell[26647, 674, 804, 15, 110, "Text",
      CellTags->"b:0.3.0"],
    Cell[27488, 694, 56, 1, 35, "Subsubsection",
      CellTags->"b:0.3.0"],
    Cell[27547, 697, 619, 10, 93, "Text",
      CellTags->"b:0.3.0"],
    Cell[28169, 709, 961, 16, 127, "Text",
      CellTags->"b:0.3.0"],
    Cell[29133, 727, 339, 6, 59, "Text",
      CellTags->"b:0.3.0"],
    Cell[29475, 735, 610, 11, 76, "Text",
      CellTags->"b:0.3.0"],
    Cell[30088, 748, 394, 9, 59, "Text",
      CellTags->"b:0.3.0"],
    Cell[30485, 759, 532, 15, 59, "InlineFormula",
      CellTags->"b:0.3.0"],
    Cell[31020, 776, 335, 6, 59, "Text",
      CellTags->"b:0.3.0"],
    Cell[31358, 784, 891, 26, 76, "InlineFormula",
      CellTags->"b:0.3.0"],
    Cell[32252, 812, 105, 3, 29, "Input",
      CellTags->"b:0.3.0"],
    Cell[32382, 819, 188, 5, 29, "Input",
      CellTags->"b:0.3.0"],
    Cell[32573, 826, 348, 6, 80, "Output",
      CellTags->"b:0.3.0"],
    Cell[32958, 837, 1593, 24, 449, "Input",
      CellTags->"b:0.3.0"],
    Cell[34554, 863, 28086, 501, 5704, "Output",
      CellTags->"b:0.3.0"],
    Cell[62689, 1370, 60, 1, 35, "Subsubsection",
      CellTags->"b:0.3.0"],
    Cell[62752, 1373, 773, 15, 93, "Text",
      CellTags->"b:0.3.0"],
    Cell[63528, 1390, 686, 13, 93, "Text",
      CellTags->"b:0.3.0"],
    Cell[64217, 1405, 450, 13, 59, "Text",
      CellTags->"b:0.3.0"],
    Cell[64670, 1420, 4259, 84, 693, "Text",
      CellTags->"b:0.3.0"]},
  "b:0.3.1"->{
    Cell[68978, 1510, 60, 1, 39, "Subsection",
      CellTags->"b:0.3.1"],
    Cell[69041, 1513, 463, 10, 59, "Text",
      CellTags->"b:0.3.1"],
    Cell[69507, 1525, 106, 3, 29, "Input",
      CellTags->"b:0.3.1"],
    Cell[69638, 1532, 179, 4, 49, "Input",
      CellTags->"b:0.3.1"],
    Cell[69820, 1538, 92, 3, 23, "Output",
      CellTags->"b:0.3.1"],
    Cell[69949, 1546, 98, 3, 29, "Input",
      CellTags->"b:0.3.1"],
    Cell[70050, 1551, 99, 3, 23, "Output",
      CellTags->"b:0.3.1"],
    Cell[70186, 1559, 107, 3, 29, "Input",
      CellTags->"b:0.3.1"],
    Cell[70296, 1564, 88, 3, 23, "Output",
      CellTags->"b:0.3.1"],
    Cell[70399, 1570, 525, 9, 76, "Text",
      CellTags->"b:0.3.1"],
    Cell[70949, 1583, 200, 5, 69, "Input",
      CellTags->"b:0.3.1"],
    Cell[71152, 1590, 136, 3, 23, "Output",
      CellTags->"b:0.3.1"],
    Cell[71325, 1598, 160, 4, 29, "Input",
      CellTags->"b:0.3.1"],
    Cell[71488, 1604, 89, 3, 23, "Output",
      CellTags->"b:0.3.1"]},
  "b:0.3.2"->{
    Cell[71626, 1613, 66, 1, 39, "Subsection",
      CellTags->"b:0.3.2"],
    Cell[71695, 1616, 552, 9, 76, "Text",
      CellTags->"b:0.3.2"],
    Cell[72250, 1627, 106, 3, 29, "Input",
      CellTags->"b:0.3.2"],
    Cell[72381, 1634, 158, 4, 29, "Input",
      CellTags->"b:0.3.2"],
    Cell[72542, 1640, 94, 3, 23, "Output",
      CellTags->"b:0.3.2"],
    Cell[72673, 1648, 136, 3, 29, "Input",
      CellTags->"b:0.3.2"],
    Cell[72812, 1653, 983, 19, 175, "Output",
      CellTags->"b:0.3.2"],
    Cell[73810, 1675, 723, 11, 110, "Text",
      CellTags->"b:0.3.2"],
    Cell[74558, 1690, 180, 4, 49, "Input",
      CellTags->"b:0.3.2"],
    Cell[74741, 1696, 1649, 27, 422, "Output",
      CellTags->"b:0.3.2"],
    Cell[76427, 1728, 128, 3, 29, "Input",
      CellTags->"b:0.3.2"],
    Cell[76558, 1733, 1649, 27, 422, "Output",
      CellTags->"b:0.3.2"],
    Cell[78222, 1763, 710, 13, 93, "Text",
      CellTags->"b:0.3.2"],
    Cell[78957, 1780, 143, 4, 29, "Input",
      CellTags->"b:0.3.2"],
    Cell[79103, 1786, 1649, 27, 422, "Output",
      CellTags->"b:0.3.2"],
    Cell[80767, 1816, 590, 12, 76, "Text",
      CellTags->"b:0.3.2"],
    Cell[81382, 1832, 110, 3, 29, "Input",
      CellTags->"b:0.3.2"],
    Cell[81495, 1837, 613, 10, 156, "Output",
      CellTags->"b:0.3.2"]},
  "b:0.3.3"->{
    Cell[82157, 1853, 67, 1, 39, "Subsection",
      CellTags->"b:0.3.3"],
    Cell[82227, 1856, 589, 12, 76, "Text",
      CellTags->"b:0.3.3"],
    Cell[82819, 1870, 991, 15, 144, "Text",
      CellTags->"b:0.3.3"],
    Cell[83813, 1887, 4554, 134, 685, "Text",
      CellTags->"b:0.3.3"],
    Cell[88370, 2023, 106, 3, 29, "Input",
      CellTags->"b:0.3.3"],
    Cell[88501, 2030, 239, 6, 69, "Input",
      CellTags->"b:0.3.3"],
    Cell[88743, 2038, 797, 16, 156, "Output",
      CellTags->"b:0.3.3"],
    Cell[89555, 2057, 295, 6, 42, "Text",
      CellTags->"b:0.3.3"],
    Cell[89875, 2067, 174, 4, 49, "Input",
      CellTags->"b:0.3.3"],
    Cell[90052, 2073, 1124, 20, 270, "Output",
      CellTags->"b:0.3.3"],
    Cell[91191, 2096, 561, 11, 76, "Text",
      CellTags->"b:0.3.3"],
    Cell[91755, 2109, 639, 16, 97, "Text",
      CellTags->"b:0.3.3"],
    Cell[92419, 2129, 367, 8, 109, "Input",
      CellTags->"b:0.3.3"],
    Cell[92789, 2139, 984, 19, 175, "Output",
      CellTags->"b:0.3.3"],
    Cell[93810, 2163, 174, 4, 49, "Input",
      CellTags->"b:0.3.3"],
    Cell[93987, 2169, 1124, 20, 270, "Output",
      CellTags->"b:0.3.3"],
    Cell[95126, 2192, 749, 15, 93, "Text",
      CellTags->"b:0.3.3"],
    Cell[95900, 2211, 377, 8, 89, "Input",
      CellTags->"b:0.3.3"],
    Cell[96280, 2221, 858, 17, 156, "Output",
      CellTags->"b:0.3.3"],
    Cell[97175, 2243, 196, 5, 49, "Input",
      CellTags->"b:0.3.3"],
    Cell[97374, 2250, 1124, 20, 270, "Output",
      CellTags->"b:0.3.3"],
    Cell[98513, 2273, 596, 12, 76, "Text",
      CellTags->"b:0.3.3"],
    Cell[99134, 2289, 110, 3, 29, "Input",
      CellTags->"b:0.3.3"],
    Cell[99247, 2294, 86, 3, 23, "Output",
      CellTags->"b:0.3.3"],
    Cell[99348, 2300, 1315, 22, 178, "Text",
      CellTags->"b:0.3.3"],
    Cell[100688, 2326, 324, 8, 69, "Input",
      CellTags->"b:0.3.3"],
    Cell[101015, 2336, 96, 3, 23, "Output",
      CellTags->"b:0.3.3"],
    Cell[101148, 2344, 108, 3, 29, "Input",
      CellTags->"b:0.3.3"],
    Cell[101259, 2349, 86, 3, 23, "Output",
      CellTags->"b:0.3.3"],
    Cell[101360, 2355, 249, 5, 42, "Text",
      CellTags->"b:0.3.3"],
    Cell[101634, 2364, 464, 10, 109, "Input",
      CellTags->"b:0.3.3"],
    Cell[102101, 2376, 100, 3, 23, "Output",
      CellTags->"b:0.3.3"],
    Cell[102238, 2384, 112, 3, 29, "Input",
      CellTags->"b:0.3.3"],
    Cell[102353, 2389, 86, 3, 23, "Output",
      CellTags->"b:0.3.3"]},
  "b:0.3.4"->{
    Cell[102488, 2398, 71, 1, 39, "Subsection",
      CellTags->"b:0.3.4"],
    Cell[102562, 2401, 725, 16, 93, "Text",
      CellTags->"b:0.3.4"],
    Cell[103290, 2419, 106, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[103421, 2426, 152, 4, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[103576, 2432, 417, 7, 99, "Output",
      CellTags->"b:0.3.4"],
    Cell[104030, 2444, 99, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[104132, 2449, 834, 14, 213, "Output",
      CellTags->"b:0.3.4"],
    Cell[104981, 2466, 326, 8, 42, "Text",
      CellTags->"b:0.3.4"],
    Cell[105310, 2476, 137, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[105472, 2483, 99, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[105574, 2488, 796, 13, 213, "Output",
      CellTags->"b:0.3.4"],
    Cell[106407, 2506, 128, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[106538, 2511, 108, 3, 23, "Output",
      CellTags->"b:0.3.4"],
    Cell[106683, 2519, 119, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[106805, 2524, 91, 3, 23, "Output",
      CellTags->"b:0.3.4"],
    Cell[106911, 2530, 251, 7, 42, "Text",
      CellTags->"b:0.3.4"],
    Cell[107165, 2539, 136, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[107326, 2546, 99, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[107428, 2551, 492, 9, 118, "Output",
      CellTags->"b:0.3.4"],
    Cell[107957, 2565, 121, 3, 29, "Input",
      CellTags->"b:0.3.4"],
    Cell[108081, 2570, 108, 3, 23, "Output",
      CellTags->"b:0.3.4"]},
  "b:0.3.5"->{
    Cell[108238, 2579, 54, 1, 39, "Subsection",
      CellTags->"b:0.3.5"],
    Cell[108295, 2582, 300, 6, 42, "Text",
      CellTags->"b:0.3.5"],
    Cell[108598, 2590, 1801, 44, 347, "Text",
      CellTags->"b:0.3.5"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:0.0", 111103, 2656},
  {"b:0.1.0", 112292, 2691},
  {"b:0.1.1", 113142, 2716},
  {"b:0.2", 114075, 2743},
  {"Type Mapping", 114433, 2754},
  {"b:0.3.0", 114541, 2757},
  {"b:0.3.1", 116246, 2804},
  {"b:0.3.2", 117259, 2833},
  {"b:0.3.3", 118575, 2870},
  {"b:0.3.4", 121047, 2939},
  {"b:0.3.5", 122585, 2982}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 28, 0, 105, "Title"],

Cell[CellGroupData[{
Cell[1829, 57, 52, 1, 56, "Section",
  CellTags->"b:0.0"],
Cell[1884, 60, 788, 15, 93, "Text",
  CellTags->"b:0.0"],

Cell[CellGroupData[{
Cell[2697, 79, 56, 1, 35, "Subsubsection",
  CellTags->"b:0.0"],
Cell[2756, 82, 657, 18, 76, "Text",
  CellTags->"b:0.0"],
Cell[3416, 102, 340, 6, 59, "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3793, 113, 58, 1, 35, "Subsubsection",
  CellTags->"b:0.0"],
Cell[3854, 116, 434, 8, 76, "Text",
  CellTags->"b:0.0"],
Cell[4291, 126, 224, 5, 42, "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4552, 136, 50, 1, 35, "Subsubsection",
  CellTags->"b:0.0"],
Cell[4605, 139, 558, 13, 76, "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5200, 157, 50, 1, 35, "Subsubsection",
  CellTags->"b:0.0"],
Cell[5253, 160, 649, 14, 76, "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5939, 179, 57, 1, 35, "Subsubsection",
  CellTags->"b:0.0"],
Cell[5999, 182, 556, 11, 76, "Text",
  CellTags->"b:0.0"],
Cell[6558, 195, 377, 7, 59, "Text",
  CellTags->"b:0.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6972, 207, 66, 1, 35, "Subsubsection",
  CellTags->"b:0.0"],
Cell[7041, 210, 251, 5, 42, "Text",
  CellTags->"b:0.0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[7341, 221, 34, 0, 56, "Section"],

Cell[CellGroupData[{
Cell[7400, 225, 55, 1, 39, "Subsection",
  CellTags->"b:0.1.0"],
Cell[7458, 228, 757, 13, 110, "Text",
  CellTags->"b:0.1.0"],
Cell[8218, 243, 105, 3, 29, "Input",
  CellTags->"b:0.1.0"],

Cell[CellGroupData[{
Cell[8348, 250, 152, 4, 29, "Input",
  CellTags->"b:0.1.0"],
Cell[8503, 256, 92, 3, 23, "Output",
  CellTags->"b:0.1.0"]
}, Open  ]],
Cell[8610, 262, 347, 6, 59, "Text",
  CellTags->"b:0.1.0"],
Cell[8960, 270, 107, 3, 29, "Input",
  CellTags->"b:0.1.0"],
Cell[9070, 275, 739, 14, 93, "Text",
  CellTags->"b:0.1.0"],
Cell[9812, 291, 346, 8, 42, "Text",
  CellTags->"b:0.1.0"],

Cell[CellGroupData[{
Cell[10183, 303, 126, 3, 29, "Input",
  CellTags->"b:0.1.0"],
Cell[10312, 308, 101, 3, 23, "Output",
  CellTags->"b:0.1.0"]
}, Open  ]],
Cell[10428, 314, 282, 8, 25, "Text",
  CellTags->"b:0.1.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10747, 327, 59, 1, 39, "Subsection",
  CellTags->"b:0.1.1"],
Cell[10809, 330, 857, 17, 110, "Text",
  CellTags->"b:0.1.1"],
Cell[11669, 349, 138, 4, 25, "Text",
  CellTags->"b:0.1.1"],
Cell[11810, 355, 105, 3, 29, "Input",
  CellTags->"b:0.1.1"],

Cell[CellGroupData[{
Cell[11940, 362, 157, 4, 29, "Input",
  CellTags->"b:0.1.1"],
Cell[12100, 368, 347, 6, 80, "Output",
  CellTags->"b:0.1.1"]
}, Open  ]],
Cell[12462, 377, 342, 8, 42, "Text",
  CellTags->"b:0.1.1"],
Cell[12807, 387, 117, 3, 29, "Input",
  CellTags->"b:0.1.1"],
Cell[12927, 392, 999, 19, 127, "Text",
  CellTags->"b:0.1.1"],

Cell[CellGroupData[{
Cell[13951, 415, 337, 6, 69, "Input",
  CellTags->"b:0.1.1"],
Cell[14291, 423, 9505, 159, 2132, "Output",
  CellTags->"b:0.1.1"]
}, Open  ]],
Cell[23811, 585, 326, 6, 59, "Text",
  CellTags->"b:0.1.1"],
Cell[24140, 593, 282, 7, 25, "Text",
  CellTags->"b:0.1.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[24471, 606, 48, 1, 56, "Section",
  CellTags->"b:0.2"],
Cell[24522, 609, 242, 6, 67, "Text",
  CellTags->"b:0.2"],
Cell[24767, 617, 261, 6, 25, "Text",
  CellTags->"b:0.2"],
Cell[25031, 625, 271, 6, 25, "Text",
  CellTags->"b:0.2"],
Cell[25305, 633, 263, 6, 25, "Text",
  CellTags->"b:0.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25605, 644, 34, 0, 56, "Section"],

Cell[CellGroupData[{
Cell[25664, 648, 74, 1, 39, "Subsection",
  CellTags->{"Type Mapping", "b:0.3.0"}],
Cell[25741, 651, 813, 16, 110, "Text",
  CellTags->"b:0.3.0"],

Cell[CellGroupData[{
Cell[26579, 671, 65, 1, 35, "Subsubsection",
  CellTags->"b:0.3.0"],
Cell[26647, 674, 804, 15, 110, "Text",
  CellTags->"b:0.3.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27488, 694, 56, 1, 35, "Subsubsection",
  CellTags->"b:0.3.0"],
Cell[27547, 697, 619, 10, 93, "Text",
  CellTags->"b:0.3.0"],
Cell[28169, 709, 961, 16, 127, "Text",
  CellTags->"b:0.3.0"],
Cell[29133, 727, 339, 6, 59, "Text",
  CellTags->"b:0.3.0"],
Cell[29475, 735, 610, 11, 76, "Text",
  CellTags->"b:0.3.0"],
Cell[30088, 748, 394, 9, 59, "Text",
  CellTags->"b:0.3.0"],
Cell[30485, 759, 532, 15, 59, "InlineFormula",
  CellTags->"b:0.3.0"],
Cell[31020, 776, 335, 6, 59, "Text",
  CellTags->"b:0.3.0"],
Cell[31358, 784, 891, 26, 76, "InlineFormula",
  CellTags->"b:0.3.0"],
Cell[32252, 812, 105, 3, 29, "Input",
  CellTags->"b:0.3.0"],

Cell[CellGroupData[{
Cell[32382, 819, 188, 5, 29, "Input",
  CellTags->"b:0.3.0"],
Cell[32573, 826, 348, 6, 80, "Output",
  CellTags->"b:0.3.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[32958, 837, 1593, 24, 449, "Input",
  CellTags->"b:0.3.0"],
Cell[34554, 863, 28086, 501, 5704, "Output",
  CellTags->"b:0.3.0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[62689, 1370, 60, 1, 35, "Subsubsection",
  CellTags->"b:0.3.0"],
Cell[62752, 1373, 773, 15, 93, "Text",
  CellTags->"b:0.3.0"],
Cell[63528, 1390, 686, 13, 93, "Text",
  CellTags->"b:0.3.0"],
Cell[64217, 1405, 450, 13, 59, "Text",
  CellTags->"b:0.3.0"],
Cell[64670, 1420, 4259, 84, 693, "Text",
  CellTags->"b:0.3.0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[68978, 1510, 60, 1, 39, "Subsection",
  CellTags->"b:0.3.1"],
Cell[69041, 1513, 463, 10, 59, "Text",
  CellTags->"b:0.3.1"],
Cell[69507, 1525, 106, 3, 29, "Input",
  CellTags->"b:0.3.1"],

Cell[CellGroupData[{
Cell[69638, 1532, 179, 4, 49, "Input",
  CellTags->"b:0.3.1"],
Cell[69820, 1538, 92, 3, 23, "Output",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[69949, 1546, 98, 3, 29, "Input",
  CellTags->"b:0.3.1"],
Cell[70050, 1551, 99, 3, 23, "Output",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[70186, 1559, 107, 3, 29, "Input",
  CellTags->"b:0.3.1"],
Cell[70296, 1564, 88, 3, 23, "Output",
  CellTags->"b:0.3.1"]
}, Open  ]],
Cell[70399, 1570, 525, 9, 76, "Text",
  CellTags->"b:0.3.1"],

Cell[CellGroupData[{
Cell[70949, 1583, 200, 5, 69, "Input",
  CellTags->"b:0.3.1"],
Cell[71152, 1590, 136, 3, 23, "Output",
  CellTags->"b:0.3.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[71325, 1598, 160, 4, 29, "Input",
  CellTags->"b:0.3.1"],
Cell[71488, 1604, 89, 3, 23, "Output",
  CellTags->"b:0.3.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[71626, 1613, 66, 1, 39, "Subsection",
  CellTags->"b:0.3.2"],
Cell[71695, 1616, 552, 9, 76, "Text",
  CellTags->"b:0.3.2"],
Cell[72250, 1627, 106, 3, 29, "Input",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{
Cell[72381, 1634, 158, 4, 29, "Input",
  CellTags->"b:0.3.2"],
Cell[72542, 1640, 94, 3, 23, "Output",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[72673, 1648, 136, 3, 29, "Input",
  CellTags->"b:0.3.2"],
Cell[72812, 1653, 983, 19, 175, "Output",
  CellTags->"b:0.3.2"]
}, Open  ]],
Cell[73810, 1675, 723, 11, 110, "Text",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{
Cell[74558, 1690, 180, 4, 49, "Input",
  CellTags->"b:0.3.2"],
Cell[74741, 1696, 1649, 27, 422, "Output",
  CellTags->"b:0.3.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[76427, 1728, 128, 3, 29, "Input",
  CellTags->"b:0.3.2"],
Cell[76558, 1733, 1649, 27, 422, "Output",
  CellTags->"b:0.3.2"]
}, Open  ]],
Cell[78222, 1763, 710, 13, 93, "Text",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{
Cell[78957, 1780, 143, 4, 29, "Input",
  CellTags->"b:0.3.2"],
Cell[79103, 1786, 1649, 27, 422, "Output",
  CellTags->"b:0.3.2"]
}, Open  ]],
Cell[80767, 1816, 590, 12, 76, "Text",
  CellTags->"b:0.3.2"],

Cell[CellGroupData[{
Cell[81382, 1832, 110, 3, 29, "Input",
  CellTags->"b:0.3.2"],
Cell[81495, 1837, 613, 10, 156, "Output",
  CellTags->"b:0.3.2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[82157, 1853, 67, 1, 39, "Subsection",
  CellTags->"b:0.3.3"],
Cell[82227, 1856, 589, 12, 76, "Text",
  CellTags->"b:0.3.3"],
Cell[82819, 1870, 991, 15, 144, "Text",
  CellTags->"b:0.3.3"],
Cell[83813, 1887, 4554, 134, 685, "Text",
  CellTags->"b:0.3.3"],
Cell[88370, 2023, 106, 3, 29, "Input",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[88501, 2030, 239, 6, 69, "Input",
  CellTags->"b:0.3.3"],
Cell[88743, 2038, 797, 16, 156, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],
Cell[89555, 2057, 295, 6, 42, "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[89875, 2067, 174, 4, 49, "Input",
  CellTags->"b:0.3.3"],
Cell[90052, 2073, 1124, 20, 270, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],
Cell[91191, 2096, 561, 11, 76, "Text",
  CellTags->"b:0.3.3"],
Cell[91755, 2109, 639, 16, 97, "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[92419, 2129, 367, 8, 109, "Input",
  CellTags->"b:0.3.3"],
Cell[92789, 2139, 984, 19, 175, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[93810, 2163, 174, 4, 49, "Input",
  CellTags->"b:0.3.3"],
Cell[93987, 2169, 1124, 20, 270, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],
Cell[95126, 2192, 749, 15, 93, "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[95900, 2211, 377, 8, 89, "Input",
  CellTags->"b:0.3.3"],
Cell[96280, 2221, 858, 17, 156, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[97175, 2243, 196, 5, 49, "Input",
  CellTags->"b:0.3.3"],
Cell[97374, 2250, 1124, 20, 270, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],
Cell[98513, 2273, 596, 12, 76, "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[99134, 2289, 110, 3, 29, "Input",
  CellTags->"b:0.3.3"],
Cell[99247, 2294, 86, 3, 23, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],
Cell[99348, 2300, 1315, 22, 178, "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[100688, 2326, 324, 8, 69, "Input",
  CellTags->"b:0.3.3"],
Cell[101015, 2336, 96, 3, 23, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[101148, 2344, 108, 3, 29, "Input",
  CellTags->"b:0.3.3"],
Cell[101259, 2349, 86, 3, 23, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],
Cell[101360, 2355, 249, 5, 42, "Text",
  CellTags->"b:0.3.3"],

Cell[CellGroupData[{
Cell[101634, 2364, 464, 10, 109, "Input",
  CellTags->"b:0.3.3"],
Cell[102101, 2376, 100, 3, 23, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[102238, 2384, 112, 3, 29, "Input",
  CellTags->"b:0.3.3"],
Cell[102353, 2389, 86, 3, 23, "Output",
  CellTags->"b:0.3.3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[102488, 2398, 71, 1, 39, "Subsection",
  CellTags->"b:0.3.4"],
Cell[102562, 2401, 725, 16, 93, "Text",
  CellTags->"b:0.3.4"],
Cell[103290, 2419, 106, 3, 29, "Input",
  CellTags->"b:0.3.4"],

Cell[CellGroupData[{
Cell[103421, 2426, 152, 4, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[103576, 2432, 417, 7, 99, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{
Cell[104030, 2444, 99, 3, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[104132, 2449, 834, 14, 213, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]],
Cell[104981, 2466, 326, 8, 42, "Text",
  CellTags->"b:0.3.4"],
Cell[105310, 2476, 137, 3, 29, "Input",
  CellTags->"b:0.3.4"],

Cell[CellGroupData[{
Cell[105472, 2483, 99, 3, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[105574, 2488, 796, 13, 213, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{
Cell[106407, 2506, 128, 3, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[106538, 2511, 108, 3, 23, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{
Cell[106683, 2519, 119, 3, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[106805, 2524, 91, 3, 23, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]],
Cell[106911, 2530, 251, 7, 42, "Text",
  CellTags->"b:0.3.4"],
Cell[107165, 2539, 136, 3, 29, "Input",
  CellTags->"b:0.3.4"],

Cell[CellGroupData[{
Cell[107326, 2546, 99, 3, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[107428, 2551, 492, 9, 118, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]],

Cell[CellGroupData[{
Cell[107957, 2565, 121, 3, 29, "Input",
  CellTags->"b:0.3.4"],
Cell[108081, 2570, 108, 3, 23, "Output",
  CellTags->"b:0.3.4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[108238, 2579, 54, 1, 39, "Subsection",
  CellTags->"b:0.3.5"],
Cell[108295, 2582, 300, 6, 42, "Text",
  CellTags->"b:0.3.5"],
Cell[108598, 2590, 1801, 44, 347, "Text",
  CellTags->"b:0.3.5"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

