(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 4.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      9191,        215]*)
(*NotebookOutlinePosition[      9897,        239]*)
(*  CellTagsIndexPosition[      9853,        235]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Amazon Manufacturer Web Services", "Title"],

Cell["\<\
This example demonstrates the Amazon Web Services.  The Amazon Web Services \
are good demenstrations of using Web Services to retrieve data.  This example \
searches the Amazon databases for books published by certain manufacturers.  \
The example code then places the results in the notebook.  The example has \
one option.  Setting the ShowPictures option to True will force the code to \
retrieve and display the cover image of the book.\
\>", "Text"],

Cell["\<\
To try this example, evaluate all the initialization cells (you can do this \
with the menu command Kernel/Evaluation/Evaluate Initialization). Then go to \
the Example section.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Code", "Section"],

Cell[BoxData[{
    \(\(\(BeginPackage["\<WebServices`Examples`Amazon`\>", \
{"\<WebServices`\>", "\<JLink`\>"}];\)\(\n\)
    \)\), "\[IndentingNewLine]", 
    \(\(ManufacturerSearch;\)\), "\n", 
    \(\(\(ShowPictures;\)\(\n\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(Begin["\<`Private`\>"];\)\(\[IndentingNewLine]\)
    \)\), "\n", 
    \(\(\(InstallService["\<http://soap.amazon.com/schemas/AmazonWebServices.\
wsdl\>"];\)\(\[IndentingNewLine]\)
    \)\), "\n", 
    \(\(\(Options[
          AuthorSearch] = {ShowPictures \[Rule] 
            False};\)\(\[IndentingNewLine]\)
    \)\), "\n", 
    \(\(\(ManufacturerSearch[keyword_String, pages_Integer, 
        options___?OptionQ] := 
      Module[{result}, 
        result = 
          ManufacturerSearchRequest[{"\<manufacturer\>" \[Rule] 
                keyword, "\<page\>" \[Rule] 
                pages, "\<mode\>" \[Rule] "\<books\>", "\<tag\>" \[Rule] \
"\<webservices-20\>", "\<type\>" \[Rule] "\<lite\>", "\<devtag\>" \[Rule] \
"\<D3VZJS46JD5U9D\>", "\<version\>" \[Rule] "\<1.0\>"}]; \
\[IndentingNewLine]If[result === $Failed, 
          Return[$Failed]]; \[IndentingNewLine]result = "\<Details\>" /. 
            result; \[IndentingNewLine]notebook = 
          EvaluationNotebook[]; \[IndentingNewLine]If[
          NotebookFind[notebook, "\<Results\>", All, {CellTags}] =!= $Failed, 
          SelectionMove[notebook, All, 
            CellGroup]; \[IndentingNewLine]NotebookDelete[
            notebook]]; \[IndentingNewLine]SelectionMove[notebook, After, 
          Notebook]; \[IndentingNewLine]NotebookWrite[notebook, 
          Cell["\<Results\>", "\<Section\>", 
            CellTags \[Rule] {"\<Results\>"}]]; \
\[IndentingNewLine]\(printDetails[#, options] &\) /@ result;]\)\(\n\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(printDetails[d_List, options___?OptionQ] := 
      Module[{url, asin, productName, authorNames, releaseDate, manufacturer, 
          imageUrlMedium, listPrice, ourPrice, usedPrice, 
          showPictures}, {showPictures} = \({ShowPictures} /. 
              Flatten[{options}]\) /. 
            Options[AuthorSearch]; \[IndentingNewLine]url = "\<Url\>" /. 
            d; \[IndentingNewLine]asin = "\<Asin\>" /. 
            d; \[IndentingNewLine]productName = "\<ProductName\>" /. 
            d; \[IndentingNewLine]authorNames = "\<Authors\>" /. 
            d; \[IndentingNewLine]releaseDate = "\<ReleaseDate\>" /. 
            d; \[IndentingNewLine]manufacturer = "\<Manufacturer\>" /. 
            d; \[IndentingNewLine]imageUrlMedium = "\<ImageUrlMedium\>" /. 
            d; \[IndentingNewLine]listPrice = "\<ListPrice\>" /. 
            d; \[IndentingNewLine]ourPrice = "\<OurPrice\>" /. 
            d; \[IndentingNewLine]usedPrice = "\<UsedPrice\>" /. 
            d; \[IndentingNewLine]authorNames = 
          ToString[authorNames]; \[IndentingNewLine]authorNames = 
          StringDrop[authorNames, 1]; \[IndentingNewLine]authorNames = 
          StringDrop[authorNames, \(-1\)]; \[IndentingNewLine]If[
          TrueQ[showPictures], 
          imageUrlMedium = 
            getImageData[imageUrlMedium]; \[IndentingNewLine]If[
            StringMatchQ[imageUrlMedium, "\<GIF*\>"], 
            imageCell = 
              ToExpression[
                ExportString[
                  ImportString[imageUrlMedium, "\<GIF\>"], "\<MGFCell\>"]], 
            imageCell = 
              ToExpression[
                ExportString[
                  ImportString[
                    imageUrlMedium, "\<JPEG\>"], "\<MGFCell\>"]]];]; \
\[IndentingNewLine]notebook = 
          EvaluationNotebook[]; \[IndentingNewLine]NotebookFind[
          notebook, "\<Results\>", 
          All, {CellTags}]; \[IndentingNewLine]SelectionMove[notebook, After, 
          Notebook]; \[IndentingNewLine]NotebookWrite[notebook, 
          Cell[CellGroupData[{Cell[
                  BoxData[
                    RowBox[{StyleBox[productName], 
                        StyleBox["\< by \>", FontWeight \[Rule] "\<Plain\>"], 
                        StyleBox[authorNames, FontWeight \[Rule] "\<Plain\>", 
                          FontSlant \[Rule] "\<Italic\>"], 
                        ButtonBox["\<(link)\>", 
                          ButtonData \[Rule] {URL[url], None}, 
                          ButtonStyle \[Rule] "\<Hyperlink\>"]}]], \
"\<Subsection\>"], 
                Cell[BoxData[
                    RowBox[{StyleBox["\<Release Date: \>", 
                          FontWeight \[Rule] "\<Bold\>"], 
                        StyleBox[releaseDate]}]]], 
                Cell[BoxData[
                    RowBox[{StyleBox["\<ASIN: \>", 
                          FontWeight \[Rule] "\<Bold\>"], StyleBox[asin]}]]], 
                Cell[BoxData[
                    RowBox[{StyleBox["\<List Price: \>", 
                          FontWeight \[Rule] "\<Bold\>"], 
                        StyleBox[listPrice]}]]], 
                Cell[BoxData[
                    RowBox[{StyleBox["\<Our Price: \>", 
                          FontWeight \[Rule] "\<Bold\>"], 
                        StyleBox[ourPrice]}]]]}, 
              Closed]]]; \[IndentingNewLine]If[
          TrueQ[showPictures], \(NotebookWrite[notebook, 
              imageCell];\)];]\)\(\n\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(getImageData[url_String] := 
      JavaBlock[
        Module[{u, connection, stream, input = "\<\>", numRead = 0, 
            bytesRead = 0, buf}, 
          u = JavaNew["\<java.net.URL\>", 
              url]; \[IndentingNewLine]connection = 
            u@openConnection[]; \[IndentingNewLine]stream = 
            connection@getInputStream[]; \[IndentingNewLine]If[
            stream \[Equal] $Failed, 
            Return[URLError["\<Failed to open URL stream\>"]]]; \
\[IndentingNewLine]buf = 
            JavaNew["\<[B\>", 
              5000]; \[IndentingNewLine]While[\((numRead = 
                  stream@read[buf])\) > 0, 
            AddTo[bytesRead, numRead]; \[IndentingNewLine]input = 
              StringJoin[input, 
                ToString[
                  FromCharacterCode[\(If[# < 0, # + 256, #] &\) /@ 
                      Take[JavaObjectToExpression[buf], 
                        numRead]]]];]; \[IndentingNewLine]stream@
            close[]; \[IndentingNewLine]input]]\)\(\n\)
    \)\), "\[IndentingNewLine]", 
    \(\(End[];\)\), "\n", 
    \(\(EndPackage[];\)\)}], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples", "Section"],

Cell[BoxData[
    \(ManufacturerSearch["\<Addison-Wesley\>", 1, 
      ShowPictures \[Rule] True]\)], "Input"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.2 for Microsoft Windows",
ScreenRectangle->{{0, 1152}, {0, 819}},
AutoGeneratedPackage->None,
WindowSize->{600, 634},
WindowMargins->{{14, Automatic}, {Automatic, 0}},
StyleDefinitions -> "Default.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 49, 0, 170, "Title"],
Cell[1828, 55, 465, 7, 90, "Text"],
Cell[2296, 64, 201, 4, 52, "Text"],

Cell[CellGroupData[{
Cell[2522, 72, 23, 0, 59, "Section"],
Cell[2548, 74, 6438, 128, 2050, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[9023, 207, 27, 0, 39, "Section"],
Cell[9053, 209, 110, 2, 30, "Input"]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

