(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 4.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     26857,        516]*)
(*NotebookOutlinePosition[     27515,        539]*)
(*  CellTagsIndexPosition[     27471,        535]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["TerraService Java UI", "Title"],

Cell["\<\
This example demonstrates the TerraService from Microsoft using a Java UI.  \
Microsoft TerraServer Web Service is a programmable interface to the popular \
Microsoft TerraServer on-line database of high resolution USGS aerial imagery \
(DOQs) and scanned USGS topographical maps (DRGs).  The user can specify the \
place to display by using the \"Place Search\" field.  The place is expected \
to be a single string containing the city or well known place name, state \
name, and country name separated by comma characters, e.g. \"San Franicisco, \
CA, USA\". Any of the three values may be missing.  The navigation panel can \
be used to move in a particular direction.  The \"Latitude\" and \
\"Longitude\" fields can be used to specify a place at a particular latitude \
and longitude.  The user can zoom-in and zoom-out using the \"Scale\" drop \
down menu.  The user can switch between aerial photographs and topological \
maps using the \"Theme\" drop down menu.  \
\>", "Text"],

Cell[TextData[{
  "More information about the TerraService may be found at ",
  ButtonBox["http://terraservice.net/.",
    ButtonData:>{
      URL[ "http://terraservice.net/"], None},
    ButtonStyle->"Hyperlink"]
}], "Text"],

Cell["\<\
To try this example, evaluate all the initialization cells (you can do this \
with the menu command Kernel/Evaluation/Evaluate Initialization). Then go to \
the Example section.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Code", "Section"],

Cell[BoxData[{
    \(\(\(BeginPackage["\<TerraService`\>", \ {"\<WebServices`\>", \ \
"\<JLink`\>"}];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(LaunchUI;\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(Begin["\<`Private`\>"];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(InstallService["\<http://terraservice.net/TerraService.asmx?WSDL \
\>"];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(getImage[id_List] := \ \(getImage[id] = 
          toolkit@createImage[
              Base64`decode[\(InvokeServiceOperation[GetTile, \ id]\)[\([2, 
                    3, 1, 3, 1, 3, 1, 3, 1]\)]]]\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(resetGetImage[] := \[IndentingNewLine]\((\[IndentingNewLine]Clear[
            getImage]; \[IndentingNewLine]getImage[
              id_List] := \ \(getImage[id] = 
              toolkit@createImage[
                  Base64`decode[\(InvokeServiceOperation[GetTile, id]\)[\([2, 
                        3, 1, 3, 1, 3, 1, 3, 
                        1]\)]]]\);\[IndentingNewLine])\);\)\(\
\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(getMetaFromLonLat[lat_, lon_, \ theme_String, \ 
          scale_String]\  := \ \(getMetaFromLonLat[lat, \ lon, \ theme, \ 
            scale]\  = \ 
          GetTileMetaFromLonLatPt[{"\<Lon\>" \[Rule] 
                longitude, "\<Lat\>" \[Rule] latitude}, theme, 
            scale]\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(getPlace[plc_String]\  := \ \(getPlace[plc]\  = \ 
          GetPlaceList[plc, 1, True]\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(lonLatToPlace[lon_, \ lat_]\  := \ \(lonLatToPlace[lon, lat]\  = \ 
          ConvertLonLatPtToNearestPlace[{"\<Lon\>" \[Rule] 
                lon, "\<Lat\>" \[Rule] lat}]\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(getMetaFromTileId[t_String, \ scl_String, \ scn_, \ x1_, 
          y1_]\  := \ \(getMetaFromTileId[t, \ scl, \ scn, \ x1, y1]\ \  = 
          GetTileMetaFromTileId[{"\<Theme\>" \[Rule] t, "\<Scale\>" \[Rule] 
                scl, "\<Scene\>" \[Rule] scn, "\<X\>" \[Rule] 
                x1, "\<Y\>" \[Rule] y1}]\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(updateMeta[] := \[IndentingNewLine]Module[{meta}, \
\[IndentingNewLine]meta = 
            getMetaFromLonLat[latitude, \ longitude, theme, 
              scale]; \[IndentingNewLine]x = "\<X\>" /. \(("\<Id\>" /. 
                  meta)\); \[IndentingNewLine]y = "\<Y\>" /. \(("\<Id\>" /. 
                  meta)\); \[IndentingNewLine]scene = "\<Scene\>" /. \(("\<Id\
\>" /. meta)\);\[IndentingNewLine]];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(setPlace[
          p_String] := \[IndentingNewLine]Module[{result, \ latLon, \ 
            plc}, \[IndentingNewLine]result = 
            getPlace[p]; \[IndentingNewLine]If[
            result =!= Null, \[IndentingNewLine]latLon = "\<Center\>" /. 
                First[result]; \[IndentingNewLine]latitude = "\<Lat\>" /. 
                latLon; \[IndentingNewLine]longitude = "\<Lon\>" /. 
                latLon; \[IndentingNewLine]plc\  = \ "\<Place\>" /. 
                First[result]; \[IndentingNewLine]location = \(("\<City\>" /. 
                    plc)\)\  <> \ "\<, \>"\  <> \ \(("\<State\>" /. 
                    plc)\)\  <> "\<, \>"\  <> \(("\<Country\>" /. 
                    plc)\)\ ;\[IndentingNewLine]]; \
\[IndentingNewLine]location\[IndentingNewLine]];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(setLatLon[lat_, \ 
          lon_] := \((\[IndentingNewLine]latitude = 
            lat; \[IndentingNewLine]longitude\  = \ 
            lon; \[IndentingNewLine]location = 
            lonLatToPlace[longitude, 
              latitude];\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(setScale[
          s_String] := \((\[IndentingNewLine]\(scale\  = 
              s;\)\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(setTheme[
          t_String] := \((\[IndentingNewLine]\(theme\  = \ 
              t;\)\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(updateXY[]\  := \[IndentingNewLine]Module[{meta}, \
\[IndentingNewLine]meta\  = \ 
            getMetaFromTileId[theme, \ scale, scene, x, 
              y]; \[IndentingNewLine]latitude\  = \ "\<Lat\>"\  /. \
\(("\<Center\>" /. meta)\); \[IndentingNewLine]latField@
            setText[ToString[
                latitude]]; \[IndentingNewLine]longitude\  = \ "\<Lon\>"\  /. \
\(("\<Center\>" /. meta)\); \[IndentingNewLine]lonField@
            setText[ToString[longitude]]; \[IndentingNewLine]location\  = \ 
            lonLatToPlace[longitude, latitude]; \[IndentingNewLine]frame@
            setTitle[location];\[IndentingNewLine]];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(LaunchUI[] \
:= \[IndentingNewLine]JavaBlock[\[IndentingNewLine]\(Module[{frm}, \
\[IndentingNewLine]frm = createWindow[]; \[IndentingNewLine]frm@
                setModal[]; \[IndentingNewLine]DoModal[]; \
\[IndentingNewLine]resetGetImage[];\[IndentingNewLine]];\)\[IndentingNewLine]]\
;\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(moveWest[] := \((\[IndentingNewLine]\(x--\); \
\[IndentingNewLine]setImages[]; \[IndentingNewLine]updateXY[];\
\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(moveEast[] := \((\[IndentingNewLine]\(x++\); \
\[IndentingNewLine]setImages[]; \[IndentingNewLine]updateXY[];\
\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(moveNorth[] := \((\[IndentingNewLine]\(y++\); \
\[IndentingNewLine]setImages[]; \[IndentingNewLine]updateXY[];\
\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(moveSouth[] := \((\[IndentingNewLine]\(y--\); \
\[IndentingNewLine]setImages[]; \[IndentingNewLine]updateXY[];\
\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(moveNorthWest[]\  := \((\[IndentingNewLine]\(x--\); \
\[IndentingNewLine]\(y++\); \[IndentingNewLine]setImages[]; \
\[IndentingNewLine]updateXY[];\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)\

    \)\), "\[IndentingNewLine]", 
    \(\(\(moveNorthEast[]\  := \((\[IndentingNewLine]\(x++\); \
\[IndentingNewLine]\(y++\); \[IndentingNewLine]setImages[]; \
\[IndentingNewLine]updateXY[];\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)\

    \)\), "\[IndentingNewLine]", 
    \(\(\(moveSouthWest[]\  := \((\[IndentingNewLine]\(x--\); \
\[IndentingNewLine]\(y--\); \[IndentingNewLine]setImages[]; \
\[IndentingNewLine]updateXY[];\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)\

    \)\), "\[IndentingNewLine]", 
    \(\(\(moveSouthEast[]\  := \((\[IndentingNewLine]\(x++\); \
\[IndentingNewLine]\(y--\); \[IndentingNewLine]setImages[]; \
\[IndentingNewLine]updateXY[];\[IndentingNewLine])\);\)\(\[IndentingNewLine]\)\

    \)\), "\[IndentingNewLine]", 
    \(\(x\  = \ 0;\)\), "\[IndentingNewLine]", 
    \(\(y\  = \ 0;\)\), "\[IndentingNewLine]", 
    \(\(scene\  = \ 0;\)\), "\[IndentingNewLine]", 
    \(\(theme\  = \ "\<Photo\>";\)\), "\[IndentingNewLine]", 
    \(\(\(scale\  = \ "\<Scale16m\>";\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(setPlace["\<Champaign, IL\>"];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(LoadJavaClass["\<org.apache.axis.encoding.Base64\>"];\)\), "\
\[IndentingNewLine]", 
    \(\(LoadJavaClass["\<java.awt.Toolkit\>"];\)\), "\[IndentingNewLine]", 
    \(\(\(toolkit\  = \ Toolkit`getDefaultToolkit[];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(createCanvas[]\  := \ \[IndentingNewLine]Module[{canvas}, \
\[IndentingNewLine]canvas\  = \ 
            JavaNew["\<com.wolfram.jlink.MathCanvas\>"]; \
\[IndentingNewLine]canvas@
            setSize[200, 
              200]; \[IndentingNewLine]canvas\[IndentingNewLine]];\)\(\
\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(createWindow[] := \[IndentingNewLine]Module[{gb, gbc, controlArea, 
            mapArea, \ navigationalPanel, \ nwButton, \ nButton, \ 
            neButton, \ wButton, \ eButton, \ swButton, \ sButton, \ 
            seButton, nwListener, \ nListener, \ neListener, \ wListener, \ 
            eListener, \ swListener, \ sListener, \ seListener, \ 
            navButton, \ placeLabel, \ latLabel, \ lonLabel, \ themeLabel, \ 
            scaleLabel, \ submitButton, \ 
            submitListener}, \[IndentingNewLine]InstallJava[]; \
\[IndentingNewLine]frame = 
            JavaNew["\<com.wolfram.jlink.MathFrame\>", 
              location]; \[IndentingNewLine]frame@
            setResizable[False]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Set\ background\ to\ black\ *) \[IndentingNewLine]LoadJavaClass["\
\<java.awt.Color\>"]; \[IndentingNewLine] (*\[IndentingNewLine]\(frame@
                setBackground[
                  Color`BLACK];\)\[IndentingNewLine]*) \[IndentingNewLine] \
(*\ Map\ Area\ panel\ setup\ *) \[IndentingNewLine]mapArea\  = \ 
            JavaNew["\<java.awt.Panel\>"]; \[IndentingNewLine]mapArea@
            setSize[600, 400]; \[IndentingNewLine]mapArea@
            setLayout[
              JavaNew["\<java.awt.GridLayout\>", 2, 3, 0, 
                0]]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ canvases\ for\ images\ *) \ \[IndentingNewLine]canvases\  \
= \ {}; \[IndentingNewLine]Do[
            canvas\  = createCanvas[]; \[IndentingNewLine]mapArea@
              add[canvas]; \[IndentingNewLine]AppendTo[canvases, \ 
              canvas];, \ {6}]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Initialize\ images\ *) \ \[IndentingNewLine]updateMeta[]; \
\[IndentingNewLine]setImages[]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Control\ Area\ panel\ setup\ *) \[IndentingNewLine]controlArea\  \
= \ JavaNew["\<java.awt.Panel\>"]; \[IndentingNewLine]controlArea@
            setSize[200, 400]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            set\ gridbag\ layout\ *) \[IndentingNewLine]gb\  = \ 
            JavaNew["\<java.awt.GridBagLayout\>"]; \[IndentingNewLine]gbc\  = \
\ JavaNew["\<java.awt.GridBagConstraints\>"]; \[IndentingNewLine]controlArea@
            setLayout[gb]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Navigation\ Panel\ *) \[IndentingNewLine]navigationPanel\  = \ 
            JavaNew["\<java.awt.Panel\>"]; \[IndentingNewLine]navigationPanel@
            setLayout[
              JavaNew["\<java.awt.GridLayout\>", 3, 3, 0, 
                0]]; \[IndentingNewLine]gbc@
              anchor\  = \ \ GridBagConstraints`NORTHWEST; \
\[IndentingNewLine]gbc@gridwidth\  = \ 
            GridBagConstraints`REMAINDER; \[IndentingNewLine]gb@
            setConstraints[navigationPanel, \ 
              gbc]; \[IndentingNewLine]controlArea@
            add[navigationPanel]; \
\[IndentingNewLine]\[IndentingNewLine]nwButton\  = \ 
            JavaNew["\<java.awt.Button\>", "\<NW\>"]; \
\[IndentingNewLine]nwListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]nwListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveNorthWest[]\>"]; \[IndentingNewLine]nwButton@
            addActionListener[nwListener]; \[IndentingNewLine]navigationPanel@
            add[nwButton]; \[IndentingNewLine]\[IndentingNewLine]nButton\  = \
\ JavaNew["\<java.awt.Button\>", "\<N\>"]; \[IndentingNewLine]nListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]nListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveNorth[]\>"]; \[IndentingNewLine]nButton@
            addActionListener[nListener]; \[IndentingNewLine]navigationPanel@
            add[nButton]; \[IndentingNewLine]\[IndentingNewLine]neButton = 
            JavaNew["\<java.awt.Button\>", "\<NE\>"]; \
\[IndentingNewLine]neListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]neListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveNorthEast[]\>"]; \[IndentingNewLine]neButton@
            addActionListener[neListener]; \[IndentingNewLine]navigationPanel@
            add[neButton]; \[IndentingNewLine]\[IndentingNewLine]wButton = 
            JavaNew["\<java.awt.Button\>", "\<W\>"]; \
\[IndentingNewLine]wListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]wListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveWest[]\>"]; \[IndentingNewLine]wButton@
            addActionListener[wListener]; \[IndentingNewLine]navigationPanel@
            add[wButton]; \[IndentingNewLine]\[IndentingNewLine]navButton = 
            JavaNew["\<java.awt.Button\>", "\<Navigation\>"]; \
\[IndentingNewLine]navigationPanel@
            add[navButton]; \[IndentingNewLine]\[IndentingNewLine]eButton = 
            JavaNew["\<java.awt.Button\>", "\<E\>"]; \
\[IndentingNewLine]eListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]eListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveEast[]\>"]; \[IndentingNewLine]eButton@
            addActionListener[eListener]; \[IndentingNewLine]navigationPanel@
            add[eButton]; \[IndentingNewLine]\[IndentingNewLine]swButton = 
            JavaNew["\<java.awt.Button\>", "\<SW\>"]; \
\[IndentingNewLine]swListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]swListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveSouthWest[]\>"]; \[IndentingNewLine]swButton@
            addActionListener[swListener]; \[IndentingNewLine]navigationPanel@
            add[swButton]; \[IndentingNewLine]\[IndentingNewLine]sButton = 
            JavaNew["\<java.awt.Button\>", "\<S\>"]; \
\[IndentingNewLine]sListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]sListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveSouth[]\>"]; \[IndentingNewLine]sButton@
            addActionListener[sListener]; \[IndentingNewLine]navigationPanel@
            add[sButton]; \[IndentingNewLine]\[IndentingNewLine]seButton = 
            JavaNew["\<java.awt.Button\>", "\<SE\>"]; \
\[IndentingNewLine]seListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]seListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`moveSouthEast[]\>"]; \[IndentingNewLine]seButton@
            addActionListener[seListener]; \[IndentingNewLine]navigationPanel@
            add[seButton]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ place\ label\ *) \[IndentingNewLine]placeLabel\  = \ 
            
            JavaNew\ ["\<java.awt.Label\>", \ "\<Place Search\>"]; \
\[IndentingNewLine]gbc@gridwidth\  = \ 
            GridBagConstraints`REMAINDER; \[IndentingNewLine]gb@
            setConstraints[placeLabel, \ gbc]; \[IndentingNewLine]controlArea@
            add[placeLabel]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ place\ text\ field\ *) \[IndentingNewLine]placeField\  = \
\ JavaNew["\<java.awt.TextField\>", \ 25]; \[IndentingNewLine]gb@
            setConstraints[placeField, \ gbc]; \[IndentingNewLine]controlArea@
            add[placeField]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ latitude\ label\ *) \[IndentingNewLine]latLabel\  = \ 
            JavaNew\ ["\<java.awt.Label\>", \ "\<Latitude\>"]; \
\[IndentingNewLine]gb@
            setConstraints[latLabel, \ gbc]; \[IndentingNewLine]controlArea@
            add[latLabel]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ latitude\ text\ field\ *) \[IndentingNewLine]latField\  = \
\ JavaNew["\<java.awt.TextField\>", \ ToString[latitude], \ 
              25]; \[IndentingNewLine]gb@
            setConstraints[latField, \ gbc]; \[IndentingNewLine]controlArea@
            add[latField]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ longitude\ label\ *) \[IndentingNewLine]lonLabel\  = \ 
            JavaNew\ ["\<java.awt.Label\>", \ "\<Longitude\>"]; \
\[IndentingNewLine]gb@
            setConstraints[lonLabel, \ gbc]; \[IndentingNewLine]controlArea@
            add[lonLabel]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ longitude\ text\ field\ *) \[IndentingNewLine]lonField\  \
= \ JavaNew["\<java.awt.TextField\>", \ ToString[longitude], \ 
              25]; \[IndentingNewLine]gb@
            setConstraints[lonField, \ gbc]; \[IndentingNewLine]controlArea@
            add[lonField]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ theme\ label\ *) \[IndentingNewLine]themeLabel\  = \ 
            JavaNew\ ["\<java.awt.Label\>", \ "\<Theme\>"]; \
\[IndentingNewLine]gbc@gridwidth\  = \ 
            GridBagConstraints`RELATIVE; \[IndentingNewLine]gb@
            setConstraints[themeLabel, \ gbc]; \[IndentingNewLine]controlArea@
            add[themeLabel]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ theme\ choice\ *) \[IndentingNewLine]themeChoice\  = \ 
            JavaNew["\<java.awt.Choice\>"]; \[IndentingNewLine]themeChoice@
            add["\<Aerial Photo\>"]; \[IndentingNewLine]themeChoice@
            add["\<Topo Map\>"]; \[IndentingNewLine]Switch[
            theme, \[IndentingNewLine]"\<Photo\>", 
            themeChoice@
              select["\<Aerial Photo\>"], \[IndentingNewLine]"\<Topo\>", \ 
            themeChoice@select["\<Topo Map\>"]]; \[IndentingNewLine]gbc@
              gridwidth\  = \ 
            GridBagConstraints`REMAINDER; \[IndentingNewLine]gb@
            setConstraints[themeChoice, \ 
              gbc]; \[IndentingNewLine]controlArea@
            add[themeChoice]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ scale\ label\ *) \[IndentingNewLine]scaleLabel\  = \ 
            JavaNew\ ["\<java.awt.Label\>", \ "\<Scale\>"]; \
\[IndentingNewLine]gbc@gridwidth\  = \ 
            GridBagConstraints`RELATIVE; \[IndentingNewLine]gb@
            setConstraints[scaleLabel, \ gbc]; \[IndentingNewLine]controlArea@
            add[scaleLabel]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            create\ theme\ choice\ *) \[IndentingNewLine]scaleChoice\  = \ 
            JavaNew["\<java.awt.Choice\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale64m\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale32m\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale16m\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale8m\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale4m\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale2m\>"]; \[IndentingNewLine]scaleChoice@
            add["\<Scale1m\>"]; \[IndentingNewLine]scaleChoice@
            select[scale]; \[IndentingNewLine]gbc@gridwidth\  = \ 
            GridBagConstraints`REMAINDER; \[IndentingNewLine]gb@
            setConstraints[scaleChoice, \ 
              gbc]; \[IndentingNewLine]controlArea@
            add[scaleChoice]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            submit\ button\ *) \[IndentingNewLine]submitButton = 
            JavaNew["\<java.awt.Button\>", "\<Submit\>"]; \
\[IndentingNewLine]submitListener = 
            JavaNew["\<com.wolfram.jlink.MathActionListener\>"]; \
\[IndentingNewLine]submitListener@
            setHandler["\<actionPerformed\>", \
"\<TerraService`Private`update[]\>"]; \[IndentingNewLine]submitButton@
            addActionListener[submitListener]; \[IndentingNewLine]controlArea@
            add[submitButton]; \[IndentingNewLine]\[IndentingNewLine]\ 
          frame@setLayout[
              JavaNew["\<java.awt.BorderLayout\>"]]; \[IndentingNewLine]frame@
            add[mapArea, 
              ReturnAsJavaObject[
                BorderLayout`EAST]]; \[IndentingNewLine]frame@
            add[controlArea, 
              ReturnAsJavaObject[
                BorderLayout`WEST]]; \[IndentingNewLine]frame@
            pack[]; \[IndentingNewLine]\[IndentingNewLine]frame@
            setLocation[100, 100]; \[IndentingNewLine]JavaShow[
            frame]; \[IndentingNewLine]frame\[IndentingNewLine]];\)\
\[IndentingNewLine]\), "\[IndentingNewLine]", 
    \(\(setImages[] := \[IndentingNewLine]Module[{i}, \[IndentingNewLine]i\  \
= \ 1; \[IndentingNewLine]Outer[\((canvases[\([i]\)]@
                  setImage[
                    getImage[{"\<Theme\>" \[Rule] theme, "\<Scale\>" \[Rule] 
                          scale, "\<Scene\>" \[Rule] 
                          scene, "\<Y\>" \[Rule] #1, "\<X\>" \[Rule] #2}]]; \
\(i++\);)\)\  &, \ Reverse[Range[y, y + 1]], 
            Range[x - 1, 
              x + 1]];\[IndentingNewLine]];\)\[IndentingNewLine]\), "\
\[IndentingNewLine]", 
    \(\(update[] := \ \[IndentingNewLine]Module[{lat, \ 
            lon}, \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Update\ place\ *) \[IndentingNewLine]If[\(! StringMatchQ[
                placeField@getText[], "\<\>"]\), \ \[IndentingNewLine]frame@
              setTitle[
                setPlace[
                  placeField@getText[]]]; \[IndentingNewLine]placeField@
              setText["\<\>"]; \[IndentingNewLine]latField@
              setText[ToString[latitude]]; \[IndentingNewLine]lonField@
              setText[ToString[longitude]], \[IndentingNewLine]lat = \ 
              ToExpression[latField@getText[], InputForm, \ 
                HoldComplete]; \[IndentingNewLine]lon = \ 
              ToExpression[lonField@getText[], InputForm, \ 
                HoldComplete]; \[IndentingNewLine]If[
              MatchQ[lat, \ 
                  HoldComplete[_Integer\  | \ _Real]]\  \
&& \[IndentingNewLine]MatchQ[lon, \ 
                  HoldComplete[_Integer\  | \ _Real]], \
\[IndentingNewLine]setLatLon[First[lat], \ 
                First[lon]]; \[IndentingNewLine]frame@
                setTitle[
                  location];\[IndentingNewLine]];\[IndentingNewLine]]; \
\[IndentingNewLine] (*\ Update\ theme\ *) \[IndentingNewLine]Switch[
            themeChoice@
              getSelectedItem[], \[IndentingNewLine]"\<Aerial Photo\>", 
            setTheme["\<Photo\>"], \[IndentingNewLine]"\<Topo Map\>", \ 
            setTheme["\<Topo\>"]]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Update\ scale\ *) \[IndentingNewLine]If[
            StringMatchQ[theme, "\<Topo\>"]\  && \ 
              StringMatchQ[
                scaleChoice@
                  getSelectedItem[], "\<Scale1m\>"], \
\[IndentingNewLine]\(scaleChoice@
                select["\<Scale2m\>"];\)\[IndentingNewLine]]; \
\[IndentingNewLine]setScale[
            scaleChoice@
              getSelectedItem[]]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
            Update\ images\ *) \[IndentingNewLine]updateMeta[]; \
\[IndentingNewLine]setImages[];\[IndentingNewLine]];\)\), \
"\[IndentingNewLine]", 
    \(\(End[];\)\), "\[IndentingNewLine]", 
    \(\(EndPackage[];\)\)}], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples", "Section"],

Cell[BoxData[
    \(\(LaunchUI[];\)\)], "Input"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.2 for X",
ScreenRectangle->{{0, 1024}, {0, 768}},
AutoGeneratedPackage->None,
WindowSize->{664, 565},
WindowMargins->{{Automatic, 172}, {Automatic, 10}}
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 37, 0, 114, "Title"],
Cell[1816, 55, 995, 14, 176, "Text"],
Cell[2814, 71, 225, 6, 32, "Text"],
Cell[3042, 79, 201, 4, 50, "Text"],

Cell[CellGroupData[{
Cell[3268, 87, 23, 0, 58, "Section"],
Cell[3294, 89, 23420, 415, 6491, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[26751, 509, 27, 0, 38, "Section"],
Cell[26781, 511, 48, 1, 27, "Input"]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

