(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 4.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      6427,        153]*)
(*NotebookOutlinePosition[      7101,        176]*)
(*  CellTagsIndexPosition[      7057,        172]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["XMethods Query Web Service ", "Title"],

Cell["\<\
This example demonstrates the XMethods Query Service.  XMethods.com is a web \
site that lists web services.  This example demonstrates using the XMethods \
Query Service to discover and use many of the web services available.  This \
example searches the XMethods database for the web services registered.  The \
example code then places the results in the notebook.  An InstallService \
function is placed in each of the Cells created for each service.  Once the \
query is finished, a user can conveniently browse through the services and \
install the services that may be interesting.\
\>", "Text"],

Cell["\<\
To try this example, evaluate all the initialization cells (you can do this \
with the menu command Kernel/Evaluation/Evaluate Initialization). Then go to \
the Example section.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Code", "Section"],

Cell[BoxData[{
    \(\(\(BeginPackage["\<WebServices`Examples`XMethods`\>", {"\<WebServices`\
\>", "\<JLink`\>"}];\)\(\[IndentingNewLine]\)
    \)\), "\[IndentingNewLine]", 
    \(\(\(GetServiceSummaries;\)\(\n\)
    \)\), "\[IndentingNewLine]", 
    \(\(Begin["\<`Private`\>"];\)\[IndentingNewLine]\), "\n", 
    \(\(InstallService["\<http://www.xmethods.net/wsdl/query.wsdl\>"];\)\
\[IndentingNewLine]\), "\n", 
    \(GetServiceSummaries[] := 
      Module[{result}, 
        result = getAllServiceSummaries[]; \[IndentingNewLine]If[
          result === $Failed, Return[$Failed]]; \[IndentingNewLine]notebook = 
          EvaluationNotebook[]; \[IndentingNewLine]If[
          NotebookFind[notebook, "\<Results\>", All, {CellTags}] === $Failed, 
          SelectionMove[notebook, After, 
            EvaluationCell]; \[IndentingNewLine]NotebookWrite[notebook, 
            Cell["\<Results\>", "\<Section\>", 
              CellTags \[Rule] {"\<Results\>"}]];]; \
\[IndentingNewLine]printResult /@ result; \[IndentingNewLine]NotebookFind[
          notebook, "\<Results\>", 
          All, {CellTags}];]\[IndentingNewLine]\), "\n", 
    \(printResult[result_List] := 
      Module[{name, id, shortDescription, wsdlURL, publisherID}, 
        name = "\<name\>" /. result; \[IndentingNewLine]id = "\<id\>" /. 
            result; \[IndentingNewLine]shortDescription = "\<shortDescription\
\>" /. result; \[IndentingNewLine]wsdlURL = "\<wsdlURL\>" /. 
            result; \[IndentingNewLine]publisherID = "\<publisherID\>" /. 
            result; \[IndentingNewLine]notebook = 
          EvaluationNotebook[]; \[IndentingNewLine]If[
          NotebookFind[notebook, id, All, {CellTags}, 
              AutoScroll \[Rule] False] === $Failed, 
          NotebookFind[notebook, "\<Results\>", All, {CellTags}, 
            AutoScroll \[Rule] False]; \[IndentingNewLine]SelectionMove[
            notebook, After, Cell, 
            AutoScroll \[Rule] False]; \[IndentingNewLine]NotebookWrite[
            notebook, 
            Cell[CellGroupData[{Cell[
                    BoxData[
                      RowBox[{StyleBox[name], 
                          StyleBox["\< from \>", 
                            FontWeight \[Rule] "\<Plain\>"], 
                          StyleBox[publisherID, 
                            FontWeight \[Rule] "\<Plain\>", 
                            FontSlant \[Rule] "\<Italic\>"], 
                          ButtonBox["\<(wsdl)\>", 
                            ButtonData \[Rule] {URL[wsdlURL], None}, 
                            ButtonStyle \[Rule] "\<Hyperlink\>"]}]], \
"\<Subsection\>", ShowGroupOpenCloseIcon \[Rule] True, 
                    CellDingbat \[Rule] None, CellTags \[Rule] {id}], 
                  Cell[shortDescription, "\<Text\>", 
                    ShowCellBracket \[Rule] False, 
                    CellMargins \[Rule] {{55, Inherited}, {Inherited, 
                          Inherited}}, CellFrame \[Rule] True, 
                    Background \[Rule] GrayLevel[0.850004]], 
                  Cell[BoxData[
                      RowBox[{"\<InstallService[\"\>" <> 
                            wsdlURL <> "\<\"]\>"}]], "\<Input\>", 
                    ShowCellBracket \[Rule] False, CellFrame \[Rule] True, 
                    CellMargins \[Rule] {{55, Inherited}, {Inherited, 
                          Inherited}}, 
                    Background \[Rule] GrayLevel[0.850004]]}, 
                Open]]];];]\[IndentingNewLine]\), "\n", 
    \(\(End[];\)\), "\n", 
    \(\(EndPackage[];\)\)}], "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["Example", "Section"],

Cell[BoxData[
    \(GetServiceSummaries[]\)], "Input"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.2 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
AutoGeneratedPackage->None,
WindowSize->{647, 564},
WindowMargins->{{128, Automatic}, {59, Automatic}}
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 44, 0, 115, "Title"],
Cell[1823, 55, 613, 9, 109, "Text"],
Cell[2439, 66, 201, 4, 52, "Text"],

Cell[CellGroupData[{
Cell[2665, 74, 23, 0, 59, "Section"],
Cell[2691, 76, 3588, 65, 1010, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[6316, 146, 26, 0, 39, "Section"],
Cell[6345, 148, 54, 1, 30, "Input"]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

